PROGRAM_NAME='statistics'
(***********************************************************)
(*  FILE CREATED ON: 07/04/2010  AT: 14:31:26              *)
(***********************************************************)
(*  FILE_LAST_MODIFIED_ON: 07/22/2016  AT: 18:00:35        *)
(***********************************************************)
/*
 * tab-width: 2
 * indent-level: 1
 *
 */
DEFINE_PROGRAM                      // rev notes
(***********************************************************)
(*  FILE REVISION: 0.17/0.16                               *)
(*  REVISION DATE: 07/22/2016  AT: 18:00:35                *)
(*                                                         *)
(*  COMMENTS:                                              *)
(*  - Added 'clean up' parameter to mixedListToArray(). If *)
(*    false, the array return is not sorted and may have   *)
(*    duplicates.                                          *)
(*  - changed ATOI() to ATOF() in mixedListToArray().      *)
(*  - some formatting updates (indents, opening braces,    *)
(*    __STATISTICS_LIB_NAME__ length).                     *)
(*  - changed two parameter names in mixedListToArray().   *)
(*  - added round() function.                              *)
(*  - Other libs used:                                     *)
(*           strings.axi v2.14                             *)
(*                                                         *)
(***********************************************************)
(*  FILE REVISION: 0.15                                    *)
(*  REVISION DATE: 02/15/2016  AT: 17:01:56                *)
(*                                                         *)
(*  COMMENTS:                                              *)
(*  - Changed removeDuplicates() return type to INTEGER.   *)
(*  - Fixed range overflow issues in mixedListToArray().   *)
(*  - In rangeToList() the array updated is a parameter,   *)
(*    and return is the number of values in the list.      *)
(*  - Added DEFINE_PROGRAM to top of Rev Notes to enable   *)
(*    code folding the notes without hiding the timestamp. *)
(*  - strings.axi added as #INCLUDE.                       *)
(*                                                         *)
(***********************************************************)
(*  FILE REVISION: 0.14, 0.13, 0.12                        *)
(*  REVISION DATE: 01/11/2016  AT: 17:38:06                *)
(*                                                         *)
(*  COMMENTS:                                              *)
(*  - Added mixedListToArray().                            *)
(*  - Added rangeToList().                                 *)
(*  - Added LONG return value to removeDuplicates().       *)
(*  - Added removeDuplicates().                            *)
(*                                                         *)
(***********************************************************)
(*  FILE REVISION: 0.11                                    *)
(*  REVISION DATE: 08/17/2014  AT: 10:50:28                *)
(*                                                         *)
(*  COMMENTS:                                              *)
(*  - Added LEVtoA() for LEVEL.VALUE formatting.           *)
(*  - getMode() nested loop now only requires one pass,    *)
(*    instead of LENGTH_ARRAY(arr) passes.                 *)
(*  - Updated version check to support NI-x200 masters.    *)
(*  - Parameters and variables upgraded to DOUBLE and LONG *)
(*    from FLOAT and INTEGER.                              *)
(*  - Updated magic constant in getSquareRoot().           *)
(*  - Don't use FORMAT('%llu') in ULtoA(). This format     *)
(*    string works but has long been undocumented. It was  *)
(*    disclosed when testing DOUBLE values on the NI-x200  *)
(*    masters. Unlike ITOA() or ITOHEX(), FORMAT() doesn't *)
(*    do any type_cast of the passed in value so it will   *)
(*    give the decimal value of the IEEE 754 bits, not the *)
(*    expected value for ULtoA().                          *)
(*                                                         *)
(***********************************************************)
(*  FILE REVISION: 0.10                                    *)
(*  REVISION DATE: 05/11/2012  AT: 15:12:06                *)
(*                                                         *)
(*  COMMENTS:                                              *)
(*  - AtoUL() function now uses the first 10 contiguous    *)
(*    digits found in the string passed in.                *)
(*    This makes it easier to use, as the string is not    *)
(*    limited to digits only.                              *)
(*    The 10 digit limit is not a great loss of function,  *)
(*    the previous version would always return 2147483647  *)
(*    if the string was longer than 10 digits.             *)
(*  - Added ULtoA() function.                              *)
(*  - Commented out getSquareRootFast() and                *)
(*    getSquareRootSlow().                                 *)
(*                                                         *)
(***********************************************************)
(*  FILE REVISION: 0.09                                    *)
(*  REVISION DATE: 10/21/2010  AT: 17:41:54                *)
(*                                                         *)
(*  COMMENTS:                                              *)
(*  - If master version is 4.x.x or greater, getPower()    *)
(*    and getSquareRoot() act as wrappers for              *)
(*    POWER_VALUE and SQRT_VALUE().                        *)
(*                                                         *)
(***********************************************************)
(*  FILE REVISION: 0.08                                    *)
(*  REVISION DATE: 10/15/2010  AT: 17:32:33                *)
(*                                                         *)
(*  COMMENTS:                                              *)
(*  - getPower() no longer modifies its parameters.        *)
(*  - added AtoUL().                                       *)
(*                                                         *)
(***********************************************************)
(*  FILE REVISION: 0.07, 0.06, 0.05                        *)
(*  REVISION DATE: 08/05/2010  AT: 20:56:15                *)
(*                                                         *)
(*  COMMENTS:                                              *)
(*  - Added getPower() function.                           *)
(*  - Added getSquareRootFast() function.                  *)
(*  - Added improved getSquareRoot() function.             *)
(*    Renamed old version getSquareRootSlow().             *)
(*                                                         *)
(***********************************************************)
(*  FILE REVISION: 0.04, 0.03                              *)
(*  REVISION DATE: 07/15/2010  AT: 17:38:04                *)
(*                                                         *)
(*  COMMENTS:                                              *)
(*  - Added getSquareRoot() and getVariance()              *)
(*    functions.                                           *)
(*  - Renamed lib from sort.axi to statistics.axi as       *)
(*    working with averages is the initial purpose of      *)
(*    these subroutines.                                   *)
(*    Can split into seperate libs later if the need       *)
(*    arises.                                              *)
(*                                                         *)
(***********************************************************)
(*  FILE REVISION: 0.02                                    *)
(*  REVISION DATE: 07/07/2010  AT: 17:56:02                *)
(*                                                         *)
(*  COMMENTS:                                              *)
(* 	  - faster version of getMode(). Same logic but with   *)
(*      consolidated loops.                                *)
(* statistics and math related functions and calls:        *)
/* 1  function AtoUL (sNum[])                                                 */
/* 2  function getMode (arr[], ndxModes[])                                    */
/* 3  function getPower(x, n)                                                 */
/* 4  function getSquareRoot (f)                                              */
/* 5  function getSquareRootFast (f)                                          */
/* 6  function getSquareRootSlow (f)                                          */
/* 7  function getVariance(arr[], mean)                                       */
/* 8  function LEVtoA (num,flag)                                              */
/* 9  function mixedListToArray(ml[], rDelim[],rDelta, lDelim[], dA[],min,max)*/
/* 10 function quickSort (arr[], left, right)                                 */
/* 11 function rangeToList(r[], rDelim[], delta, lDelim[], sList[])           */
/* 12 function removeDuplicates (arr[], left, right)                          */
/* 13 function round (double)                                                 */
/* 14 function ULtoA (num)                                                    */

(***********************************************************)
#IF_NOT_DEFINED __statistics_lib__
#DEFINE __statistics_lib__

DEFINE_VARIABLE                     // #INCLUDE
#INCLUDE 'strings'

(***********************************************************)
(*               CONSTANT DEFINITIONS GO BELOW             *)
(***********************************************************)
DEFINE_CONSTANT

// VERSION FOR CODE

CHAR __STATISTICS_LIB_NAME__[20]     = '      statistics.axi';
CHAR __STATISTICS_LIB_VERSION__[]    = '0.17';

(***********************************************************)
(*               VARIABLE DEFINITIONS GO BELOW             *)
(***********************************************************)
DEFINE_VARIABLE

#IF_NOT_DEFINED __display_lib_version__
VOLATILE __display_lib_version__ = TRUE
#END_IF

#IF_NOT_DEFINED __juggernaut_bits__
VOLATILE INTEGER __juggernaut_bits__ = 0
#END_IF

(***********************************************************)
(*           SUBROUTINE DEFINITIONS GO BELOW               *)
(***********************************************************)
(* EXAMPLE: DEFINE_FUNCTION <RETURN_TYPE> <NAME> (<PARAMETERS>) *)
(* EXAMPLE: DEFINE_CALL '<NAME>' (<PARAMETERS>) *)

DEFINE_FUNCTION LONG AtoUL(CHAR str[])
/* function AtoUL()
 * purpose: Converts ASCII representation of a LONG decimal whole number value
 *          to an unsigned LONG number.
 *          Works with values > 2147483647:
 *            '2147483647' < sNum <= '4294967295'
 *
 *          10 digit values > 4294967295 are truncated to a 32 bit value,
 *          for example: A2UL('4294967297') returns 1
 * parameters:
 *      str[] - string to be converted
 * return:
 *      32 bit value of the decimal string
 */
{
 STACK_VAR LONG i                   // loop counter
 STACK_VAR LONG iLengthStr          // length of str
 STACK_VAR LONG digits              // length of first set of contiguous digits in str
 STACK_VAR CHAR sNum[10]            // storage for digits found in str

 iLengthStr = LENGTH_ARRAY(str)
 FOR (i = 1; i <= iLengthStr; i++)
 {
	IF ((str[i] >= '0') && (str[i] <= '9'))
	{
	 digits++
	 IF (digits > 10)
	 {
		SEND_STRING 0,"'statistics. axi ',ITOA(__LINE__),' AtoUL(): value truncated to 10 digits'"
		digits = 10
		break
	 }
	 ELSE
	 {
		sNum[digits] = str[i]
	 }
	}
	ELSE IF (digits) { break }        // if a non-digit charactor has been found after digits, stop
 }
 SET_LENGTH_ARRAY (sNum,digits)

 IF (digits == 10)                  // if sNum is possibly greater than 2147483647
 {
	RETURN ((ATOI("sNum[1]") * 1000000000) + ATOI(RIGHT_STRING(sNum,9)))
 }
 ELSE
 {
	RETURN (ATOI(sNum))
 }
}

DEFINE_FUNCTION LONG[2] getMode(DOUBLE arr[], LONG ndxModes[])
/* function getMode()
 * purpose: find the most often occuring value(s) in an array, and how many times they occured
 * parameters:
 *      arr[]    - sorted array of values
 *      ndxModes - array of indexes to unique mode values, populated by this function
 * return:
 *      return_value[1] - count of unique mode values
 *      return_value[2] - count of number of occurances of each unique mode value
 *
 * This subroutine greatly speed up not having i and j retrace the whole range multiple times.
 * In
 *
 *
 */
{
 STACK_VAR LONG i;
 STACK_VAR LONG j;
 STACK_VAR LONG m; // maximum count
 STACK_VAR LONG n; // number of unique values at count m
 STACK_VAR LONG len;
 STACK_VAR LONG counts[100000]
 STACK_VAR LONG return_value[2] // [<n>,<m>]

 m = 0                              // init max count
 n = 0                              // init modes count
 len = LENGTH_ARRAY(arr)

 for(i = len, j = len; i; i = j)    // count how many times we see each unique value
																	  /* speed tweaks (i,j tag team each other):
																		 * set j initial value here, do not reset j to max each pass of inner loop
																		 * at end of outer loop set i = j to bypass values already counted, instead of decrementing i by 1
																		 * thus break the inner loop if arr[j] <> arr[i], to restart outer loop at next unique value
																		 *
																		 */
 {
//SEND_STRING 0,"'i',ITOA(i),',j',ITOA(j)"
	counts[i] = 0                   // init counts

	for (; j; j--)                    // increment counts
	{
//SEND_STRING 0,"'j',ITOA(j)"
	 if(arr[j] == arr[i])
	 {
		IF (i >= j)                     // if value in j hasn't already been counted...
		{
//SEND_STRING 0,"ITOA(arr[j]),'@',ITOA(J)"
		  counts[i]++                   // ...increment count
																		// set max count, if necessary
		 if(counts[i] > m)  {m = counts[i]};
		}
		ELSE                            // else null duplicates and break
		{
//SEND_STRING 0,"'break@',ITOA(J)"
		 counts[i] = 0
		 break
		}
	 }
	 ELSE { BREAK }
	}
 }
 for(i = len; i; i--)
 {
	if(counts[i] = m)                  // if count[i] is the max count
	{
	 n++
	 ndxModes[n] = i;                 // store the index of each unique mode value found
	}
 }
 SET_LENGTH_ARRAY(ndxModes,n)

 return_value[1] = n
 return_value[2] = m
 SET_LENGTH_ARRAY(return_value,MAX_LENGTH_ARRAY(return_value))
 return return_value;
}

DEFINE_FUNCTION DOUBLE getPower (DOUBLE base, DOUBLE exponent)
/* function: getPower()
 * purpose:  calculates x^n power
 *           does not calcalute roots (n must be a whole number)
 * parameters:
 *     x     - initial value
 *     n     - whole number exponent (must be float so n/2 does not truncate)
 */
{
 IF (__juggernaut_bits__)           // If master supports POWER_VALUE()
 {
	RETURN POWER_VALUE(base,exponent)
 }
 ELSE
 {
	STACK_VAR DOUBLE x
	STACK_VAR DOUBLE n
	STACK_VAR DOUBLE z

	x = base
	n = exponent

	if (n < 0)
	{
	 n = -n
	 x = 1.0 / x
	}

	FOR (z = 1.0; n >= 1; n = n/2)
	{
	 if (n % 2 == 1)
	 {
		z = z*x
		}
		x = x*x
	}

	return z
 }
}

DEFINE_FUNCTION DOUBLE getSquareRoot(DOUBLE f)
/* function: getSquareRoot()
 * purpose: quickly get the square root of ABS_VALUE(f)
 *          based on Quake III implementation of bitwise operations on IEEE-754 floating point values.
 *          About as accurate as getSquareRootSlow(): FTOA(gSR()) usually gives same value as FTOA(gSRS()) but they sometimes differ at LSD
 *                                                    In those cases FTOA(gSR()) gives the same value as FTOA(SQRT_VALUE())
 *          ~1.5x-2x slower than getSquareRootFast(), ~5x-10x faster than getSquareRootSlow()
 * parameters:
 *     f - the initial value
 * returns:
 *     The square root of the initial value (may be approximate in some cases)
 */
{
 IF (__juggernaut_bits__)           // If master supports SQRT_VALUE()
 {
	RETURN (SQRT_VALUE(f))
 }
 ELSE IF ((f == 0) || (f == 1)) { return f }
 ELSE
 {
	STACK_VAR long tmp;
	STACK_VAR DOUBLE x;               // pre-compute f*0.5 to speed approximation steps
	STACK_VAR DOUBLE y;               // storage for inverse square root
	STACK_VAR DOUBLE sqrt;            // storage for square root
	STACK_VAR LONG i;

	y  = ABS_VALUE(f)
	x = y * 0.5

	tmp = hextoi(FORMAT('%x',y))      // hextoi(format('%x',f)) gives us the bits for a 64 bit float,
									  // but we only have 32 bit longs in NetLinx. 32 bits of mantissa truncated

	tmp  = $5fe6eb50 - ( tmp >> 1 )   // magic number is 64-bit float $5fe6ec85e7de30da, truncated (32-bit float algorithm use 0x5f3759df)
								  // magic number updated 1/14/2014: $5fe6eb50c7b537a9
									// Note: this trick doesn't work on NI-x200 masters, they're little-endian instead of big-endian
	y = ATOF(FORMAT('%f',tmp))        // atof(format('%tmp',tmp)) to convert 32bit long back to 64 bit float
									  // NOTE: this trick does not work on NI-x000 masters - they're little-endian instead of big-endian.

	WHILE (sqrt*sqrt <> f)
	{
	 y  = y * ( 1.5 - ( x * y * y ) ) // Newton approximation (error is typically ~1% to 2% after the first pass)
	 sqrt = 1/y                       //  square root = f * inverse square root is faster, but 1/y does not need the time to RETURN ABS_VALUE(sqrt)

	 i++
	 IF (i > 4) { BREAK }             // failsafe loop breaker
	}
	RETURN sqrt
 }
}

//DEFINE_FUNCTION DOUBLE getSquareRootFast(DOUBLE f)
///* function: getSquareRootFast()
// * purpose: quickly approximate the square root of ABS_VALUE(f),
// *          using bitwise operations on IEEE-754 floating point values.
// * parameters:
// *     f - the initial value
// * returns:
// *     The square root of the initial value, error ~ -3.5% to 3.5%
// */
//{
//  STACK_VAR long tmp
//
//  tmp = hextoi(FORMAT('%x',f))      // hextoi(format('%x',f)) gives us the bits for a 64 bit float,
//                                    // but we only have 32 bit longs in NetLinx. 32 bits of mantissa truncated
//  tmp = tmp & $7FFFFFFF             // clear sign bit to get ABS_VALUE(f)
//
//  tmp = tmp - (1<<20)               /* Remove last exponent bit so 1.0 gives 1.0 */
//  tmp = tmp >> 1                    /* divide by 2 */
//  tmp = tmp + (1<<29)               /* add 512 to exponent: (e+1023)/2 = (e/2)+511, */
//									/* that represents (e/2)-512 but we want e/2 */
//  tmp = tmp - $9800                 // error bias adjust = $4C000>>3
//  return  ATOF(FORMAT('%f',tmp))    // atof(format('%f',tmp)) to convert back to 64 bit float
//									  // NOTE: this trick does not work on NI-x000 masters - they're little-endian instead of big-endian.
//}

//DEFINE_FUNCTION DOUBLE getSquareRootSlow(DOUBLE f)
///* function: getSquareRootSlow()
// * purpose: calculates the square root of the ABS_VALUE(f)
// * parameters:
// *     f - the initial value
// * returns:
// *     The square root of the initial value (may be approximate in some cases)
// */
//{
//  IF ((f == 0) || (f == 1)) { return f }
//  ELSE
//  {
//	STACK_VAR DOUBLE a
//	STACK_VAR DOUBLE p
//	STACK_VAR DOUBLE k
//	STACK_VAR LONG i
//
//	k = ABS_VALUE(f)
//	a = k
//	p = a*a
//
//	while(FTOA(p) <> FTOA(k))
//	{
//	  i++
//	  a=(a+(k/a))/2
//	  p=a*a
//
//	  IF (i >= 32) { BREAK }   // failsafe loop breaker
//	}
//
//	return a
//  }
//}

DEFINE_FUNCTION DOUBLE getVariance(DOUBLE arr[], DOUBLE mean)
/* function: getVariance
 * purpose: calculates the variance in an array of values
 * parameters:
 *     arr   - the array of values
 *     mean  - mean of the values.
 */
{
 STACK_VAR DOUBLE sumSquaredDelta;
 STACK_VAR DOUBLE delta;
 STACK_VAR LONG i;
 STACK_VAR LONG len;

 len = LENGTH_ARRAY(arr)

 // set median at 0 and return if not enough valid datapoints
 if (len <= 1)
 {
	return 0;
 }
 for (i = len; i; i--)
 {
	delta = mean - arr[i]
	sumSquaredDelta = sumSquaredDelta + (delta*delta)
 }
 return sumSquaredDelta / (len-1);
}

DEFINE_FUNCTION CHAR[32] LEVtoA (DOUBLE num,LONG flag)
/* function LEVtoA()
 * purpose: Formats a LEVEL.VALUE in 3 possible ways
 *          Works with values > 2147483647:
 *            2147483647 < lNum <= 4294967295
 * parameters:
 *      level value: UByte,SByte,UInt,SInt,ULong,SLong,Float,Double
 *      flag: 0 = always use ULtoA() logic
 *            1 = use FTOA() for values < ABS(1000000), ULtoA() for larger
 *            2 = always use FTOA()
 *            3 = always use FORMAT('%f')
 * return:
 *      level value as a string
 */
{
 SWITCH (flag)
 {
	CASE 1:                           // Use FTOA() conveniently.
	{
									// where FTOA() will not print whole numbers in 1.5f exponential
									// notation, use FTOA()
									// Note: FTOA() does not print more than 6 digits,
									// and rounds up/down instead of truncating the hidden digits.
	 // FTOA() rounds up ABS_VALUE(999999.46864 to 1000000.0) to 1E+06 on NI-x200 (found on v1.2.296)
	 // FTOA() rounds up ABS_VALUE(999999.9687499999 to 1000000.0) to 1E+06 on NXC-ME v2.20.117
	 // FTOA() rounds up ABS_VALUE(999999.9687499999 to 1000000.0) to 1E+06 on NI-700 v4.1.404

	 IF ((1000000 > num) && (-1000000 < num)) {
		RETURN FTOA(num)
	 }
									// Values > 6 whole number digits FTOA prints in exponential notation,
									// so use ULtoA() logic instead
	 ELSE IF (num <= $7FFFFFFF) {
		RETURN (ITOA(num))
	 }
	 ELSE {
		RETURN "ITOA(num/1000000000),ITOA(num%1000000000)"
	 }
	}
	CASE 2: {       // always use FTOA()
	 RETURN FTOA(num)
	}
	CASE 3: {       // always use FORMAT('%f')
	 RETURN FORMAT('%f',num)
	}
	DEFAULT: {      // always use ULtoA() logic
	 IF (Num > $7FFFFFFF) {
		RETURN "ITOA(num/1000000000),ITOA(num%1000000000)" }
	 ELSE {
		RETURN (ITOA(num))
	 }
	}
 }
}
DEFINE_FUNCTION INTEGER mixedListToArray(CHAR sList[], char rDelim[], DOUBLE rDelta, char lDelim[], 
																			   DOUBLE dReturn[], DOUBLE min, DOUBLE max, INTEGER nCleanUp)
{
/* function: mixedListToArray()
 * purpose: convert a list of values and ranges presented as a string, e.g.'3,4,5,7*11',
 *          to an array of values. (Formatting back to a string would add overhead, and likely would need to be undone later.)
 *          Created originally for parsing level numbers in '?LEVEL_STATUS-' query commands.
 * parameters:
 *     sList   - the mixed list as a string
 *     rDelim  - delimiter of the range string(s) in ml
 *     rDelta  - the difference between values in the range(s)
 *     lDelim  - delimiter between items in ml
 *     dReturn - the array of values to be returned
 *     min     - minimum value allowed in the array
 *     max     - maximum value allowed in the array
 *     clean   - true = sort and remove duplicates, i.e. the previous version's behavior
 *               false = do not sort, do not remove duplicates. For use with channel on/off/pulse/push/release functions.
 * affects:
 *     dReturn - the array of values to be returned
 * return:
 *     l       - length of dA, i.e. the number of values it contains (max 600, for now)
 */
 STACK_VAR INTEGER nMl
 STACK_VAR CHAR txt[15999]
 STACK_VAR INTEGER i
 STACK_VAR INTEGER j
 STACK_VAR CHAR sVAL[600][24]      // comma delimited array of values and ranges
 STACK_VAR CHAR sVAL_TMP[600][12]  // temp storage for output of StringTokenizer(rangeToList())
 STACK_VAR INTEGER nTMP_L
 STACK_VAR INTEGER l
 STACK_VAR DOUBLE d
																	 // tokenize sList to sVAL, update nMl
 nMl = StringTokenizer(sList,lDelim,sVAL)

 IF (nMl)
 {
	txt = ''
	FOR (i = 1; i <= nMl; i++)   			// parse thru the mixed list
	{
																		// if a range is found...
	 IF (FIND_STRING(sVAL[i],rDelim,1))
	 {
	 																	// tokenize the range to txt
		rangeToList(sVAL[i],rDelim,rDelta,lDelim,txt)
		nTMP_L = StringTokenizer(txt,lDelim,sVAL_TMP)
		FOR (j = 1; j <= nTMP_L; j++)		// add the values to the array
		{
		 d = ATOF(sVAL_TMP[j])
		 IF (d >= min && d <= max)
		 {
			l++
			IF (l > MAX_LENGTH_ARRAY(dReturn)) { BREAK } // range check
			dReturn[l] = d
		 }
		}
	 }
	 ELSE                             // add the next value to add to the return array
	 {
		d =  ATOF(sVAL[i])
		IF (d >= min && d <= max)
		{
		 l++
		 IF (l > MAX_LENGTH_ARRAY(dReturn)) { BREAK } // range check
		 dReturn[l] = d
		}
	 }
	}
	SET_LENGTH_ARRAY(dReturn,l)
	IF (l)
	{
	 IF (l > MAX_LENGTH_ARRAY(dReturn))
	 {
		l = MAX_LENGTH_ARRAY(dReturn)
		SEND_STRING 0,"'statistics.axi ',ITOA(__LINE__),' mixedListToArray() return array overflow'"
	 }
	 IF (nCleanUp)
	 {
		quickSort(dReturn,1,l)
		l = removeDuplicates(dReturn,1,l)
	 }
	}
 }
 RETURN l
}

DEFINE_FUNCTION quickSort(DOUBLE arr[], LONG left, LONG right)
/* function: quickSort()
 * purpose: quicksorts an array of values from low to high
 * parameters:
 *     arr   - the array to be sorted
 *     left  - index of low side. Start with 1 to sort entire array.
 *     right - index of high side. Start with length_array(arr) to sort entire array.
 */
{
 STACK_VAR LONG i;
 STACK_VAR LONG j;
 STACK_VAR DOUBLE tmp;
 STACK_VAR DOUBLE pivot;

 i = left
 j = right

 pivot = arr[(left + right) / 2]

 /* partition */
 while (i <= j)
 {
	while (arr[i] < pivot) { i++ }
	while (arr[j] > pivot) { j-- }

	if (i <= j)
	{
	 tmp = arr[i]
	 arr[i] = arr[j]
	 arr[j] = tmp
	 i++
	 j--
	}
 }

  /* recursion */
 if (left < j)  { quickSort(arr, left, j) }
 if (i < right) { quickSort(arr, i, right) }
}

DEFINE_FUNCTION INTEGER rangeToList(CHAR r[], CHAR rDelim[], DOUBLE delta, CHAR lDelim[], CHAR sList[])
{
/* function: rangeToList()
 * purpose: convert a range of values presented as a string, e.g. '3 to 5', to a list of values presented as a string: '3,4,5'
 * parameters:
 *     r       - the range string
 *     rDelim  - delimiter of the range string
 *     delta   - the difference between values in the list
 *     lDelim  - delimiter of the list string
 *     sList   - the list string. Values are sorted low to high if delta is positive, high to low if delta is negative
 * affects:
 *     sList   - the list string. Values are sorted low to high if delta is positive, high to low if delta is negative
 * return:
 *     nValues - the number of delimited values in the list string.
 *
 */
 STACK_VAR DOUBLE min
 STACK_VAR DOUBLE max
 STACK_VAR DOUBLE d
 STACK_VAR INTEGER nValues          // count of the values in sList
 STACK_VAR CHAR sTmp[32]
 IF (FIND_STRING(r,rDelim,1))       // if range is delimited
 {
	min = ATOF(LEFT_STRING(r,FIND_STRING(r,rDelim,1)-1))
	max = ATOF(RIGHT_STRING(r,LENGTH_ARRAY(r)-FIND_STRING(r,rDelim,1)))
	IF (max < min)                    // if range is inverted, flip it
	{
	 d = max
	 max = min
	 min = d
	}
	sList = "FTOA(min),lDelim"        // enter initial value in list
	nValues++
	IF (delta > 0)
	{
	 FOR (d = min; d < max; )
	 {
		d = d + delta
		IF ((d >= min) || (d <= max))
		{
		 sTmp = "FTOA(d),lDelim"
		 IF ((LENGTH_ARRAY(sList) + LENGTH_ARRAY(sTmp)) <= MAX_LENGTH_ARRAY(sList)) {
			sList = "sList,sTmp"
			nValues++
		 }
		 ELSE {
			SEND_STRING 0,"'statistics.axi ',ITOA(__LINE__),' rangeToList() return array overflow'"
			BREAK
		 }
		} 
	 }
	 SET_LENGTH_ARRAY(sList,LENGTH_ARRAY(sList)-LENGTH_ARRAY(lDelim))
	 RETURN nValues
	}
	ELSE IF (delta < 0)
	{
	 FOR (d = max; d < min; )
	 {
		d = d + delta
		IF ((d >= min) || (d <= max))
		{
		 sTmp = "FTOA(d),lDelim"
		 IF ((LENGTH_ARRAY(sList) + LENGTH_ARRAY(sTmp)) <= MAX_LENGTH_ARRAY(sList))
		 {
			sList = "sList,sTmp"
			nValues++
		 }
		 ELSE
		 {
			SEND_STRING 0,"'statistics.axi ',ITOA(__LINE__),' rangeToList() return array overflow'"
			BREAK
		 }
		}
	 }
	 SET_LENGTH_ARRAY(sList,LENGTH_ARRAY(sList)-LENGTH_ARRAY(lDelim))
	 RETURN nValues
	}
	ELSE                              // if delta == 0
	{
	 sList = FTOA(ATOF(r))
	 RETURN 1
	}
 }
 ELSE                               // no delimiter for range
 {
  sList = FTOA(ATOF(r))
	RETURN 1
 }
}

DEFINE_FUNCTION INTEGER removeDuplicates (DOUBLE arr[], LONG left, LONG right)
{
/* function: removeDuplicates()
 * purpose: remove duplicate values from an array
 * parameters:
 *     arr   - the array to process
 *     left  - index of low side. Start with 1 to process entire array.
 *     right - index of high side. Start with length_array(arr) to process entire array.
 * affects:
 *     arr   - the array to process
 * return:
 *     LENGTH_ARRAY(arr)
 */
 STACK_VAR LONG i;
 STACK_VAR LONG j;
 STACK_VAR LONG k;
 STACK_VAR LONG r;
 STACK_VAR INTEGER len;

 i = left
 r = right
 len = LENGTH_ARRAY(arr)

 FOR (i = left; i <= r; i++)
 {
	FOR (j = i + 1; j <= r;)
	{
	 IF (arr[j] == arr[i])
	 {
		FOR (k = j; k < len; k++)
		{
		 arr[k] = arr[k+1]
		}
		r--
		len--
		SET_LENGTH_ARRAY(arr,len)
	 }
	 ELSE {	j++ }
	}
 }
 RETURN len
}

DEFINE_FUNCTION slong round(double d)
{
/* function: round()
 * purpose: round float value to the nearest whole number.
 * parameters:
 *     d     - the volue to process
 * return:
 *     r     - rounded value, cast to slong
 */
//(int)(num < 0) ? (num - 0.5) : (num + 0.5)
 STACK_VAR DOUBLE r
 r = d                              // don't change passed-in value
 IF (r >= 0) { r = r + 0.5 }        // bump up positive values so truncation rounds to whole number closest to d
 ELSE IF (r < 0) { r = r - 0.5 }    // bump down negative values so truncation rounds to whole number closest to d
 RETURN TYPE_CAST(r)                // cast to slong return truncates to whole number
}


DEFINE_FUNCTION CHAR[16] ULtoA (LONG num)
/* function ULtoA()
 * purpose: Converts an unsigned LONG number to its ASCII representation.
 *          Works with values > 2147483647:
 *            2147483647 < lNum <= 4294967295
 * parameters:
 *      32 bit value
 * return:
 *      32 bit value as a decimal string
 */
{
 IF (Num > 2147483647)              // if value is greater than SLONG
 {
	RETURN "ITOA(num/1000000000),ITOA(num%1000000000)"
 }
 ELSE
 {
	RETURN (ITOA(num))
 }
}

(***********************************************************)
(*                 STARTUP CODE GOES BELOW                 *)
(***********************************************************)
DEFINE_START

{ // local scope for STACK_VAR
  // Declare
  STACK_VAR DEV_INFO_STRUCT uDevInfo
  STACK_VAR CHAR    cVer[50]
  STACK_VAR CHAR    cTemp[50]
  STACK_VAR INTEGER nMajor
  STACK_VAR INTEGER nMinor
  STACK_VAR INTEGER nBuild

  // Check master version
  DEVICE_INFO(0:0:0,uDevInfo)
  cVer = uDevInfo.VERSION
  cTemp = REMOVE_STRING(cVer,'.',1)
  nMajor = ATOI(cTemp)
  cTemp = REMOVE_STRING(cVer,'.',1)
  nMinor = ATOI(cTemp)
  nBuild = ATOI(cVer)

  IF (uDevInfo.DEVICE_ID >= $18C)   // NI-x200 master or newer
	{
	 __juggernaut_bits__ = 5000+nMajor
	}
	ELSE IF (nMajor >= 4)             // Older master v4.xx.xx
  {
	 __juggernaut_bits__ = nMajor
  }
  ELSE                              // Older master v3.xx.xx, v2.xx.xx
  {
	 __juggernaut_bits__ = FALSE
  }
} // END local scope for STACK_VAR


(***********************************************************)
(*                THE EVENTS GOES BELOW                    *)
(***********************************************************)
DEFINE_EVENT

DATA_EVENT[0:1:0]
{
 ONLINE:
 {
	IF (__display_lib_version__ == TRUE)
	{
	 wait 50                          // WAIT FOR DEVICE HOLDOFF
	 {
		  // WHAT VERSION?
		 SEND_STRING 0:0:0,"'Using lib ',__STATISTICS_LIB_NAME__,
												', v',__STATISTICS_LIB_VERSION__";
	 }
	}
 }
}

#END_IF
(***********************************************************)
(*                     END OF PROGRAM                      *)
(*        DO NOT PUT ANY CODE BELOW THIS COMMENT           *)
(***********************************************************)
