PROGRAM_NAME='strings'
(***********************************************************)
(*  FILE CREATED ON: 07/27/2005  AT: 13:43:23              *)
(***********************************************************)
(*  FILE_LAST_MODIFIED_ON: 05/12/2017  AT: 12:15:32        *)
(***********************************************************)
/*
 * tab-width: 4
 * indent-level: 2
 *
 */
DEFINE_PROGRAM                      // rev notes
(***********************************************************)
(*  FILE REVISION: 2.15                                    *)
(*  REVISION DATE: 05/12/2017  AT: 12:15:32                *)
(*                                                         *)
(*  COMMENTS:                                              *)
(*  - nDebug values 3 & 4 now required for debug output    *)
(*    from StringBuffer(), instead of 1 & 3.               *)
(*  - Fixed problem with functions that would return a     *)
(*    false value when ndxMax was too small.               *)
(*                                                         *)
(***********************************************************)
(*  FILE REVISION: 2.14/2.13                               *)
(*  REVISION DATE: 07/16/2016  AT: 12:02:52                *)
(*                                                         *)
(*  COMMENTS:                                              *)
(*  - __STRINGS_LIB_NAME__ now lines up with other libs.   *)
(*  - Added range checking to stringTokenizerEsc().        *)
(*  - Added range checking to stringTokenizer().           *)
(*  - All function names start with 's', not 'S'.          *)
(*  - Added DEFINE_PROGRAM to top of Rev Notes to enable   *)
(*    code folding the notes without hiding the timestamp. *)
(*                                                         *)
(***********************************************************)
(*  FILE REVISION: 2.12                                    *)
(*  REVISION DATE: 03/20/2014  AT: 09:52:32                *)
(*                                                         *)
(*  COMMENTS:                                              *)
(*  - Added functions stringGetXmlElement(),               *)
(*    stringCount(), __stringCountXmlEmptyElement__(),     *)
(*    and stringTrimWhiteSpace().                          *)
(*                                                         *)
(***********************************************************)
(*  FILE REVISION: 2.11                                    *)
(*  REVISION DATE: 08/26/2013  AT: 17:56:58                *)
(*                                                         *)
(*  COMMENTS:                                              *)
(*  - Fixed bug in stringTokenizer() where old data was    *)
(*    kept if new data started with empty strings.         *)
(*                                                         *)
(***********************************************************)
(*  FILE REVISION: 2.10                                    *)
(*  REVISION DATE: 03/31/2013  AT: 22:19:20                *)
(*                                                         *)
(*  COMMENTS:                                              *)
(*  - Added function stringGetLast().                      *)
(*  - stringTokenizer() args parameter can now start with  *)
(*    deliminators, i.e. empty tokens.                     *)
(*                                                         *)
(***********************************************************)
(*  FILE REVISION: 2.09                                    *)
(*  REVISION DATE: 11/05/2012  AT: 13:24:11                *)
(*                                                         *)
(*  COMMENTS:                                              *)
(*  - added stringAppend() subroutine.                     *)
(*  - added stringBuffer() subroutine.                     *)
(*  - Changed name back to 'strings.axi' from              *)
(*    'tokenizer.axi'.                                     *)
(*                                                         *)
(***********************************************************)
(*  FILE REVISION: 2.08                                    *)
(*  REVISION DATE: 10/31/2008  AT: 18:55:08                *)
(*                                                         *)
(*  COMMENTS:                                              *)
(*  - Speed tweaks                                         *)
(*    * Reduce FOR and WHILE loop control expressions to   *)
(*      only the variable name where possible.             *)
(*    * Avoid use of LENGTH_ARRAY or LENGTH_STRING() in    *)
(*      FOR or WHILE loop control expressions.             *)
(*    For example:                                         *)
(*    Bad  -   FOR (i = 1; i <= LENGTH_STRING(str); i++)   *)
(*    Good -   len = LENGTH_STRING(str)                    *)
(*             FOR (i = 1; i <= len; i++)                  *)
(*    Better - FOR (i = LENGTH_STRING(str); i > 0; i--)    *)
(*    Best -   FOR (i = LENGTH_STRING(str); i; i--)        *)
(*                                                         *)
(***********************************************************)
(*  FILE REVISION: 2.07                                    *)
(*  REVISION DATE: 10/06/2008  AT: 07:26:34                *)
(*                                                         *)
(*  COMMENTS:                                              *)
(*  - Added __display_lib_version__ variable.              *)
(*                                                         *)
(***********************************************************)
(*  FILE REVISION: Rev 2.06                                *)
(*  REVISION DATE: 08/26/2008  AT: 08:31:14                *)
(*                                                         *)
(*  COMMENTS:                                              *)
(*  - Removed deprecated function slstringTokenizer().     *)
(*  - string-to-be-processed parameters of subroutines     *)
(*    were set to 15999 to workaround a feature that       *)
(*    truncates string expressions to 2048 bytes if passed *)
(*    in via an unbounded array.                           *)
(*    With the parameter size set to 15999 they are not    *)
(*    truncated below this "normal" NetLinx interpreter    *)
(*    limit.                                               *)
(*    string arrays, for example the output of FILE_READ() *)
(*    or GET_BUFFER_STRING(), will still pass through      *)
(*    unaffected at 16000 bytes or greater.                *)
(*                                                         *)
(***********************************************************)
(*  FILE REVISION: Rev 2.05                                *)
(*  REVISION DATE: 08/25/2008  AT: 12:27:42                *)
(*                                                         *)
(*  COMMENTS:                                              *)
(*  - Changed name from 'strings.axi' to 'tokenizer.axi'.  *)
(*                                                         *)
(***********************************************************)
(*  FILE REVISION: Rev 2.04                                *)
(*  REVISION DATE: 04/10/2007  AT: 18:24:12                *)
(*                                                         *)
(*  COMMENTS:                                              *)
(*  - added stringTokenizerEsc(). Data between two escape  *)
(*    characters will not be parsed for delim. Only        *)
(*    supports single-byte delims.                         *)
(***********************************************************)
(*  FILE REVISION: 2.03                                    *)
(*  REVISION DATE: 03/17/2007  AT: 14:47:00                *)
(*                                                         *)
(*  COMMENTS:                                              *)
(*  - stringTokenizer returns the number tokens found as   *)
(*    integer.                                             *)
(***********************************************************)
(*  FILE REVISION: Rev 2.02                                *)
(*  REVISION DATE: 03/12/2007  AT: 13:31:25                *)
(*                                                         *)
(*  COMMENTS:                                              *)
(*  - stringTokenizer now returns results to a passed-in   *)
(*    array of strings. This will allow more flexibility   *)
(*    as the caller can then convert the data by context.  *)
(*  - To simplify parameters, the function now expects any *)
(*    message header to be stripped off before calling the *)
(*    function.                                            *)
(*  - Old stringTokenizer function has been renamed to     *)
(*    slstringTokenizer.                                   *)
(***********************************************************)
(*  FILE REVISION: Rev 2.01                                *)
(*  REVISION DATE: 03/09/2007  AT: 17:12:30                *)
(*                                                         *)
(*  COMMENTS:                                              *)
(*  - Improvements to stringTokenizer:                     *)
(*    * No longer requires a delimiter between msgHeader   *)
(*      and arg1.                                          *)
(*    * delim may be more than 1 character long.           *)
(*    * returns slong, not integer.                        *)
(***********************************************************)
(*!!FILE REVISION: 2.00 GLM                                *)
(*  REVISION DATE: 07/27/2005                              *)
(*                                                         *)
(*  COMMENTS:                                              *)
(* string functions and calls                              *)
/* 1  function stringAppend (sBB[], DATA[], max_len_bb)                       */
/* 2  function stringBuffer (dvD, sBB[], DATA[], max_len_bb, nDebug)          */
/* 3  function stringCount (buffer[], data[], ndx, ndxMax)                    */
/* 4  function __stringCountXMLEmptyElement__ (buffer[], data[], ndx, ndxMax) */
/* 5  function stringGetLast (buffer[], data[], ndx, ndxMax)                  */
/* 6  function stringGetXmlElement (xml[], name[], start, element[], trim)    */
/* 7  function stringTokenizer (args[], delim[], results[][])                 */
/* 8  function stringTokenizerEsc (args[], delim, escape, results[][])        */
/* 9  function stringTrimWhiteSpace (data[])                                  */
(***********************************************************)
#IF_NOT_DEFINED __strings_lib__
#DEFINE __strings_lib__
(***********************************************************)
(*               CONSTANT DEFINITIONS GO BELOW             *)
(***********************************************************)
DEFINE_CONSTANT

// VERSION FOR CODE

CHAR __STRINGS_LIB_NAME__[20]     = '         strings.axi'
CHAR __STRINGS_LIB_VERSION__[]    = '2.15'

(***********************************************************)
(*               VARIABLE DEFINITIONS GO BELOW             *)
(***********************************************************)
DEFINE_VARIABLE

#IF_NOT_DEFINED __display_lib_version__
VOLATILE __display_lib_version__ = TRUE
#END_IF

#IF_NOT_DEFINED __MAX_LEN_STR__
VOLATILE __MAX_LEN_STR__ = 15999    // NetLinx string Interpreter limit
#END_IF

(***********************************************************)
(*           SUBROUTINE DEFINITIONS GO BELOW               *)
(***********************************************************)
(* EXAMPLE: DEFINE_FUNCTION <RETURN_TYPE> <NAME> (<PARAMETERS>) *)
(* EXAMPLE: DEFINE_CALL '<NAME>' (<PARAMETERS>) *)

DEFINE_FUNCTION INTEGER stringAppend(char sBB[], char DATA[], LONG max_len_bb)
/********************************/
/* DEFINE_FUNCTION stringAppend */
/********************************/
/*
 * purpose: Appends a string to a string variable
 *          Works around the 15999 byte string expression limit via VARIABLE_TO_STRING()
 * parameters:
 *          sBB        - the big buffer for data
 *          DATA       - new data
 *          max_len_bb - maximum desired length of sBB (may be smaller than MAX_LENGTH_ARRAY(sBB))
 *                       VARIABLE_TO_STRING() error -2 on v2.20.117 when max_len_bb > MAX_LENGTH_ARRAY(sBB)-1
 * return:  True if successful, false otherwise
 * notes:   Works around a limitation in the maximum return size of two string
 *          literals being appended. The 65535 byte chunks come from a
 *          documented 64kB limitation of the VARIABLE_TO_STRING() stream
 *          marshaller.
 *          Code based on
 *          http://www.amxforums.com/showthread.php?2085-Netlinx-Ajax-interface-module&p=63432
 */
{
  STACK_VAR LONG len_bb
  STACK_VAR LONG len_new
  STACK_VAR LONG len_pad
  STACK_VAR CHAR sPieces[3]
  STACK_VAR LONG len_dat            // Note: DATA.TEXT and CUSTOM.TEXT are limited to 2048 bytes
  STACK_VAR LONG len_VtoS
  STACK_VAR LONG lPos

  len_bb = LENGTH_ARRAY(sBB)
  len_dat = LENGTH_ARRAY(DATA)
  len_new = len_bb + len_dat

  IF((len_new <= __MAX_LEN_STR__) && (len_new <= max_len_bb))
  {
    sBB = "sBB, DATA"
    RETURN TRUE
  }
  ELSE
  {
    IF(len_bb < 3) { len_pad = 3 - len_bb }

    IF(len_new +len_pad <= max_len_bb)
    {
      IF(len_bb) { sPieces = RIGHT_STRING(sBB, 3-len_pad) }
      len_VtoS = len_dat
      WHILE(len_VtoS > 65535)
      {
        len_VtoS = len_VtoS - 65532
        lPos = len_bb + len_pad + len_VtoS - 2
        VARIABLE_TO_STRING(MID_STRING(DATA, len_VtoS + 1, 65532), sBB, lPos)
      }
      lPos = len_bb + len_pad - 2
      VARIABLE_TO_STRING(LEFT_STRING(DATA, len_VtoS), sBB, lPos)
      IF(len_pad)
      {
        FOR(lPos = 1; lPos <= 3 - len_pad; lPos++) {
          sBB[lPos + len_pad] = sPieces[lPos]
        }
        SET_LENGTH_ARRAY(sBB, len_new + len_pad)
        sBB = RIGHT_STRING(sBB, len_new)
      }
      ELSE
      {
        sBB[len_bb - 2]  = sPieces[1]
        sBB[len_bb - 1]  = sPieces[2]
        sBB[len_bb]      = sPieces[3]
        set_length_array(sBB, len_new)
      }
      RETURN TRUE
    }
  }
  RETURN FALSE
}
DEFINE_FUNCTION INTEGER stringBuffer (DEV dvD, char sBB[], char DATA[], LONG max_len_bb, INTEGER nDebug)
/********************************/
/* DEFINE_FUNCTION stringBuffer */
/********************************/
/*
 * purpose: Buffers data, working around the 15999 byte string expression limit
 *          If new data is too large to fit into the buffer, old data is pushed out to make room
 * parameters:
 *          dvD        - debug device
 *          sBB        - the big buffer for data
 *          DATA       - new data
 *          max_len_bb - maximum desired length of sBB (may be smaller than MAX_LENGTH_ARRAY(sBB))
 *          nDebug     - debug level
 * return:
 *          TRUE if all data can fit in buffer without pushing out old data
 *          FALSE if new data pushes out any old data
 */
{
  STACK_VAR LONG len_new
  STACK_VAR LONG len_dat            // Note: DATA.TEXT and CUSTOM.TEXT are limited to 2048 bytes
  STACK_VAR LONG len_bb
  STACK_VAR INTEGER return_value

  len_bb = LENGTH_ARRAY(sBB)
  len_dat = LENGTH_ARRAY(DATA)
  len_new = len_bb + len_dat

  SELECT
  {
									// if new length if larger than the NetLinx string expression limit,
									// but will still fit into the sBB
	ACTIVE ((max_len_bb > __MAX_LEN_STR__) && (len_new > __MAX_LEN_STR__) && (len_new <= max_len_bb)):
	{
	  IF (len_bb < __MAX_LEN_STR__)
	  {
		sBB = "sBB,GET_BUFFER_STRING(DATA,__MAX_LEN_STR__-len_bb)"
	  }
	  ELSE
	  {
		IF (nDebug >= 3) {
		  SEND_STRING dvD,"__STRINGS_LIB_NAME__,' ',ITOA(__LINE__),' Adding ',ITOA(len_dat),' bytes, buffer size ',ITOA(len_bb),'...'"
		}
	  }

	  IF (LENGTH_ARRAY(DATA))
	  {
		stringAppend(sBB,DATA,max_len_bb)
	  }
   	  IF (len_bb >= __MAX_LEN_STR__)
	  {
		IF (nDebug >= 3) { SEND_STRING dvD,"__STRINGS_LIB_NAME__,' ',ITOA(__LINE__),' ...done'" }
	  }
	  RETURN TRUE
	}
									// if new length if larger than the NetLinx string expression limit,
									// and larger than the sBB can hold...
									// new data pushes old data out
	ACTIVE ((max_len_bb > __MAX_LEN_STR__) && (len_new > __MAX_LEN_STR__) && (len_new > max_len_bb)):
	{
	  STACK_VAR LONG len_over

	  len_over = len_new - max_len_bb
	  IF (nDebug >= 3) {
		SEND_STRING dvD,"__STRINGS_LIB_NAME__,' Adding ',ITOA(len_dat),' bytes with overrun of ',ITOA(len_over),' bytes, buffer size ',ITOA(max_len_bb),'...'"
	  }

									// As long as we don't use "", these string functions won't truncate
	  IF (len_bb < max_len_bb) {
		IF (nDebug >= 4) {
		  SEND_STRING dvD,"__STRINGS_LIB_NAME__,' ',ITOA(__LINE__),' RIGHT_STRING(',ITOA(len_bb),',',ITOA(len_bb-len_over),')'"
		}
		sBB = RIGHT_STRING(sBB,len_bb-len_over)
	  }
	  ELSE {
		IF (nDebug >= 4) {
		  SEND_STRING dvD,"__STRINGS_LIB_NAME__,' ',ITOA(__LINE__),' RIGHT_STRING(',ITOA(len_bb),',',ITOA(max_len_bb-len_dat),')'"
		}
		sBB = RIGHT_STRING(sBB,max_len_bb-len_dat)
	  }
	  stringAppend(sBB,DATA,max_len_bb)

	  IF (nDebug >= 3) { SEND_STRING dvD,"__STRINGS_LIB_NAME__,' ',ITOA(__LINE__),' ...done'" }
	  RETURN FALSE
	}
								/*****************************************************/
								/* new length <= the NetLinx string expression limit */
								/*****************************************************/
								// new data pushes old data out
	ACTIVE (LENGTH_ARRAY(DATA) > max_len_bb):
	{
	  sBB = RIGHT_STRING(DATA,max_len_bb)
	  RETURN FALSE
	}
								  // new data pushes old data out
	ACTIVE (len_new > max_len_bb):
	{
	  STACK_VAR LONG len_over
								  // normal operation
	  IF (len_bb <= max_len_bb)
	  {
		len_over = len_new - max_len_bb
		sBB = "RIGHT_STRING(sBB,len_bb-len_over),DATA"
	  }
	  ELSE                        // else if we've decreased the size of max_len_bb to max_len_bb,
	  {                           // but not yet received new data to decrease LENGTH_ARRAY(sBB)
		len_over = len_new - len_bb
		sBB = "RIGHT_STRING(sBB,__MAX_LEN_STR__-len_over),DATA"
	  }
	  RETURN FALSE
	}
	ACTIVE (TRUE):                // if there is enough room for all data
	{
	  sBB = "sBB,DATA"
	  RETURN TRUE
	}
  }
}
DEFINE_FUNCTION LONG stringCount (CHAR buffer[],CHAR data[],LONG ndx,LONG ndxMax)
{
/*******************************/
/* DEFINE_FUNCTION stringCount */
/*******************************/
/*
 * purpose:   return the number of times <data> is found in buffer, between ndx and ndxMax
 *            This function is used by stringGetXmlElement(), and may have other uses
 * parameters:
 * 	          buffer - buffer to search
 *	          data   - data to search for
 *            ndx    - where to start search
 *            ndxMax - where search should stop
 * return:    number of times data was found
 *            if data is found in buffer, ndx is updated to the location where it was found
 */
  STACK_VAR LONG w                  // where data is found in buffer
  STACK_VAR LONG r                  // value to return
  STACK_VAR LONG lenDat             // length of data
  STACK_VAR LONG lenMax             // if lenDat > 1, make sure data end doesn't go over ndxMax

  w = FIND_STRING(buffer,data,ndx)
  IF (!w) { RETURN 0 }
  lenDat = LENGTH_ARRAY(data)
  IF (ndxMax <= lenDat + 1) { RETURN 0 }
  lenMax = ndxMax - lenDat + 1

  IF (w > lenMax)                   // if ndx is not valid
  {
	RETURN 0
  }
  ELSE IF (w)                       // forward search for another <data>, jumping ahead via FIND_STRING()
  {
	STACK_VAR LONG i

	i = w
	r++
	WHILE (w)
	{
	  w = i                         // this is needed for 2nd, 3rd, nth passes
	  i = FIND_STRING(buffer,data,i + lenDat)
	  IF (i = 0) { BREAK }
	  IF (i > lenMax) {	BREAK }
	  if (i > w) { r++ }
	}
	ndx = w
	RETURN r
  }
  IF (w) { ndx = w }
  RETURN r
}
DEFINE_FUNCTION LONG __stringCountXMLEmptyElement__ (CHAR XML[],CHAR name[],LONG ndx,LONG ndxMax)
{
/**************************************************/
/* DEFINE_FUNCTION __stringCountXMLEmptyElement__ */
/**************************************************/
/*
 * purpose:   return the number of times an empty XML element is found in XML, between ndx and ndxMax
 *            This function is used by stringGetXmlElement(), probably not much use elsewhere
 * parameters:
 * 	          XML    - XML buffer to search
 *	          name   - name of the element to search for
 *            ndx    - where to start search
 *            ndxMax - where search should stop
 * return:    number of times name was found
 *            if name is found in XML, ndx is updated to the location where it was found
 */
  STACK_VAR LONG w                  // where "'<',name" is found in XML
  STACK_VAR LONG e                  // where '/>' is found in XML
  STACK_VAR LONG s_check            // where next '<' is found in XML
  STACK_VAR LONG r                  // value to return
  STACK_VAR LONG lenDat             // length of name
  STACK_VAR LONG lenMax             // if lenDat > 1, make sure name end doesn't go over ndxMax

  w = FIND_STRING(XML,"'<',name",ndx)
  IF (!w) { RETURN 0 }
  e = FIND_STRING(XML,"'/>'",w)  // look for end of empty tag
  IF (!e) { RETURN 0 }

  lenDat = LENGTH_ARRAY(name) + 3   // + 3 for '<' and '/>'
  IF (ndxMax <= lenDat + 1) { RETURN 0 }
  lenMax = ndxMax - lenDat + 1
  IF (w > lenMax)                   // if ndx is not valid
  {
	RETURN 0
  }
  ELSE                              // forward search for another name, jumping ahead via FIND_STRING()
  {
	STACK_VAR LONG i

	i = w
	s_check = FIND_STRING(XML,'<',i+1)
									// if no other start tag was found before the end empty tag, increment r
	IF (s_check && (s_check < e)) {}
	ELSE { r++ }
	WHILE (w)
	{
	  w = i                         // this is needed for 2nd, 3rd, nth passes
	  i = FIND_STRING(XML,"'<',name",i + lenDat)
	  IF (i = 0) { BREAK }
	  e = FIND_STRING(XML,"'/>'",i)
	  IF (e = 0) { BREAK }
	  IF (i > lenMax) {	BREAK }
	  if (i > w) {
		s_check = FIND_STRING(XML,'<',i+1)
									// if no other start tag was found before the end empty tag, increment r
		IF (s_check && (s_check < e)) {}
		ELSE { r++ }
	  }
	}
	ndx = w
	RETURN r
  }
  IF (w) { ndx = w }
  RETURN r
}
DEFINE_FUNCTION LONG stringGetLast (CHAR buffer[],CHAR data[],LONG ndx,LONG ndxMax)
{
/*********************************/
/* DEFINE_FUNCTION stringGetLast */
/*********************************/
/*
 * purpose:   return the index into buffer where the last <data> starts
 * parameters:
 * 	          buffer - buffer to search
 *	          data   - data to search for
 *            ndx    - where to start search
 *            ndxMax - where search should stop
 * return:    ndx where last data starts if found, 0 if not found
 */
  STACK_VAR LONG r                  // value to return
  STACK_VAR LONG lenDat             // length of data
  STACK_VAR LONG lenMax             // if lenDat > 1, make sure data end doesn't go over ndxMax

  r = FIND_STRING(buffer,data,ndx)
  IF (!r) { RETURN 0}
  lenDat = LENGTH_ARRAY(data)
  IF (ndxMax <= lenDat + 1) { RETURN 0 }
  lenMax = ndxMax - lenDat + 1

  IF (r > lenMax)                   // if ndx is not valid
  {
	RETURN 0
  }
  ELSE IF (r)                       // forward search for another <data>, jumping ahead via FIND_STRING()
  {
	STACK_VAR LONG i

	i = r
	WHILE (i)
	{
	  r = i                         // this is needed for 2nd, 3rd, nth passes
	  i = FIND_STRING(buffer,data,i + lenDat)
	  if (i > lenMax) { BREAK }
	}
	RETURN r
  }
  RETURN r
}
DEFINE_FUNCTION LONG stringGetXmlElement(CHAR XML[],CHAR NAME[],LONG START,CHAR ELEMENT[],LONG TRIM)
/*
 * FUNCTION: stringGetXmlElement()
 * This subroutine can be run in a loop, searching for multiple XML elements with the same tag name
 * PARAMETERS:
 *   XML     - the data to be parsed
 *   NAME    - tag name of the element(s) to look for
 *   START   - start position in XML to look for NAME
 *   ELEMENT - array to hold returned XML data
 *   TRIM    - TRUE = return ELEMENT without the start and end tags
 * RETURN    - next starting position to look for i, past the element found
 */
{
  // This subroutine ignores child tags named the same as the parent...
  // If you want to parse for those, use TRIM to return the ELEMENT without the parent tag, then reparse.
  // This() also assumes you know the context of the tag you are looking for, i.e. parent tag
  // ALSO, THIS DOES NOT SUPPORT CDATA ESCAPING OR CHARACTER ENTITY REFERENCES

  STACK_VAR LONG i
  STACK_VAR Long failsafe
  STACK_VAR LONG S                  // Start of tag found
  STACK_VAR LONG ES                 // End of Start tag
  STACK_VAR LONG BE                 // Beginning of End tag
  STACK_VAR LONG E                  // End of end tag
  STACK_VAR LONG N                  // Number of characters in tag
  STACK_VAR LONG Sc                 // Start of child tag found
  STACK_VAR LONG ESc                // End of child tag found
  STACK_VAR LONG BEc                // Beginning of child End tag
  STACK_VAR LONG Ec                 // End of child end tag
  STACK_VAR LONG dName              // Depth of nested tags with <name>

  IF (START == 0) { i = 1 }         // NetLinx indexes start at 1, not 0
  ELSE            { i = START }

  IF (FIND_STRING(XML,"'<',NAME",i))
  {
									// FIND START OF TAG WITHOUT ATTRIBUTES
	S=FIND_STRING(XML,"'<',NAME,'>'",i)
	IF(!S) {                        // FIND START OF TAG WITH ATTRIBUTES, or EMPTY ELEMENT without ATTRIBUTES
	  S=FIND_STRING(XML,"'<',NAME",i)
	}
	IF(S) {
	  ES=FIND_STRING(XML,'>',S)     // END OF START TAG FOUND
	  BE=FIND_STRING(XML,'/>',S)
	  IF (BE = ES-1)                // If this is an empty element
	  {
		E = BE+2                    // SET END OF END TAG (actually end + 1 so we don't have to add 1 later)
		ES = S+LENGTH_ARRAY(NAME)
		N = E-S                     // NUMBER OF CHARACTERS TO TAKE (INCLUSIVE OF ENDPOINTS)
		ELEMENT = MID_STRING(XML,S,N)
	  }
	  ELSE
	  {
		dName++
		i = S
		ESc = ES
		WHILE (dName) {             // look for child tags with same name
		  failsafe++
								    // FIND BEGINNING OF END TAG
		  BEc=FIND_STRING(XML,"'</',NAME",i)
		  IF(BEc) {
			Ec = BEc+LENGTH_STRING(NAME)+3// SET END OF END TAG (actually end + 1 so we don't have to add 1 later)
		  }
		  ELSE {                    // IT'S AN EMPTY ELEMENT, POSSIBLY WITH ATTRIBUTES...
			BEc = FIND_STRING(XML,'/>',i)// LOOK FOR END OF ELEMENT AFTER START
			IF (BEc) {
			  Ec = BEc+2            // SET END OF END TAG (actually end + 1 so we don't have to add 1 later)
			  ESc = Sc+LENGTH_ARRAY(NAME)// no use setting ES if BE wasn't found
			}
		  }
		  IF (Ec) {                 // complete end tag found, make ELEMENT candidate
			N = Ec-S                // NUMBER OF CHARACTERS TO TAKE (EXCLUSIVE OF ENDPOINTS)
			ELEMENT = MID_STRING(XML,S,N)
									// if ELEMENT candidate has an equal number of start and end tags of NAME, we're done
			dName = stringCount(ELEMENT,"'<',NAME",1,LENGTH_ARRAY(ELEMENT))                // + # start tags
					- stringCount(ELEMENT,"'</',NAME,'>'",1,LENGTH_ARRAY(ELEMENT))         // - # "normal" end tag
					- __stringCountXMLEmptyElement__(ELEMENT,NAME,1,LENGTH_ARRAY(ELEMENT)) // - # empty element end tags

			Sc = stringGetLast(ELEMENT,"'<',NAME",ESc,BEc)
			i = Ec
			E = Ec
			BE = BEc
		  }
		  IF (failsafe > 32) { BREAK }
		}
	  }

	  IF(E)                         // complete end tag found
	  {
		IF (TRIM) {
		  N = BE-ES-1               // NUMBER OF CHARACTERS TO TAKE (EXCLUSIVE OF ENDPOINTS)
		  ELEMENT = MID_STRING(XML,ES+1,N)
		}
//		ELSE {                      // this was already done, in the WHILE() or the empty element section
//		  N = E-S                   // NUMBER OF CHARACTERS TO TAKE (INCLUSIVE OF ENDPOINTS)
//		  ELEMENT = MID_STRING(XML,S,N)
//		}
		i = E + 1
		RETURN i
	  }
	  ELSE {                        // start not found
		ELEMENT = ""
		i = 0
		RETURN i
	  }
	}
  }
  ELSE                              // start not found
  {
	ELEMENT = ""
	i = 0
	RETURN i
  }
}
DEFINE_FUNCTION INTEGER stringTokenizer(CHAR args[15999], CHAR delim[], CHAR sArgArray[][])
/***********************************/
/* DEFINE_FUNCTION stringTokenizer */
/***********************************/
/*
 * this function requires the data be of
 * the form "<arg1><delim>...<delim><argX>"
 *
 * Note v2.10 - if the data starts with one or more <delim>, args starts with an empty value(s).
 *              i.e. LENGTH_ARRAY(sArgArray) will be 1, LENGTH_ARRAY(sArgArray[1]) will be 0.
 *              previous versions would fail if args started with <delim>
 *
 * purpose:
 *  non-destructively parse delimited string values from a larger character string.
 *
 * parameters:
 *  args[]        - raw string with delimited data.
 *  delim[]       - delimiter between data.
 *  sArgArray[][] - results are stored in this array.
 *
 * returns:
 *  number of tokens found
 *
 */
{
  STACK_VAR INTEGER nStart;
  STACK_VAR INTEGER nNbrOfValues;
									// Get the start position past the header
  nStart = 1;			            // make sure start position is non-zero

									// if args starts with delim(s), initial value(s) is empty
  WHILE (FIND_STRING(args,"delim",nStart) = nStart)
  {
	sArgArray[nStart] = ''          // be sure to clear old data
									// set nStart to beginning of data string
	nStart = nStart + LENGTH_ARRAY(delim);
	nNbrOfValues++;                 // increment value counter/index
									// range check
	IF (nNbrOfValues > MAX_LENGTH_ARRAY(sArgArray))
	{
	  SEND_STRING 0,"'strings.axi ',ITOA(__LINE__),' stringTokenizer() return array overflow'"
	  nNbrOfValues = MAX_LENGTH_ARRAY(sArgArray)
	  BREAK
	}
  }

									// if there is more than one value, there will be at least
									// one delimiter.
  IF (FIND_STRING(args,"delim",nStart) > nStart)
  {
	nNbrOfValues++;                 // increment value counter/index
	IF (nNbrOfValues <= MAX_LENGTH_ARRAY(sArgArray))
	{
	  sArgArray[nNbrOfValues] = MID_STRING(args,
										  nStart,
										  FIND_STRING(args,"DELIM",nStart) - nStart);

									// find the next delimiter, set nStart to that value
	  nStart = FIND_STRING(args,"DELIM",nStart);

									// while there's a delimiter at nStart
	  WHILE (FIND_STRING(args,"delim",nStart))
	  {
									// set nStart to beginning of data string
		nStart = nStart + LENGTH_ARRAY(delim);
		nNbrOfValues++;             // increment value counter/index
									// range check
		IF (nNbrOfValues > MAX_LENGTH_ARRAY(sArgArray))
		{
		 SEND_STRING 0,"'strings.axi ',ITOA(__LINE__),' stringTokenizer() return array overflow'"
		 nNbrOfValues = MAX_LENGTH_ARRAY(sArgArray)
		 BREAK
		}

									// if this is not the last value
		IF (FIND_STRING(args,"DELIM",nStart))
		{
									// store the value
		  sArgArray[nNbrOfValues] = MID_STRING(args,
											  nStart,
											  FIND_STRING(args,"DELIM",nStart) - nStart);

									// find the next delimiter, set nStart to that value
		  nStart = FIND_STRING(args,"DELIM",nStart);
		}
		ELSE                        // if this is the last value
		{
									// store the value
		  sArgArray[nNbrOfValues] = RIGHT_STRING(args,LENGTH_ARRAY(args) - (nStart - 1))
		}
	  }
	}
	ELSE
	{
	  SEND_STRING 0,"'strings.axi ',ITOA(__LINE__),' stringTokenizer() return array overflow'"
	  nNbrOfValues = MAX_LENGTH_ARRAY(sArgArray)
	}
  }
									// If there is only one value and no delim
  ELSE IF (FIND_STRING(args,"delim",nStart) = 0 && (LENGTH_ARRAY(args) >= nStart))
  {
	nNbrOfValues++;                 // increment value counter/index
	IF (nNbrOfValues <= MAX_LENGTH_ARRAY(sArgArray)) // range check
	{
									// store the value
	  sArgArray[nNbrOfValues] = RIGHT_STRING(args,LENGTH_ARRAY(args) - (nStart - 1))
	}
	ELSE
	{
	  SEND_STRING 0,"'strings.axi ',ITOA(__LINE__),' stringTokenizer() return array overflow'"
	  nNbrOfValues = MAX_LENGTH_ARRAY(sArgArray)
	}
  }
  SET_LENGTH_ARRAY(sArgArray,nNbrOfValues);
  RETURN nNbrOfValues;
}

DEFINE_FUNCTION INTEGER stringTokenizerEsc(CHAR args[15999], CHAR delim, CHAR escape, CHAR sArgArray[][])
/**************************************/
/* DEFINE_FUNCTION stringTokenizerEsc */
/**************************************/
/*
 * this function requires the data be of
 * the form "<arg1><delim>...<delim><argX>"
 *
 * purpose:
 *	non-destructively parse delimited string values from a larger character string.
 *
 * parameters:
 *	args[]      - raw string with delimited data.
 *	delim		- single-byte delimiter between data.
 *  escape      - single-byte escape character. Double-quote is typically used.
 *                data inside escaped area is not parsed for delim's.
 *	result[][]  - results are stored in this array.
 *
 * returns:
 *	number of tokens found
 *
 */
{
  STACK_VAR INTEGER i;              // outer loop counter
  STACK_VAR INTEGER byteCounter;
  STACK_VAR INTEGER nTokenCount;
  STACK_VAR INTEGER bEscaped;
  STACK_VAR INTEGER len;

  IF (LENGTH_ARRAY(args))
  {
	nTokenCount = 1
	byteCounter = 0
	len = LENGTH_ARRAY(args)

	FOR (i = 1; i <= len; i++)
	{
	  SELECT
	  {
		ACTIVE (bEscaped == TRUE):  // if inside an escaped area, don't search for delim's...
		{
		  IF (args[i] = escape)     // if end of escape is found...
			bEscaped = FALSE        // ...clear the escaped flag...

		  byteCounter++             // ...add data to sArgArray[nTokenCount]
		  IF (byteCounter > MAX_LENGTH_ARRAY(sArgArray[nTokenCount]))
		  {
			SEND_STRING 0,"'strings.axi ',ITOA(__LINE__),' stringTokenizerEsc() return array[',ITOA(nTokenCount),'] overflow'"
			byteCounter = MAX_LENGTH_ARRAY(sArgArray[nTokenCount])
			BREAK
		  }
		  sArgArray[nTokenCount][byteCounter] = args[i]
		}
		ACTIVE (args[i] == escape): // else if a start of escape area is found...
		{
		  bEscaped = TRUE           // ...set escaped flag...

		  byteCounter++             // ...add data to sArgArray[nTokenCount]
		  IF (byteCounter > MAX_LENGTH_ARRAY(sArgArray[nTokenCount]))
		  {
			SEND_STRING 0,"'strings.axi ',ITOA(__LINE__),' stringTokenizerEsc() return array[',ITOA(nTokenCount),'] overflow'"
			byteCounter = MAX_LENGTH_ARRAY(sArgArray[nTokenCount])
			BREAK
		  }
		  sArgArray[nTokenCount][byteCounter] = args[i]
		}
		ACTIVE (args[i] == delim):	// else if a new delimiter is found...
		{
									// ...set the byte count of token...
		  SET_LENGTH_ARRAY(sArgArray[nTokenCount],byteCounter)

		  nTokenCount++             // ...increment token count...
		  IF (nTokenCount > MAX_LENGTH_ARRAY(sArgArray))
		  {
			SEND_STRING 0,"'strings.axi ',ITOA(__LINE__),' stringTokenizerEsc() return array overflow'"
			nTokenCount = MAX_LENGTH_ARRAY(sArgArray)
			BREAK
		  }
		  byteCounter = 0           // ...and reset byteCounter
		}
		ACTIVE (1):                 // else...
		{
		  byteCounter++             // ...add data to sArgArray[nTokenCount]
		  IF (byteCounter > MAX_LENGTH_ARRAY(sArgArray[nTokenCount]))
		  {
			SEND_STRING 0,"'strings.axi ',ITOA(__LINE__),' stringTokenizerEsc() return array[',ITOA(nTokenCount),'] overflow'"
			byteCounter = MAX_LENGTH_ARRAY(sArgArray[nTokenCount])
			BREAK
		  }
		  sArgArray[nTokenCount][byteCounter] = args[i]
		}
	  }
	}
									// set the length of final token
	SET_LENGTH_ARRAY(sArgArray[nTokenCount],byteCounter)
  }

  SET_LENGTH_ARRAY(sArgArray,nTokenCount);
  RETURN nTokenCount;
}
DEFINE_FUNCTION stringTrimWhiteSpace(CHAR DATA[]){
  STACK_VAR LONG i
									// remove whitespace from start of DATA
  WHILE (LENGTH_ARRAY(DATA) && (
	(DATA[1] = $09) ||              // horizontal tab
	(DATA[1] = $0A) ||              // line feed
	(DATA[1] = $0B) ||              // vertical tab
	(DATA[1] = $0C) ||              // form feed
	(DATA[1] = $0D) ||              // carriage return
	(DATA[1] = $20) ||              // space
	(DATA[1] = $85) ||              // next line (unicode)
	(DATA[1] = $A0))) {             // nbsp 
	GET_BUFFER_CHAR(DATA)
  }
									// remove whitespace from end of DATA
  i = LENGTH_ARRAY(DATA)
  WHILE (i) {
	IF ((DATA[i] = $09) ||
		(DATA[i] = $0A) ||
		(DATA[i] = $0B) ||
		(DATA[i] = $0C) ||
		(DATA[i] = $0D) ||
		(DATA[i] = $20) ||
		(DATA[i] = $85) ||
		(DATA[i] = $A0))
		 { i-- }
	ELSE { BREAK }
  }
  SET_LENGTH_ARRAY(DATA,i)
}

(***********************************************************)
(*                THE EVENTS GOES BELOW                    *)
(***********************************************************)
DEFINE_EVENT

DATA_EVENT[0:1:0]
{
  ONLINE:
  {
	IF (__display_lib_version__ == TRUE)
	{
	  wait 50                       // WAIT FOR DEVICE HOLDOFF
	  {
		  // WHAT VERSION?
		SEND_STRING 0:0:0,"'Using lib ',__STRINGS_LIB_NAME__,
								  ', v',__STRINGS_LIB_VERSION__";
	  }
	}
  }
}

#END_IF
(***********************************************************)
(*                     END OF PROGRAM                      *)
(*        DO NOT PUT ANY CODE BELOW THIS COMMENT           *)
(***********************************************************)
