PROGRAM_NAME='HiQnet_Comm_test'
(***********************************************************)
(*  FILE CREATED ON: 12/01/2020  AT: 16:42:22              *)
(***********************************************************)
(*  FILE_LAST_MODIFIED_ON: 02/20/2021  AT: 09:25:13        *)
(***********************************************************)
(*  FILE REVISION: Rev 0.5.6                               *)
(*  REVISION DATE: 02/20/2021  AT: 09:25:13                *)
(*                                                         *)
(*  COMMENTS:                                              *)
(*  * Very basic HiQnet disco support, with TP control.    *)
(*  * 3rd Party Controller AA strings: fixed bug that      *)
(*    added "$00,$00" to outgoing string.                  *)
(*  * Separated Hello Query Payloads from the big structs. *)
(*  * Improved array tracking. No more one-offs.           *)
(*  * Improved touch panel support on disco page.          *)
(*  * Removed WIDECHAR support. Just use the raw UTF16 as  *)
(*    it comes over HiQnet.                                *)
(*                                                         *)
(***********************************************************)
(*  FILE REVISION: Rev 0.4                                 *)
(*  REVISION DATE: 01/25/2021  AT: 07:51:27                *)
(*                                                         *)
(*  COMMENTS:                                              *)
(*  * Added HiQnet disco querier device.                   *)
(*  * Added ACK response to ACK REQUEST.                   *)
(*  * Separated Disco Payloads from the big structs.       *)
(*                                                         *)
(***********************************************************)
(*  FILE REVISION: Rev 0.3                                 *)
(*  REVISION DATE: 01/19/2021  AT: 12:31:12                *)
(*                                                         *)
(*  COMMENTS:                                              *)
(*  - Added generic subscribe to all non-sensors after     *)
(*    connection...to mimic AA behavior. For now, this     *)
(*    seems to be the workaround to get information from   *)
(*    CDi Drivecore and DSi 2.0 amplifiers.                *)
(*                                                         *)
(***********************************************************)
(*  FILE REVISION: Rev 0.2                                 *)
(*  REVISION DATE: 01/11/2021  AT: 08:23:27                *)
(*                                                         *)
(*  COMMENTS:                                              *)
(*  * Able to connect and start a working Keep Alive for   *)
(*    ITHD4, DCiN, DCiHD, DCiDA.                           *)
(*  * Inconsistent Session and Get Attributes responses    *)
(*    from CDi DC, DSi2.0 amps. Keep Alives do not seem to *)
(*    be necessary for these amps, so maybe not much of an *)
(*    issue.                                               *)
(*  * Able to decode errors from e.g. ITHD4, DCiN, BLU806. *)
(*    This still needs a bit of work though...             *)
(*                                                         *)
(***********************************************************)
(* System Type : NetLinx                                   *)
(***********************************************************)
#warn 'todo 1: more work on disco...'
#warn 'todo 2: more work on attributes'
#warn 'todo 3: code cleanup'
#warn 'todo 4: add control examples'
/*
 * tab-width: 2
 * indent-level: 1
 *
 */
#INCLUDE 'statistics'
#INCLUDE 'strings'
#INCLUDE 'DEV_TO_STRING'
#INCLUDE 'debug'
(***********************************************************)
(*          DEVICE NUMBER DEFINITIONS GO BELOW             *)
(***********************************************************)
DEFINE_DEVICE

dvSelf       = 0:256:0              // NX loopback connection to retrieve MAC address
dvHQnDiscoBC = 0:257:0              // HiQnet Disco Broadcast
dvHQN_2      = 0:2:0                // These port numbers don't have to match up with the HQN numbers...
dvHQN_3      = 0:3:0                // But it may be convenient for troubleshooting, closing ports, etc.
dvHQN_4      = 0:4:0
dvHQN_5      = 0:5:0
dvHQN_6      = 0:6:0
dvHQN_7      = 0:7:0
dvHQN_8      = 0:8:0
dvHQN_9      = 0:9:0
dvHQN_10     = 0:10:0
dvHQN_11     = 0:11:0
dvHQN_12     = 0:12:0
dvHQN_13     = 0:13:0
dvHQN_14     = 0:14:0
dvHQN_15     = 0:15:0
dvHQN_16     = 0:16:0
dvHQN_17     = 0:17:0
dvHQN_18     = 0:18:0
dvHQN_19     = 0:19:0
dvHQN_20     = 0:20:0
dvHQN_21     = 0:21:0
dvHQN_22     = 0:22:0
dvHQN_23     = 0:23:0
dvHQN_24     = 0:24:0
dvHQN_25     = 0:25:0


dvTP1 = 10001:1:0
dvTP2 = 10002:1:0
dvTP3 = 10003:1:0

vdvControl = 33000:1:0

(***********************************************************)
(*               CONSTANT DEFINITIONS GO BELOW             *)
(***********************************************************)
DEFINE_CONSTANT                     // version

CHAR __HIQNET_COMM_NAME__[20]   = 'HiQnet_Comm_test.axs'
CHAR __HIQNET_COMM_VERSION__[]  = '0.5.6'

DEFINE_CONSTANT                     // HiQnet

                                    // OUI's used with the _sHiQnet_Network_Info struct
OUI_PHAST      = $00609F            // AMX products
OUI_CROWN      = $000A40            // ITHD and older Crown products
OUI_SOUNDCRAFT = $000FD4            // Current HARMAN audio products including Crown DCi, CDi Drivecore, JBL DSi 2.0

PORT_HQN = 3804

// MID == Method ID == Message ID
// Message Flags
HIQ_MID_FLAG_REQUEST_ACK    = $0001
HIQ_MID_FLAG_ACK            = $0002
HIQ_MID_FLAG_INFORMATION    = $0004
HIQ_MID_FLAG_ERROR          = $0008
HIQ_MID_FLAG_GUARANTEED     = $0020 // GUARANTEED == TCP ONLY == don't use UDP,
																		// but "...some messages, in some circumstances,
																		// the Routing Layer will...send the message
																		// out on a different service [than specified by this flag].
HIQ_MID_FLAG_MULTI_PART_MSG = $0040
HIQ_MID_FLAG_SESSION_NUMBER = $0100

HIQ_ROUTING_NID_ETHERNET    = 1     // for DISCO INFO Network ID
HIQ_ROUTING_NID_RS232       = 4     // for DISCO INFO Network ID
HIQ_ROUTING_NID_USB_P2P     = 5     // for DISCO INFO Network ID

HIQ_ROUTING_COST_ETHERNET = 1       // for DISCO INFO cost
HIQ_ROUTING_COST_USB      = 10      // for DISCO INFO cost
HIQ_ROUTING_COST_RS232    = 100     // for DISCO INFO cost

// Routing Layer Methods
HIQ_MID_ROUTING_DISCO_INFO       = $0000 // routing layer method
HIQ_MID_ROUTING_GET_NETWORK_INFO = $0002 // routing layer method
HIQ_MID_ROUTING_REQUEST_ADDRESS  = $0004 // routing layer method
HIQ_MID_ROUTING_ADDRESS_USED     = $0005 // routing layer method
HIQ_MID_ROUTING_SET_ADDRESS      = $0006 // routing layer method
HIQ_MID_ROUTING_GOODBYE          = $0007 // routing layer method
HIQ_MID_ROUTING_HELLO            = $0008 // routing layer method

// Device Methods
HIQ_MID_DEVICE_GET_ATTRIBUTES    = $010D // device method
HIQ_MID_DEVICE_GET_VD_LIST       = $011A // device method
HIQ_MID_DEVICE_STORE             = $0124 // device method
HIQ_MID_DEVICE_RECALL            = $0125 // device method
HIQ_MID_DEVICE_LOCATE            = $0129 // device method

HIQ_MID_EVENT_LOG_SUBSCRIBE      = $0115 // event log method
HIQ_MID_EVENT_LOG_UNSUBSCRIBE    = $012B // event log method
HIQ_MID_EVENT_LOG_REQUEST        = $012C // event log method

// Multi Parameter Methods
HIQ_MID_MULTISV_SET               = $0100 // multi parameter method
HIQ_MID_MULTIVDOSV_SET            = $0101 // multi parameter method, e.g. ITHD Output Level Subscribe returns this
HIQ_MID_MULTISV_SET_PERCENT       = $0102 // multi parameter method
HIQ_MID_MULTISV_GET               = $0103 // multi parameter method
HIQ_MID_MULTISV_SUBSCRIBE         = $010F // multi parameter method
HIQ_MID_MULTISV_SUBSCRIBE_PERCENT = $0111 // multi parameter method
HIQ_MID_MULTISV_UNSUBSCRIBE       = $0112 // multi parameter method
HIQ_MID_MULTISV_SUBSCRIBE_ALL     = $0113 // Wireshark
HIQ_MID_MULTISV_UNSUBSCRIBE_ALL   = $0114 // Wireshark

// Subscription Types
HIQ_SUBSCRIPTION_VALUE = $00
HIQ_SUBSCRIPTION_EVENT = $01

// Subscribe_All Types - bitmapped
HIQ_SUB_TYPE_SV_NON_SENSOR = $01
HIQ_SUB_TYPE_SV_SENSOR     = $02
HIQ_SUB_TYPE_ATTRIBUTES    = $04

HIQ_SUB_SENSOR_RATE_NX  = $07D0 // 2000ms for Netlinx(?)... HiQnet default is 50ms

// Subscribe All Flags
HIQ_SUB_FLAG_SEND_INITIAL  = $01

// Parameters

HIQ_LOCATE_OFF      = $0000
HIQ_LOCATE_ON       = $FFFF

// Data_Types
HIQ_BYTE      = 0                   // CHAR 1 byte
HIQ_UBYTE     = 1                   // CHAR 1 byte
HIQ_WORD      = 2                   // SINTEGER 2 bytes
HIQ_UWORD     = 3                   // INTEGER 2 bytes
HIQ_LONG      = 4                   // SLONG 4 bytes
HIQ_ULONG     = 5                   // LONG 4 bytes
HIQ_FLOAT32   = 6                   // FLOAT 4 bytes
HIQ_FLOAT64   = 7                   // DOUBLE 8 bytes
HIQ_BLOCK     = 8                   // STRING array 0 - 65535 bytes
HIQ_STRING    = 9                   // Raw UTF16BE array 0 - 32767 characters
HIQ_LONG64    = 10                  // 2x LONG(?) 8 bytes
HIQ_ULONG64   = 11                  // 2x LONG 8 bytes
HIQ_BIG_BLOCK = 12                  // LBYTE STRING array 0 - 4,294,967,926 bytes

// Errors aka Control Network Event IDs
HIQ_ERROR_BAD_VERSION   = $0001 // Invalid Version	The version number in the HiQnet header is unknown.
HIQ_ERROR_BAD_LENGTH    = $0002 // Invalid Length: The header length specified in the packet is wrong.
																//   There are not enough bytes in packet payload to hold message type.
HIQ_ERROR_BAD_VIRTUAL   = $0003 // Invalid Virtual Device: Tried to Create VD on an invalid VD.
																//   Set/Get/Subscribe/Attributes referenced an invalid VD.
HIQ_ERROR_BAD_OBJECT    = $0004 // Invalid Object: Set/Get/Subscription referenced an invalid object.
HIQ_ERROR_BAD_SV        = $0005 // Invalid Parameter: Set/Get/Subscribe/Attributes referenced an invalid State Variable.
HIQ_ERROR_BAD_MSG_ID    = $0006 // Invalid Message ID: Received a message with an unknown Message ID.
HIQ_ERROR_BAD_VALUE     = $0007 // Invalid Value: Tried to set an attribute with a value that is out of range.
																//   Referenced an invalid scene number, or encountered an invalid scene data length.
																//   Referenced an invalid preset number.
HIQ_ERROR_UNAVAILABLE   = $0008 // Resource Unavailable: Running out of internal software resources.
																//   Device is in a state where it cannot process the current request (flashing, configuration, etc).
																//   Cannot set node ID, subscription, SV value, or the configuration is locked.
HIQ_ERROR_MSG_NA        = $0009 // Unsupported: Received a message that is considered obsolete.
																//   Trying to flash a device that does not support flashing.
																//   Trying to add unsupported or invalid (too long) attribute types.
HIQ_ERROR_BAD_VD_CLASS  = $000A // Invalid Virtual Device Class: Referenced an invalid VD Class.
HIQ_ERROR_BAD_OBJ_CLASS = $000B // Invalid Object Class: Referenced an invalid Object Class.
HIQ_ERROR_BAD_SV_CLASS  = $000C // Invalid Parameter Class: Referenced an invalid State Variable Class.
HIQ_ERROR_BAD_ATTRIBUTE = $000D // Invalid Attribute ID: Get/Set referenced an invalid Attribute ID.
HIQ_ERROR_BAD_DATA_TYPE = $000E // Invalid DataType:Set referenced an invalid data type (class).
HIQ_ERROR_BAD_CONFIG    = $000F // Invalid Configuration: Began creating a new VD and never finished.
																//   Lost connection with owner while creating a new VD.
																//   Trying to configure a node that is owned by a different node.
HIQ_ERROR_FLASH         = $0010 // Flash Error: An error was encountered during the requested flash operation.
HIQ_ERROR_NOT_ROUTER    = $0011 // Not a Router:AhdrOther node has requested a guaranteed delivery connection through this node to a third node, and this node is not a router.

(***********************************************************)
(*              DATA TYPE DEFINITIONS GO BELOW             *)
(***********************************************************)
DEFINE_TYPE                         // NetLinx

STRUCT _sNetLinx_HiQNet_control_info
{
 CHAR hostname[64]                  // A friendly name for this connection, the hostname is recommended
 CHAR node[5]                       // HiQNet address as a string in decimal format
 // IP_CLIENT_OPEN info
 INTEGER dvPort                     // local device 0:port:0 for IP_CLIENT_OPEN, cannot use 0 or 1
 CHAR ipAddress[16]                 // ipV4 address
// INTEGER ipPort                     // 3804 for HiQnet
// INTEGER ipProtocol // IP_TCP for HiQnet most 3rd party control, IP_UDP_2WAY for Disco Broadcast
                                         
}

//FYI already defined in NetLinx.axi
// USED with GET_IP_ADDRESS(), SET_IP_ADDRESS()
//STRUCTURE IP_ADDRESS_STRUCT
//{
//  CHAR     FLAGS;          // Configuration flags DHCP = 1, Static = 0, other flags not used
//  CHAR     HOSTNAME[128];  // Host name
//  CHAR     IPADDRESS[32];  // IP address unit
//  CHAR     SUBNETMASK[32]; // subnet mask
//  CHAR     GATEWAY[32];    // IP address of gateway
//}

DEFINE_TYPE                         // HiQnet
/***************************************************************************/
/* NX to HiQ device: data type -> RAW_BE() -> message to HiQ device.       */
/* HiQ device to NX: from HiQ device -> STRING_TO_VARIABLE() -> data type. */
/***************************************************************************/

STRUCT _sHiQNet_Common_Header
{
 CHAR    A_cVersion                 // 1 always "$02"
 CHAR    B_cLengthHeader            // 2
 LONG    C_lLengthMessage           // 3,4,5,6
 INTEGER D_nSourceNode              // 7,8
 LONG    E_lSourceVDObject          // 9,10,11,12
 INTEGER F_nDestinationNode         // 13,14
 LONG    G_lDestinationVDObject     // 15,16,17,18
 INTEGER H_nMessageID               // 19,20
 INTEGER I_nFlags                   // 21,22
 CHAR    J_cHop_count               // 23 usually "$05"
 INTEGER K_nSequence                // 24,25
}

STRUCT _sHiQnet_Error_Header
{
 INTEGER L_nCode                    // see DEFINE_CONSTANT
// Strings are Unicode and stored using the String data type.
// The actual string data is prefixed with a UWORD count indicating the length of the string in bytes.
// Strings sent over the network are to include the NULL character at the end of the string.
// Because we are using Unicode, the length used in the String format is 2 * (strlen + 1). 
// For example:  for the string Hello World, the count will be 24.
 INTEGER M_nLength
 CHAR N_utf16[4000]
 CHAR str[2000]                     // The Unicode string reformatted as ASCII (upper bytes stripped)
}

STRUCT _sHiQnet_MultiPart_Header
{
 INTEGER O_nStart_sequence          // initially hqData.hdrCommon.K_nSequence
 LONG P_lBytes_remaining            // if lBytes_remaining = # bytes left in this message
																		// after the optional Session # and Flag Mask, this is the last one...
}

STRUCT _sHiQnet_Session_Header
{
 INTEGER Q_nIncoming                // Session header received from HiQnet device in communication
 INTEGER Q_nOther                   // Session header to be used in communication to the other device
}

STRUCT _sHiQNet_Attribute_UC
{
 INTEGER  nID
 CHAR     cDataType
 INTEGER  nLength
 CHAR     utf16[256]
 CHAR     str[256]
}

STRUCT _sHiQNet_Attribute_UWord
{
 INTEGER nID
 CHAR    cDataType
 INTEGER nWord
}

STRUCT _sHiQNet_Attribute_Block
{
 INTEGER nID
 CHAR    cDataType
 INTEGER nLength
 CHAR    str[256]
}

STRUCT _sHiQNet_Attributes_Device_Manager
{
 _sHiQNet_Attribute_UC classname    // .nID = $0000 Type of Device expected,
 _sHiQNet_Attribute_UC name         // .nID = $0001 Device Name expected
 _sHiQNet_Attribute_UWord flags     // .nID = $0002
 _sHiQNet_Attribute_Block sn        // .nID = $0003 Serial Number
 _sHiQNet_Attribute_UC version      // .nID = $0004 Software Version 
}

struct _sHiQnet_session_payload     // the first part of HELLO QUERY/INFO payload
{
 INTEGER a_nSelf
}

STRUCT _sHiQnet_Flag_Mask
{
 INTEGER b_nFlags                   // the second part of HELLO QUERY/INFO payload
}

struct _sHiQnet_HELLO_PAYLOAD
{
 _sHiQnet_session_payload session
 _sHiQnet_Flag_Mask mask
}

STRUCT _sHiQnet_Network_Info
{
 LONG h_lMAC_OUI                      // 24 bits Organizationally Unique Identifier
 LONG i_lMAC_NIC                      // 24 bits NIC specific bytes
 CHAR     J_cflagDHCP                 // 1 = DHCP, 0 = STATIC
 LONG     k_lIPAddress                // IP bytes
 LONG     l_lSubnetMask               // Subnet Mask bytes
 LONG     m_lGateway                  // GATEWAY bytes, use 0000 for unspecified ('0,0,0,0')
}

STRUCT _sHiQNet_DISCO_PAYLOAD
{
//_sHiQNet_Common_Header common // use common.nDestionNode == $FFFF for DISCO INFO broadcast
																		// DISCO Query == (.hdrCommon.I_nFlags & HIQ_MID_FLAG_INFORMATION ($0004) = 0)
																		// DISCO Info  == (.hdrCommon.I_nFlags & HIQ_MID_FLAG_INFORMATION ($0004) = 4)
 INTEGER a_nSourceNode              // HiQnet address of sender == .hdrCommon.D_nSourceNode
 CHAR    b_cCost                    // HIQ_ROUTING_COST_ETHERNET for NetLinx TCP connection. If proxies are in play(?): "Aggregated cost of route back to src".
 INTEGER c_nLengthSerialNumber      // 16 is expected. Serial number is a "block", 2 bytes for length, then the body in a string array of that length.
 CHAR    d_sSerialNumber[64]        // Typically 10x 0's and the MAC address, if the device has a NIC
 LONG    e_lMaxMessageSize          // Max Msg size sender can handle
 INTEGER f_nKeepAlivePeriod         // Keep Alive rate in ms. 10000ms is default, 250ms is min
 CHAR    g_cNetworkID               // HIQ_ROUTING_NID_ETHERNET for Netlinx TCP connection
 _SHiQnet_Network_Info network
}

STRUCT _sHiQnet_headers
{
 _sHiQNet_Common_Header hdrCommon
 _sHiQnet_Error_Header hdrError
 _sHiQnet_MultiPart_Header hdrMultiPart
 _sHiQnet_Session_Header hdrSession
}

(***********************************************************)
(*               VARIABLE DEFINITIONS GO BELOW             *)
(***********************************************************)
DEFINE_VARIABLE

VOLATILE DEV dvTP[]   = {dvTP1,dvTP2,dvTP3}
VOLATILE DEV dvTPG4[8]
VOLATILE DEV dvTPG5[8]
CONSTANT INTEGER DEVICE_ID_G5[] =   // from firmware release readme.txt
{
 $0196, $0197, $0198, $0199, $019A, $019B, $019C, $019D, $019E, $019F, $01A0, $01A1, $01D6
}

VOLATILE DEV dvHQN[] = {dvSelf,     // dvSelf	is used only once at bootup, is here as a placeholder to keep the arrays aligned
												dvHQN_2,dvHQN_3,dvHQN_4,dvHQN_5,dvHQN_6,dvHQN_7,
												dvHQN_8,dvHQN_9,dvHQN_10,dvHQN_11,dvHQN_12,dvHQN_13,
												dvHQN_14,dvHQN_15,dvHQN_16,dvHQN_17,dvHQN_18,dvHQN_19,
												dvHQN_20,dvHQN_21,dvHQN_22,dvHQN_23,dvHQN_24,dvHQN_25}

INTEGER flagHQN_online[25]

																		// 0 = off, 1 = minimal, 2 = verbose, 3 = more verbose
CONSTANT DEBUG_TERSE = 1
CONSTANT DEBUG_VERBOSE = 2
CONSTANT DEBUG_VVV = 3
CONSTANT DEBUG_VVVV = 4
INTEGER debugHQN = DEBUG_VERBOSE

INTEGER ndxSelf = 1
INTEGER nodeSelf = 51               // HiQnet default 3rd party control node

																		// initially set in DEFINE_START
_sNetLinx_HiQNet_control_info      ipHiQNet[25] // 1 is dvSelf
_sHiQnet_headers                   hqDataIn[25] 
_sHiQnet_headers                   hqDataOut[25]
_sHiQnet_HELLO_PAYLOAD             hqPH_In[25]
_sHiQnet_HELLO_PAYLOAD             hqPH_Out[25]

_sHiQNet_DISCO_PAYLOAD             hqPD_DvMgr[25]
_sHiQNet_Attributes_Device_Manager hqAtr_DM[25] // .name = ipHiQNet.hostname


IP_ADDRESS_STRUCT ipNetLinx
CHAR flagsNX_GET_MAC                // $1 = GET MAC, $2 = ONLINE. For NX MAC Address retrieval via DATA_EVENT[dvSelf]

CONSTANT tidHQnDiscoBC = 1
CONSTANT tidHQN_2  = 2              // if the HiQNet TID's are changed,
CONSTANT tidHQN_3  = 3              // more code will have to be added to
CONSTANT tidHQN_4  = 4              // deal with array offset #'s
CONSTANT tidHQN_5  = 5
CONSTANT tidHQN_6  = 6
CONSTANT tidHQN_7  = 7
CONSTANT tidHQN_8  = 8
CONSTANT tidHQN_9  = 9
CONSTANT tidHQN_10 = 10
CONSTANT tidHQN_11 = 11
CONSTANT tidHQN_12 = 12
CONSTANT tidHQN_13 = 13
CONSTANT tidHQN_14 = 14
CONSTANT tidHQN_15 = 15
CONSTANT tidHQN_16 = 16
CONSTANT tidHQN_17 = 17
CONSTANT tidHQN_18 = 18
CONSTANT tidHQN_19 = 19
CONSTANT tidHQN_20 = 20
CONSTANT tidHQN_21 = 21
CONSTANT tidHQN_22 = 22
CONSTANT tidHQN_23 = 23
CONSTANT tidHQN_24 = 24
CONSTANT tidHQN_25 = 25

LONG tidHQN[25] = // DO NOT use this array in a TIMELINE_EVENT[] statement
{
 tidHQnDiscoBC,
 tidHQN_2,tidHQN_3,tidHQN_4,tidHQN_5,tidHQN_6,tidHQN_7,
 tidHQN_8,tidHQN_9,tidHQN_10,tidHQN_11,tidHQN_12,tidHQN_13,
 tidHQN_14,tidHQN_15,tidHQN_16,tidHQN_17,tidHQN_18,tidHQN_19,
 tidHQN_20,tidHQN_21,tidHQN_22,tidHQN_23,tidHQN_24,tidHQN_25
}

LONG lHQnDiscoBC[1] = { 2000 }      // disco broadcast timeline
LONG lHQN_KeepAlive[25][1] =
{
 { 10000 }, // index 1 is placeholder
 { 10000 }, { 10000 }, { 10000 }, { 10000 }, { 10000 }, { 10000 },
 { 10000 }, { 10000 }, { 10000 }, { 10000 }, { 10000 }, { 10000 },
 { 10000 }, { 10000 }, { 10000 }, { 10000 }, { 10000 }, { 10000 },
 { 10000 }, { 10000 }, { 10000 }, { 10000 }, { 10000 }, { 10000 }
}
FLOAT lHQN_KAP_p[25] =               // What proportion of KAP to send KA message? 0.5 to 0.9 expected
{
 0.9, // index 1 is placeholder
 0.9,0.9,0.9,0.9,0.9,0.9,
 0.9,0.9,0.9,0.9,0.9,0.9,
 0.9,0.9,0.9,0.9,0.9,0.9,
 0.9,0.9,0.9,0.9,0.9,0.9
}

flagHQN_NoSession[25] =              // TRUE = don't use sessions
{
 1, // index 1 is placeholder
 1,1,1,1,1,1 ,1,1,1,1,1,1, 1,1,1,1,1,1, 1,1,1,1,1,1
}

flagSubscribeAll_after_GA[25] =      // set to true with initial connection, set false after the subscribe is sent
{
 0,  // 1 is placeholder
 0,0,0,0,0,0, 0,0,0,0,0,0, 0,0,0,0,0,0, 0,0,0,0,0,0
}

INTEGER nHQnDiscoBC                 // HiQnet specifies 5 repetitions of Disco BC... 1 often suffices.
INTEGER flagHQnDiscoBC_Online
INTEGER flagHQnDiscoListener = TRUE

INTEGER rowTP

INTEGER btnDiscoSelectCol1[] = { 9, 13,17,21,25, 29,33,37,41, 45,49,53,57, 61,65,69,73, 77,81,85,89, 93, 97,101,105}
INTEGER btnDiscoSelectCol2[] = {10, 14,18,22,26, 30,34,38,42, 46,50,54,58, 62,66,70,74, 78,82,86,90, 94, 98,102,106}
INTEGER btnDiscoSelectCol3[] = {11, 15,19,23,27, 31,35,39,43, 47,51,55,59, 63,67,71,75, 79,83,87,91, 95, 99,103,107}
INTEGER btnDiscoSelectCol4[] = {12, 16,20,24,28, 32,36,40,44, 48,52,56,60, 64,68,72,76, 80,84,88,92, 96,100,104,108}

INTEGER adrHQnDeviceType[] = { 9, 13,17,21,25, 29,33,37,41, 45,49,53,57, 61,65,69,73, 77,81,85,89, 93, 97,101,105}
INTEGER adrHQnDeviceNode[] = {10, 14,18,22,26, 30,34,38,42, 46,50,54,58, 62,66,70,74, 78,82,86,90, 94, 98,102,106}
INTEGER adrHQn_IPAddress[] = {11, 15,19,23,27, 31,35,39,43, 47,51,55,59, 63,67,71,75, 79,83,87,91, 95, 99,103,107}
INTEGER adrHQnDeviceName[] = {12, 16,20,24,28, 32,36,40,44, 48,52,56,60, 64,68,72,76, 80,84,88,92, 96,100,104,108}

INTEGER nDiscoTxtMax = 18           // max chars of text on the disco page row buttons
INTEGER nDiscoUniMax = 72           // 4x nDiscoTxtMax

CHAR sDiscoFontLargeG4[] = '43'     // Lucida Sans Unicode 16
CHAR sDiscoFontSmallG4[] = '32'     // Lucida Sans Unicode 12
CHAR sDiscoFontLargeG5[] = 'Lucida Sans Unicode:16'
CHAR sDiscoFontSmallG5[] = 'Lucida Sans Unicode:12'

INTEGER btnDiscoMatched[] =
{
 110,
 111,112,113,114,115,116,117,118,119,120,121,122,
 123,124,125,126,127,128,129,130,131,132,133,134
}

INTEGER adrDiscoRowNumber[] =
{
 110,
 111,112,113,114,115,116,117,118,119,120,121,122,
 123,124,125,126,127,128,129,130,131,132,133,134
}

INTEGER btnHQnOnline[] =
{
 135,
 136,137,138,139,140,141,142,143,144,145,146,147,
 148,149,150,151,152,153,154,155,156,157,158,159
}

INTEGER adrHQnSavedRowNumber[] =
{
 135,
 136,137,138,139,140,141,142,143,144,145,146,147,
 148,149,150,151,152,153,154,155,156,157,158,159
}


INTEGER flagDiscoRowSelected[25]

																		// disco broadcast outgoing
// ipHiQDiscoBC.hostname  = 'HQN BROADCAST'
// ipHiQDiscoBC.node      = '65535'
// ipHiQDiscoBC.dvPort    = dvHQnDiscoBC.PORT
// ipHiQDiscoBC.ipAddress = '255.255.255.255' // this could also be the local subnet broadcast IP, e.g. 192.168.1.255
_sNetLinx_HiQNet_control_info ipHiQDiscoBC
                                    // set from incoming disco broadcasts
_sHiQNet_DISCO_PAYLOAD hqPD_DiscoBC[25]// HiQnet disco payloads - discovered via broadcast
INTEGER flagDiscoMatched[25]
 
(***********************************************************)
(*               LATCHING DEFINITIONS GO BELOW             *)
(***********************************************************)
DEFINE_LATCHING

(***********************************************************)
(*       MUTUALLY EXCLUSIVE DEFINITIONS GO BELOW           *)
(***********************************************************)
DEFINE_MUTUALLY_EXCLUSIVE

(***********************************************************)
(*        SUBROUTINE/FUNCTION DEFINITIONS GO BELOW         *)
(***********************************************************)
(* EXAMPLE: DEFINE_FUNCTION <RETURN_TYPE> <NAME> (<PARAMETERS>) *)
(* EXAMPLE: DEFINE_CALL '<NAME>' (<PARAMETERS>) *)

DEFINE_FUNCTION LONG lengthHiQDiscoPayload (INTEGER nLengthSerialNumber)
{ 
 //  2 - 2 bytes hqD.disco.a_nSourceNode
 //  3 - 1 bytes hqD.disco.b_cCost
 //  5 - 2 bytes hqD.disco.c_nLengthSerialNumber
 //  9 - 4 bytes hqD.disco.e_lMaxMessageSize
 // 11 - 2 bytes hqD.disco.f_nKeepAlivePeriod
 // 12 - 1 bytes hqD.disco.g_cNetworkID
 // 15 - 3 bytes hqD.disco.network.h_lMAC_OUI
 // 18 - 3 bytes hqD.disco.network.i_lMAC_NIC
 // 19 - 1 bytes hqD.disco.network.J_cflagDHCP
 // 23 - 4 bytes hqD.disco.network.k_lIPAddress
 // 27 - 4 bytes hqD.disco.network.l_lSubnetMask
 // 31 - 4 bytes hqD.disco.network.m_lGateway
 RETURN 31 + nLengthSerialNumber    // $2F (47) if SN is 10x 0's and MAC address
}

DEFINE_FUNCTION CHAR lengthHiQHeaders (INTEGER nFlags)
{
 STACK_VAR CHAR ret
 ret = $19                          // 25 bytes common header
 IF (nFlags & HIQ_MID_FLAG_ERROR == HIQ_MID_FLAG_ERROR)
 {
	ret = ret + 2 /* error code */    // if ERROR the other headers probably don't matter as headers...but session # may be part of the error string as terminator after "$00,$00", e.g. 4x3500HD, DCiN
 }
 IF (nFlags & HIQ_MID_FLAG_MULTI_PART_MSG == HIQ_MID_FLAG_MULTI_PART_MSG)
 {
	ret = ret + 6
 }
 IF (nFlags & HIQ_MID_FLAG_SESSION_NUMBER == HIQ_MID_FLAG_SESSION_NUMBER)
 {
	ret = ret + 2
 }
 RETURN ret
}

DEFINE_FUNCTION buildHiQHeaders (_sHiQNet_Common_Header hqCH,
																 INTEGER nFlags,
																 INTEGER nErrorCode,
																 _sHiQnet_MultiPart_Header hqMPH,
																 INTEGER nSessionHeader,
																 CHAR sOutgoing[])
{
 																		// start with common header bytes
 sOutgoing = "hqCH.A_cVersion,
							hqCH.B_cLengthHeader,
							RAW_BE(hqCH.C_lLengthMessage),
							RAW_BE(hqCH.D_nSourceNode),
							RAW_BE(hqCH.E_lSourceVDObject),
							RAW_BE(hqCH.F_nDestinationNode),
							RAW_BE(hqCH.G_lDestinationVDObject),
							RAW_BE(hqCH.H_nMessageID),
							RAW_BE(hqCH.I_nFlags),
							hqCH.J_cHop_count,
							RAW_BE(hqCH.K_nSequence)"
 IF (nflags & HIQ_MID_FLAG_ERROR == HIQ_MID_FLAG_ERROR)
 {
	sOutgoing = "sOutgoing,RAW_BE(nErrorCode)"
 }
 IF (nFlags & HIQ_MID_FLAG_MULTI_PART_MSG == HIQ_MID_FLAG_MULTI_PART_MSG)
 {
	sOutgoing = "sOutgoing,RAW_BE(hqMPH.O_nStart_sequence),RAW_BE(hqMPH.P_lBytes_remaining)"
 }
 IF (nFlags & HIQ_MID_FLAG_SESSION_NUMBER == HIQ_MID_FLAG_SESSION_NUMBER)
 {
	sOutgoing = "sOutgoing,RAW_BE(nSessionHeader)"
 }
}

DEFINE_FUNCTION buildHiQDisco (_sHiQNet_DISCO_PAYLOAD hqPD, _sHiQnet_headers hqDO, INTEGER nFlags, CHAR sOutgoing[], INTEGER nLine, CHAR sDebug[])
// I_nFlags: Expected HIQ_MID_FLAG_GUARANTEED | HIQ_MID_FLAG_INFORMATION for Disco Info
//         HIQ_MID_FLAG_GUARANTEED for Disco Query
//         other flags as necessary...
{
 // build DISCO INFO command
 IF (!hqDO.hdrCommon.A_cVersion) { hqDO.hdrCommon.A_cVersion = $02 }
                                    // set the lengths...
 hqDO.hdrCommon.B_cLengthHeader = lengthHiQHeaders(nFlags)
 																		// bytes remaining = length of hqDataOut[i].disco
 hqDO.hdrMultiPart.P_lBytes_remaining = lengthHiQDiscoPayload(hqPD.c_nLengthSerialNumber)
 hqDO.hdrCommon.C_lLengthMessage = TYPE_CAST(hqDO.hdrCommon.B_cLengthHeader + 
																						 hqDO.hdrMultiPart.P_lBytes_remaining)

																		// if you accidentally set H_nMessageID to HIQ_MID_DEVICE_GET_ATTRIBUTES here,
																		// 4x3500HD will generate some interesting errors...
 hqDO.hdrCommon.H_nMessageID = HIQ_MID_ROUTING_DISCO_INFO
//#warn'ITHD4 error testing'
// hqDO.hdrCommon.H_nMessageID = HIQ_MID_DEVICE_GET_ATTRIBUTES
 
 hqDO.hdrCommon.I_nFlags = nFlags
 
 IF (!hqDO.hdrCommon.J_cHop_count) { hqDO.hdrCommon.J_cHop_count = $05 }
 
 hqDO.hdrCommon.K_nSequence++
 IF (!hqDO.hdrCommon.K_nSequence) { hqDO.hdrCommon.K_nSequence++ }
 
 IF (hqDO.hdrCommon.I_nFlags & HIQ_MID_FLAG_MULTI_PART_MSG == HIQ_MID_FLAG_MULTI_PART_MSG)
 {
	hqDO.hdrMultiPart.O_nStart_sequence = hqDO.hdrCommon.K_nSequence
	// bytes remaining was already set when we calculated the lengths
 }

 IF (debugHQN >= DEBUG_VERBOSE)
 {
	printHiQHeaders(hqDO,hqDO.hdrSession.Q_nOther,__LINE__,sDebug)
 }
 
 buildHiQHeaders(hqDO.hdrCommon,
								 hqDO.hdrCommon.I_nFlags,
								 hqDO.hdrError.L_nCode,
								 hqDO.hdrMultiPart,
								 hqDO.hdrSession.Q_nOther,
								 sOutgoing)

 IF (debugHQN >= DEBUG_VERBOSE)
 {
	printHiQPayloadDisco(hqPD,__LINE__,sDebug)
 }
																		// add the Disco payload
 sOutgoing = "sOutgoing,
							RAW_BE(hqPD.a_nSourceNode),
										 hqPD.b_cCost,
							RAW_BE(hqPD.c_nLengthSerialNumber),
										 hqPD.d_sSerialNumber,
							RAW_BE(hqPD.e_lMaxMessageSize),
							RAW_BE(hqPD.f_nKeepAlivePeriod),
										 hqPD.g_cNetworkID,
 RIGHT_STRING(RAW_BE(hqPD.network.h_lMAC_OUI),3),
 RIGHT_STRING(RAW_BE(hqPD.network.i_lMAC_NIC),3),
							RAW_BE(hqPD.network.J_cflagDHCP),
							RAW_BE(hqPD.network.k_lIPAddress),
							RAW_BE(hqPD.network.l_lSubnetMask),
							RAW_BE(hqPD.network.m_lGateway)"

 IF (debugHQN >= DEBUG_VERBOSE)
 {
	printToDebug(0:1:0,"ITOA(nLine),
											FORMAT(' $%X ',LENGTH_ARRAY(sOutgoing)),
											sDebug,' '",
											sOutgoing,
											131,0,PRINT_HEX)
 }
}


DEFINE_FUNCTION buildHiQGetAttributes (_sHiQnet_headers hqDO, INTEGER nAttributeCount, INTEGER nAID[], CHAR sOutgoing[], CHAR sDebug[])
{
 STACK_VAR INTEGER i
 STACK_VAR INTEGER nFlags
 nFlags = hqDO.hdrCommon.I_nFlags
 
 IF (!hqDO.hdrCommon.A_cVersion) { hqDO.hdrCommon.A_cVersion = $02 }
 hqDO.hdrCommon.B_cLengthHeader = lengthHiQHeaders(nFlags)
 hqDO.hdrCommon.C_lLengthMessage = hqDO.hdrCommon.B_cLengthHeader
																	 // add 2 to length for attribute count
                                   + 2
																	 // and 2 for each attribute polled
																	 + LENGTH_VARIABLE_TO_STRING(nAID) - 3
 IF (!hqDO.hdrCommon.D_nSourceNode) { hqDO.hdrCommon.D_nSourceNode = nodeSelf }

 hqDO.hdrCommon.H_nMessageID = HIQ_MID_DEVICE_GET_ATTRIBUTES
 hqDO.hdrCommon.I_nFlags = HIQ_MID_FLAG_GUARANTEED
 IF (!hqDO.hdrCommon.J_cHop_count) { hqDO.hdrCommon.J_cHop_count = $05 }
 hqDO.hdrCommon.K_nSequence++
 IF (!hqDO.hdrCommon.K_nSequence) { hqDO.hdrCommon.K_nSequence++ }

																		// build the headers
 buildHiQHeaders(hqDO.hdrCommon,
								 hqDO.hdrCommon.I_nFlags,
								 hqDO.hdrError.L_nCode,
								 hqDO.hdrMultiPart,
								 hqDO.hdrSession.Q_nOther,
								 sOutgoing)
 
																		// then the attribute count...
 sOutgoing = "sOutgoing,RAW_BE(nAttributeCount)"

																		// some devices (ITHD4, DCIN are ok with multiple attributes polled at init; others not CDi DC
 FOR (i = 1; i <= nAttributeCount; i++)
 {
	sOutgoing = "sOutgoing,RAW_BE(nAID[i])"
 }

 // debugging
 IF (debugHQN >= DEBUG_VERBOSE)
 {
	STACK_VAR CHAR sAID[128]
	VARIABLE_TO_STRING(nAID,sAID,1)
	sAID = RIGHT_STRING(sAID,LENGTH_ARRAY(sAID)-3) // strip marshalling bytes
	
	printHiQHeaders(hqDO,hqDO.hdrSession.Q_nOther,__LINE__,sDebug)
	printToDebug(0:1:0,"ITOA(__LINE__),
										 ' ',sDebug,' Attribute Payload '",
										 "RAW_BE(nAttributeCount),sAID",131,0,PRINT_HEX)
	printToDebug(0:1:0,"ITOA(__LINE__),
										 FORMAT(' $%X ',LENGTH_ARRAY(sOutgoing)),
										 sDebug,' '",
										 sOutgoing,131,0,PRINT_HEX)
 }
}

DEFINE_FUNCTION buildHiQHello (_sHiQnet_HELLO_PAYLOAD hqPHO,_sHiQnet_headers hqDO, INTEGER nFlags, CHAR sOutgoing[],INTEGER nLine, CHAR sDebug[])
// I_nFlags: Expected HIQ_MID_FLAG_GUARANTEED | HIQ_MID_FLAG_INFORMATION for Hello Info
//         HIQ_MID_FLAG_GUARANTEED for Hello Query
//         other flags as necessary...
{
 IF (!hqDO.hdrCommon.A_cVersion) { hqDO.hdrCommon.A_cVersion = $02 }
																	  // Set lengths
 hqDO.hdrCommon.B_cLengthHeader = lengthHiQHeaders(nFlags)
 																	  // add 2 to message length for session payload
																		// add 2 to message length for flags mask
 hqDO.hdrCommon.C_lLengthMessage = hqDO.hdrCommon.B_cLengthHeader + 4
 
 IF (!hqDO.hdrCommon.D_nSourceNode) { hqDO.hdrCommon.D_nSourceNode = nodeSelf }
 
 hqDO.hdrCommon.H_nMessageID = HIQ_MID_ROUTING_HELLO
 hqDO.hdrCommon.I_nFlags = nFlags   // initial HELLO QUERY not flagged for SESSION in HiQNet Spec...
 IF (!hqPHO.mask.b_nflags)         // IF! to allow for testing other flag settings
 {
	hqPHO.mask.b_nflags = $01FF      // The minimum support for the flag mask is: 0x01FF
 }
 IF (!hqDO.hdrCommon.J_cHop_count) { hqDO.hdrCommon.J_cHop_count = $05 }
 hqDO.hdrCommon.K_nSequence++
 IF (!hqDO.hdrCommon.K_nSequence) { hqDO.hdrCommon.K_nSequence++ }
  
 //#warn 'todo: add Multi Part message support?

																		// build the Hello string
																		// build the headers
 buildHiQHeaders(hqDO.hdrCommon,
								 hqDO.hdrCommon.I_nFlags,
								 hqDO.hdrError.L_nCode,
								 hqDO.hdrMultiPart,
								 hqDO.hdrSession.Q_nOther,
								 sOutgoing)

 sOutgoing = "sOutgoing,
							RAW_BE(hqPHO.session.a_nSelf),
							RAW_BE(hqPHO.mask.b_nflags)"

 // debugging
 IF (debugHQN >= DEBUG_TERSE)
 {
 	printHiQHeaders(hqDO,0, nLine,sDebug)
	AMX_LOG(AMX_WARNING,"ITOA(nLine),' ',sDebug,FORMAT(' Payload session $%04X TO OTHER',hqPHO.session.a_nSelf)")
	AMX_LOG(AMX_WARNING,"ITOA(nLine),' ',sDebug,' mask ',returnHiqMessageFlags(hqPHO.mask.b_nflags)")
 }
 IF (debugHQN >= DEBUG_VERBOSE)
 {
	printToDebug(0:1:0,"ITOA(nLine),
											FORMAT(' $%X ',LENGTH_ARRAY(sOutgoing)),
											sDebug,' '",
											sOutgoing,131,0,PRINT_HEX)
 }
}

DEFINE_FUNCTION buildHiQHelloError (_sHiQnet_headers hqDO, INTEGER nFlags, CHAR sOutgoing[], INTEGER nLine, CHAR sDebug[])
// I_nFlags: Expected HIQ_MID_FLAG_GUARANTEED | HIQ_MID_FLAG_INFORMATION | HIQ_MID_FLAG_ERROR
//         other flags as necessary...
{
 IF (!hqDO.hdrCommon.A_cVersion) { hqDO.hdrCommon.A_cVersion = $02 }
																	  // Set lengths
 hqDO.hdrCommon.B_cLengthHeader = lengthHiQHeaders(nFlags)
 																	  // add 2 to message length for Error Code
																		// add 2 to message length for Error Length
 hqDO.hdrCommon.C_lLengthMessage = hqDO.hdrCommon.B_cLengthHeader + 4

 IF (!hqDO.hdrCommon.D_nSourceNode) { hqDO.hdrCommon.D_nSourceNode = nodeSelf }

 hqDO.hdrCommon.H_nMessageID = HIQ_MID_ROUTING_HELLO
 hqDO.hdrCommon.I_nFlags = nFlags // | HIQ_MID_FLAG_ERROR | HIQ_MID_FLAG_GUARANTEED | HIQ_MID_FLAG_INFORMATION
                                    // no flag mask necessary per HiQnet 3rd party quickstart guide
// IF (!hqDO.mask.b_nflags)           // IF! to allow for testing other flag settings
// {
//	hqDO.mask.b_nflags = $01FF        // The minimum support for the flag mask is: 0x01FF
// }
 IF (!hqDO.hdrCommon.J_cHop_count) { hqDO.hdrCommon.J_cHop_count = $05 }
 hqDO.hdrCommon.K_nSequence++
 IF (!hqDO.hdrCommon.K_nSequence) { hqDO.hdrCommon.K_nSequence++ }
 hqDO.hdrError.L_nCode = HIQ_ERROR_BAD_LENGTH // $0002 per HiQnet 3rd party quickstart guide
 hqDO.hdrError.M_nLength = $0000    // "$00,$00" per HiQnet 3rd party quickstart guide 

																		// build the Hello string
																		// build the headers
 buildHiQHeaders(hqDO.hdrCommon,
								 hqDO.hdrCommon.I_nFlags,
								 hqDO.hdrError.L_nCode,
								 hqDO.hdrMultiPart,
								 hqDO.hdrSession.Q_nOther,
								 sOutgoing)

 sOutgoing = "sOutgoing,
							RAW_BE(hqDO.hdrError.L_nCode),
							RAW_BE(hqDO.hdrError.M_nLength)"

 // debugging
 IF (debugHQN >= DEBUG_TERSE)
 {
 	printHiQHeaders(hqDO,0, nLine,sDebug)
 }
 IF (debugHQN >= DEBUG_VERBOSE)
 {
	printToDebug(0:1:0,"ITOA(nLine),
											FORMAT(' $%X ',LENGTH_ARRAY(sOutgoing)),
											sDebug,' '",
											sOutgoing,131,0,PRINT_HEX)
 }
}

DEFINE_FUNCTION buildHiQMultiParameterGet (_sHiQnet_headers hqDO, CHAR sOutgoing[], CHAR sDebug[])
{
 STACK_VAR INTEGER nFlags
 STACK_VAR INTEGER nParameterCount
 // Build Multi-Parameter Get : "Common header, maybe Session header, Parameter Count"
 // this first version is for the initial query when going online, after DISCO INFO
 // Common header unique data needed:
 // Message Length
 // Parameter Count
 // flags
 // Message ID
 // Session unique data needed:
 // UseOther Sessions flag, if true...
 // Session Other header
 nFLags = hqDO.hdrCommon.I_nFlags | HIQ_MID_FLAG_GUARANTEED
 hqDO.hdrCommon.B_cLengthHeader = lengthHiQHeaders(nFlags)
 hqDO.hdrCommon.C_lLengthMessage = hqDO.hdrCommon.B_cLengthHeader
																 + LENGTH_VARIABLE_TO_STRING(nParameterCount)
																 - 1
 #warn 'add support for Parameter Counts > 0'
 hqDO.hdrCommon.H_nMessageID = HIQ_MID_MULTISV_GET
 hqDO.hdrCommon.I_nFlags = nFlags
 hqDO.hdrCommon.K_nSequence++
 IF (!hqDO.hdrCommon.K_nSequence) { hqDO.hdrCommon.K_nSequence++ }
 
																		// build the headers
 buildHiQHeaders(hqDO.hdrCommon,
								 hqDO.hdrCommon.I_nFlags,
								 hqDO.hdrError.L_nCode,
								 hqDO.hdrMultiPart,
								 hqDO.hdrSession.Q_nOther,
								 sOutgoing)

 sOutgoing = "sOutgoing,RAW_BE(nParameterCount)"
 // todo: add parameters...

 // debugging
 IF (debugHQN >= DEBUG_TERSE)
 {
	printHiQHeaders(hqDO,hqDO.hdrSession.Q_nOther,__LINE__,sDebug)
	AMX_LOG(AMX_WARNING,"ITOA(__LINE__),' ',sDebug,FORMAT(' Parameter Count $%04X',nParameterCount)")
 }
 IF (debugHQN >= DEBUG_VERBOSE)
 {
	printToDebug(0:1:0,"ITOA(__LINE__),FORMAT(' $%X ',LENGTH_ARRAY(sOutgoing)),sDebug,' '",sOutgoing,131,0,PRINT_HEX)
 }
}

DEFINE_FUNCTION HiQGoodbye (DEV dv, _sHiQnet_headers hqDO, CHAR sDebug[])
{
 STACK_VAR CHAR sOutgoing[256]

 // Build Goodbye string : "Common header, maybe Session header, Source Node"
 // Common header unique data needed:
 // Message Length
 // Source Node
 // flags
 // Message ID
 // Session unique data needed:
 // UseOther Sessions flag, if true...
 // Session Other header
 IF (!hqDO.hdrCommon.A_cVersion) { hqDO.hdrCommon.A_cVersion = $02 }
 hqDO.hdrCommon.B_cLengthHeader = $19
 hqDO.hdrCommon.C_lLengthMessage = hqDO.hdrCommon.B_cLengthHeader + 2 // 2 for source node payload
 IF (!hqDO.hdrCommon.D_nSourceNode) { hqDO.hdrCommon.D_nSourceNode = nodeSelf }
 hqDO.hdrCommon.H_nMessageID = HIQ_MID_ROUTING_GOODBYE
 IF (hqDO.hdrCommon.I_nFlags & HIQ_MID_FLAG_GUARANTEED  <> HIQ_MID_FLAG_GUARANTEED)
 {
	hqDO.hdrCommon.I_nFlags = hqDO.hdrCommon.I_nFlags | HIQ_MID_FLAG_GUARANTEED
 }
 IF (!hqDO.hdrCommon.J_cHop_count) { hqDO.hdrCommon.J_cHop_count = $05 }
 hqDO.hdrCommon.K_nSequence++

 IF (hqDO.hdrCommon.I_nFlags & HIQ_MID_FLAG_SESSION_NUMBER == HIQ_MID_FLAG_SESSION_NUMBER)
 {
	hqDO.hdrCommon.B_cLengthHeader = hqDO.hdrCommon.B_cLengthHeader + 2
	hqDO.hdrCommon.C_lLengthMessage = hqDO.hdrCommon.C_lLengthMessage + 2
 }
 
																		// build the headers
 buildHiQHeaders(hqDO.hdrCommon,
								 hqDO.hdrCommon.I_nFlags,
								 hqDO.hdrError.L_nCode,
								 hqDO.hdrMultiPart,
								 hqDO.hdrSession.Q_nOther,
								 sOutgoing)
																	 // add the goodbye node
 sOutgoing = "sOutgoing,           // end the string with NX HiQnet source node
							RAW_BE(hqDO.hdrCommon.D_nSourceNode)"

 // debugging
 IF (debugHQN >= DEBUG_TERSE)
 {
	printHiQHeaders(hqDO,hqDO.hdrSession.Q_nOther,__LINE__,sDebug)
	printToDebug(0:1:0,"ITOA(__LINE__),
										 FORMAT(' $%X ',LENGTH_ARRAY(sOutgoing)),
										 sDebug,' GOODBYE '",
										 sOutgoing,131,0,PRINT_HEX)
 }

 SEND_STRING dv,sOutgoing

 IP_CLIENT_CLOSE(dv.port)
 IF (debugHQN >= DEBUG_VVV)
 {
	AMX_LOG(AMX_ERROR,"ITOA(__LINE__),' ',sDebug,' IP_CLIENT_CLOSE(',ITOA(dv.port),')'")
 }
}

DEFINE_FUNCTION LONG parseHiQAttributesDeviceManager (_sHiQNet_Attributes_Device_Manager hqADM, CHAR buffer[], INTEGER nLine, CHAR sDebug[])
{
// FUNCTION: parseHiQGetAttributes()
// PURPOSE : destructively parse HiQNet attribute data from left side of buffer[],
//           setting values using the NetLinx Marshalling Protocol
// PARAMETERS:
//         LP - start variable at 0 for initial function call.
//              pass into successive parseHiQ functions to get the total message Length Parsed
// RETURN: #bytes parsed + #bytes previously parsed (from lp parameter)
 STACK_VAR LONG ret
 STACK_VAR INTEGER nAttributeCount
 STACK_VAR INTEGER nAttributeID
 STACK_VAR INTEGER i
 ret = LENGTH_ARRAY(buffer)

 IF (debugHQN >= DEBUG_VERBOSE)
 {
	AMX_LOG(AMX_WARNING,"ITOA(nLine),' ',sDebug,' parseHiQAttributesDeviceManager() start...'")
 }

 IF (debugHQN >= DEBUG_VVV)
 {
	printToDebug(0:1:0,"ITOA(__LINE__),' TBD (HEX) -> '",buffer,131,0,PRINT_HEX)
	printToDebug(0:1:0,"ITOA(__LINE__),' TBD (ASC) -> '",buffer,131,0,PRINT_ASCII_HEX)
	printToDebug(0:1:0,"ITOA(__LINE__),' TBD (DEC) -> '",buffer,131,0,PRINT_DECIMAL)
 }

 STRING_TO_VARIABLE(nAttributeCount,"$E2,GET_BUFFER_STRING(buffer,2)",1)
 
 IF (debugHQN >= DEBUG_TERSE)
 {
	AMX_LOG(AMX_ERROR,"ITOA(__LINE__),' hqADM Attribute Count ',ITOA(nAttributeCount)")
 }
 
 FOR (i = 1; i <= nAttributeCount; i++)
 {
	STRING_TO_VARIABLE(nAttributeID,"$E2,GET_BUFFER_STRING(buffer,2)",1)

	SWITCH (nAttributeID)
	{
	 CASE 4:
	 {
		hqADM.version.nID = nAttributeID
		STRING_TO_VARIABLE(hqADM.version.cDataType,"$E1,GET_BUFFER_STRING(buffer,1)",1)
		STRING_TO_VARIABLE(hqADM.version.nLength,"$E2,GET_BUFFER_STRING(buffer,2)",1)
		
		hqADM.version.utf16 = GET_BUFFER_STRING(buffer,hqADM.version.nLength)
		SET_LENGTH_ARRAY(hqADM.version.utf16,LENGTH_ARRAY(hqADM.version.utf16) - 2)
		UTF16BE_TO_ASCII(hqADM.version.utf16,hqADM.version.str)

		IF (debugHQN >= DEBUG_VERBOSE)
		{
		 AMX_LOG(AMX_ERROR,"ITOA(__LINE__),
											 FORMAT(' Attribute ID $%04X, ',nAttributeID),
											 returnHiQDataType(hqADM.version.cDataType),
											 ', SOFTWARE VERSION ''',hqADM.version.str,''''")
		}
	 }
	 CASE 3:
	 {
		hqADM.sn.nID = nAttributeID
		STRING_TO_VARIABLE(hqADM.sn.cDataType,"$E1,GET_BUFFER_STRING(buffer,1)",1)
		STRING_TO_VARIABLE(hqADM.sn.nLength,"$E2,GET_BUFFER_STRING(buffer,2)",1)

		hqADM.sn.str = GET_BUFFER_STRING(buffer,hqADM.sn.nLength)

		IF (debugHQN >= DEBUG_VERBOSE)
		{
		 printToDebug(0:1:0,"ITOA(__LINE__),
									 FORMAT(' Attribute ID $%04X, ',nAttributeID),
									 returnHiQDataType(hqADM.sn.cDataType),
									', SN '",
									hqADM.sn.str,131,0,PRINT_HEX)
		}
	 }
	 CASE 2:
	 {
		hqADM.flags.nID = nAttributeID
		STRING_TO_VARIABLE(hqADM.flags.cDataType,"$E1,GET_BUFFER_STRING(buffer,1)",1)
		STRING_TO_VARIABLE(hqADM.flags.nWord,"$E2,GET_BUFFER_STRING(buffer,2)",1)
		
		IF (debugHQN >= DEBUG_VERBOSE)
		{
		 AMX_LOG(AMX_ERROR,"ITOA(__LINE__),
											 FORMAT(' Attribute ID $%04X, ',nAttributeID),
											 returnHiQDataType(hqADM.flags.cDataType),
											', FLAGS ',returnHiQMessageFlags(hqADM.flags.nWord)")
		}
	 }
	 CASE 1:
	 {
		hqADM.name.nID = nAttributeID
		STRING_TO_VARIABLE(hqADM.name.cDataType,"$E1,GET_BUFFER_STRING(buffer,1)",1)
		STRING_TO_VARIABLE(hqADM.name.nLength,"$E2,GET_BUFFER_STRING(buffer,2)",1)
		
		hqADM.name.utf16 = GET_BUFFER_STRING(buffer,hqADM.name.nLength)
		SET_LENGTH_ARRAY(hqADM.name.utf16,LENGTH_ARRAY(hqADM.name.utf16) - 2)
		UTF16BE_TO_ASCII(hqADM.name.utf16,hqADM.name.str)
		
		IF (debugHQN >= DEBUG_VERBOSE)
		{
		 AMX_LOG(AMX_ERROR,"ITOA(__LINE__),
											 FORMAT(' Attribute ID $%04X, ',nAttributeID),
											 returnHiQDataType(hqADM.name.cDataType),
											 ', NAME ''',hqADM.name.str,''''")
		}
	 }
	 CASE 0:
	 {
		hqADM.classname.nID = nAttributeID
		STRING_TO_VARIABLE(hqADM.classname.cDataType,"$E1,GET_BUFFER_STRING(buffer,1)",1)
		STRING_TO_VARIABLE(hqADM.classname.nLength,"$E2,GET_BUFFER_STRING(buffer,2)",1)
		
		hqADM.classname.utf16 = GET_BUFFER_STRING(buffer,hqADM.classname.nLength)
		SET_LENGTH_ARRAY(hqADM.classname.utf16,LENGTH_ARRAY(hqADM.classname.utf16) - 2)
		UTF16BE_TO_ASCII(hqADM.classname.utf16,hqADM.classname.str)
		
		IF (debugHQN >= DEBUG_VERBOSE)
		{
		 AMX_LOG(AMX_ERROR,"ITOA(__LINE__),
											 FORMAT(' Attribute ID $%04X, ',nAttributeID),
											 returnHiQDataType(hqADM.classname.cDataType),
											 ', CLASS NAME ''',hqADM.classname.str,''''")
		}
	 }
	 DEFAULT:
	 {
		IF (debugHQN >= DEBUG_TERSE)
		{
		 AMX_LOG(AMX_ERROR,"ITOA(nLine),' ',
												sDebug,' errLine(',ITOA(__LINE__),') ',
												'parse Attribute ID ',ITOA(nAttributeID),' ToDo...'")	}
		}
	}
 }
 ret = ret - LENGTH_ARRAY(buffer)
 RETURN ret
}

DEFINE_FUNCTION LONG parseHiQData (_sHiQnet_headers hqDI,
																	 _sHiQnet_HELLO_PAYLOAD hqPHI,
																	 _sHiQNet_DISCO_PAYLOAD hqPD,
																	 _sHiQNet_Attributes_Device_Manager hqADM,
																	 CHAR buffer[], 
																	 INTEGER nLine, CHAR sDebug[])

// FUNCTION: parseHiQData()
// PURPOSE : destructively parse HiQNet data from left side of buffer[],
//           setting values in hqDI using the NetLinx Marshalling Protocol
// PARAMETERS:
// RETURN: #bytes parsed here (ret)
{
 STACK_VAR LONG ret
 ret = LENGTH_ARRAY(buffer)

 IF (debugHQN >= DEBUG_VERBOSE)
 {
	AMX_LOG(AMX_WARNING,"ITOA(nLine),' ',sDebug,' parseHiQData() start...'")
 }
 IF (buffer[1] == 2)                // if the HiQnet protocol version is correct
 {
																		/*****************/
																		/* parse Headers */
																		/*****************/
	parseHiQHeaders(hqDi,buffer,__LINE__,sDebug)
	printHiQHeaders(hqDI,hqDI.hdrSession.Q_nIncoming,__LINE__,sDebug)

																		/*******************************/
																		/* parse message specific data */
																		/*******************************/
	IF (debugHQN >= DEBUG_VERBOSE)
	{
	 AMX_LOG(AMX_WARNING,"ITOA(__LINE__),
									 ' ',sDebug,
									 ' message ID ',returnHiQMessageID(hqDI.hdrCommon.H_nMessageID),',',
									 ' flags ',returnHiQMessageFlags(hqDI.hdrCommon.I_nFlags),' parsing...'")
	}
	SWITCH (hqDI.hdrCommon.H_nMessageID)
	{
	 CASE HIQ_MID_ROUTING_DISCO_INFO:
	 {
		parseHiQPayloadDisco(hqPD,buffer,__LINE__,sDebug)
		printHiQPayloadDisco(hqPD,__LINE__,sDebug)
	 }
	 CASE HIQ_MID_ROUTING_HELLO:
	 {
																	 /*****************************************************/
																	 /* if this is the HELLO INFO response to HELLO QUERY */
																	 /*****************************************************/
		IF (hqDI.hdrCommon.I_nFlags & HIQ_MID_FLAG_INFORMATION == HIQ_MID_FLAG_INFORMATION)
		{
																	 // if information flag is set, the "Self" session # is expected as a payload
																	 // store incoming session number
		 STRING_TO_VARIABLE(hqPHI.session.a_nSelf,"$E2,GET_BUFFER_STRING(buffer,2)",1)
		 IF (debugHQN >= DEBUG_TERSE)
		 {
			AMX_LOG(AMX_WARNING,"ITOA(__LINE__),' ',sDebug,
																					' PAYLOAD SESSION ',
																					 FORMAT('$%04X from OTHER',hqPHI.session.a_nSelf),
																					 ' (',ITOA(hqPHI.session.a_nSelf),')'")
		 }
		}
																		// EXPECTED: Flags Mask is last two bytes of the HELLO INFO msg
		STRING_TO_VARIABLE(hqPHI.mask.b_nflags,"$E2,GET_BUFFER_STRING(buffer,2)",1)
		IF (debugHQN >= DEBUG_TERSE)
		{
		 IF (hqPHI.mask.b_nflags & hqDI.hdrCommon.I_nFlags <> hqDI.hdrCommon.I_nFlags)
		 {
			 AMX_LOG(AMX_ERROR,"ITOA(__LINE__),' ',sDebug,' flag mask expected ',
																				 FORMAT('$%04X',hqDI.hdrCommon.I_nFlags),
																				 ' flag mask received ',
																				 FORMAT('$%04X',hqPHI.mask.b_nflags),
																				 ' NOT all expected flags are in mask'")
		 }
		 ELSE IF (hqPHI.mask.b_nflags <> hqDI.hdrCommon.I_nFlags)
		 {
			AMX_LOG(AMX_WARNING,"ITOA(__LINE__),' ',sDebug,' flag mask expected ',
																					FORMAT('$%04X',hqDI.hdrCommon.I_nFlags),
																					' flag mask received ',
																					FORMAT('$%04X',hqPHI.mask.b_nflags),
																					' all expected flags OK'")
		 }
		 IF (debugHQN >= DEBUG_VERBOSE)
		 {
			AMX_LOG(AMX_WARNING,"ITOA(__LINE__),' ',sDebug,' mask ',returnHiqMessageFlags(hqPHI.mask.b_nflags)")
			IF (hqPHI.mask.b_nflags & hqDI.hdrCommon.I_nFlags)
			{
			 AMX_LOG(AMX_WARNING,"ITOA(__LINE__),' ',sDebug,
																					 ' flags common to mask...',
																					 returnHiqMessageFlags(hqPHI.mask.b_nflags & hqDI.hdrCommon.I_nFlags)")
			}
			IF (hqPHI.mask.b_nflags ^ hqDI.hdrCommon.I_nFlags)
			{
			 AMX_LOG(AMX_WARNING,"ITOA(__LINE__),' ',sDebug,
																					 ' flags NOT common to mask...',
																					 returnHiqMessageFlags(hqPHI.mask.b_nflags ^ hqDI.hdrCommon.I_nFlags)")
			}
		 }
		}
	 } // END CASE HIQ_MID_ROUTING_HELLO:
	 
//	 CASE HIQ_MID_ROUTING_GET_NETWORK_INFO:
//	 CASE HIQ_MID_ROUTING_REQUEST_ADDRESS:
//	 CASE HIQ_MID_ROUTING_ADDRESS_USED:
//	 CASE HIQ_MID_ROUTING_SET_ADDRESS:
//	 CASE HIQ_MID_ROUTING_GOODBYE:
//	 
//	 // Device Methods
//	 CASE HIQ_MID_DEVICE_GET_ATTRIBUTES:
//	 CASE HIQ_MID_DEVICE_GET_VD_LIST:
//	 CASE HIQ_MID_DEVICE_STORE:
//	 CASE HIQ_MID_DEVICE_RECALL:
//	 CASE HIQ_MID_DEVICE_LOCATE:
//
//	 // Log Methods
//	 CASE HIQ_MID_EVENT_LOG_SUBSCRIBE:
//	 CASE HIQ_MID_EVENT_LOG_UNSUBSCRIBE:
//	 CASE HIQ_MID_EVENT_LOG_REQUEST:
//
//	 // Multi Parameter Methods
//	 CASE HIQ_MID_MULTISV_SET:
//	 CASE HIQ_MID_MULTISV_SET_PERCENT:
//	 CASE HIQ_MID_MULTISV_GET:
//	 CASE HIQ_MID_MULTISV_SUBSCRIBE:
//	 CASE HIQ_MID_MULTISV_SUBSCRIBE_PERCENT:
//	 CASE HIQ_MID_MULTISV_UNSUBSCRIBE:
	 CASE HIQ_MID_DEVICE_GET_ATTRIBUTES:
	 {
																	 /***************************/
																	 /* if this is INFORMATION  */
																	 /***************************/
		IF (hqDI.hdrCommon.I_nFlags & HIQ_MID_FLAG_INFORMATION == HIQ_MID_FLAG_INFORMATION)
		{
		// how do we know this is the device manager data? Check source dev & VDO...
		 #warn'todo: filter attributes parsing'
		 parseHiQAttributesDeviceManager(hqADM, buffer, nLine, sDebug)
		}
	 }
	 DEFAULT:
	 {
		IF (debugHQN >= DEBUG_VVV)
		{
		 AMX_LOG(AMX_WARNING,"ITOA(__LINE__),' ',sDebug,' message ID ',returnHiQMessageID(hqDI.hdrCommon.H_nMessageID),' no additional parsing'")
		}
	 }
	}
	IF (debugHQN >= DEBUG_VERBOSE)
	{
	 IF (LENGTH_ARRAY(buffer))      // if there's anything left...
	 {
		AMX_LOG(AMX_WARNING,"ITOA(__LINE__),' ',sDebug,' len = ',ITOA(LENGTH_ARRAY(buffer))")
		IF (debugHQN >= DEBUG_VVV)
		{
		 printToDebug(0:1:0,"ITOA(__LINE__),' TBD (HEX) -> '",buffer,131,0,PRINT_HEX)
		 printToDebug(0:1:0,"ITOA(__LINE__),' TBD (ASC) -> '",buffer,131,0,PRINT_ASCII_HEX)
		 printToDebug(0:1:0,"ITOA(__LINE__),' TBD (DEC) -> '",buffer,131,0,PRINT_DECIMAL)
		}
	 }
	}
 }
 ELSE
 {
	AMX_LOG(AMX_ERROR,"ITOA(__LINE__),' ',sDebug,' HiQnet Protocol Version ',ITOA(buffer[1]),' is incorrect'")
 }
 ret = ret - LENGTH_ARRAY(buffer)   // ret = # bytes removed from buffer

 IF (debugHQN >= DEBUG_TERSE)
 {
	AMX_LOG(AMX_WARNING,"ITOA(nLine),' ',sDebug,
																	 ' HiQData bytes expected ',ITOA(hqDI.hdrCommon.C_lLengthMessage),
																	 ', HiQData bytes parsed ',ITOA(ret),
																	 ' ...parseHiQData() end'")
 }

 RETURN ret
}

DEFINE_FUNCTION LONG parseHiQHeaderCommon (_sHiQNet_Common_Header hqCH, CHAR buffer[], INTEGER nLine, CHAR sDebug[])
// FUNCTION: parseHiQHeaderCommon()
// PURPOSE : destructively parse HiQNet common header data from left side of buffer[],
//           setting values in hqCH using the NetLinx Marshalling Protocol
// PARAMETERS:
// RETURN: #bytes parsed here (ret)
{
 STACK_VAR LONG ret
 IF (debugHQN >= DEBUG_VVVV) { AMX_LOG(AMX_WARNING,"ITOA(nLine),' ',sDebug,' parseHiQHeaderCommon() start...'") }
 ret = LENGTH_ARRAY(buffer)
// return updated , in case of successive or recurring parseHiQ...() functions
// initial incoming data, store into _sHiQNet_Common_Header struct using STRING_TO_VARIABLE Marshalling Protocol
																		// 1
 STRING_TO_VARIABLE(hqCH.A_cVersion,"$E1,GET_BUFFER_STRING(buffer,1)",1) 
																		// 2
 STRING_TO_VARIABLE(hqCH.B_cLengthHeader,"$E1,GET_BUFFER_STRING(buffer,1)",1)
																		// 3,4,5,6
 STRING_TO_VARIABLE(hqCH.C_lLengthMessage,"$E3,GET_BUFFER_STRING(buffer,4)",1)
																		// 7,8
 STRING_TO_VARIABLE(hqCH.D_nSourceNode,"$E2,GET_BUFFER_STRING(buffer,2)",1)
 																		// 9,10,11,12
 STRING_TO_VARIABLE(hqCH.E_lSourceVDObject,"$E3,GET_BUFFER_STRING(buffer,4)",1)
 																		// 13,14
 STRING_TO_VARIABLE(hqCH.F_nDestinationNode,"$E2,GET_BUFFER_STRING(buffer,2)",1)
 																		// 15,16,17,18
 STRING_TO_VARIABLE(hqCH.G_lDestinationVDObject,"$E3,GET_BUFFER_STRING(buffer,4)",1)
 																		// 19,20
 STRING_TO_VARIABLE(hqCH.H_nMessageID,"$E2,GET_BUFFER_STRING(buffer,2)",1)
 																		// 21,22
 STRING_TO_VARIABLE(hqCH.I_nFlags,"$E2,GET_BUFFER_STRING(buffer,2)",1)
 																		// 23
 STRING_TO_VARIABLE(hqCH.J_cHop_count,"$E1,GET_BUFFER_STRING(buffer,1)",1)
 																		// 24,25
 STRING_TO_VARIABLE(hqCH.K_nSequence,"$E2,GET_BUFFER_STRING(buffer,2)",1)
 ret = ret - LENGTH_ARRAY(BUFFER)   // ret = # bytes removed from buffer
 // end common header

 IF (debugHQN >= DEBUG_VVVV)
 {
	AMX_LOG(AMX_WARNING,"ITOA(nLine),' ',sDebug,' header bytes expected ',ITOA(hqch.B_cLengthHeader),
																							', common header bytes parsed ',ITOA(ret),
																							' ...parseHiQHeaderCommon() end'")
 }
 RETURN ret
}

DEFINE_FUNCTION LONG parseHiQHeaders (_sHiQnet_headers hqDI, CHAR buffer[], INTEGER nLine, CHAR sDebug[]) 
// FUNCTION: parseHiQHeaders()
// PURPOSE : destructively parse HiQNet Header data from left side of buffer[],
//           setting values in hqDI using the NetLinx Marshalling Protocol
// PARAMETERS:
// RETURN: #bytes parsed here (ret)
{
 STACK_VAR LONG ret
 ret = LENGTH_ARRAY(buffer)
 																		/***********************/
																		/* parse Common header */
																		/***********************/
																		// expected: ret = 25
 parseHiQHeaderCommon(hqDI.hdrCommon,buffer,nLine,sDebug)

																		// check flags for other headers
																		/**********************/
																		/* parse ERROR header */
																		/**********************/
 IF (hqDI.hdrCommon.I_nFlags & HIQ_MID_FLAG_ERROR == HIQ_MID_FLAG_ERROR)
 {
	STRING_TO_VARIABLE(hqDI.hdrError.L_nCode,"$E2,GET_BUFFER_STRING(buffer,2)",1)
	STRING_TO_VARIABLE(hqDI.hdrError.M_nLength,"$E2,GET_BUFFER_STRING(buffer,2)",1)
	IF (hqDI.hdrError.M_nLength)
	{
	 hqDI.hdrError.N_utf16 = GET_BUFFER_STRING(buffer,hqDI.hdrError.M_nLength)
	 UTF16BE_TO_ASCII(hqDI.hdrError.N_utf16,hqDI.hdrError.str)
	}
 }

																		/***********************************/
																		/* parse MULTI PART MESSAGE header */
																		/***********************************/
 IF (hqDI.hdrCommon.I_nFlags & HIQ_MID_FLAG_MULTI_PART_MSG == HIQ_MID_FLAG_MULTI_PART_MSG)
 {
	STRING_TO_VARIABLE(hqDI.hdrMultiPart.O_nStart_sequence,"$E2,GET_BUFFER_STRING(buffer,2)",1)
	STRING_TO_VARIABLE(hqDI.hdrMultiPart.P_lBytes_remaining,"$E3,GET_BUFFER_STRING(buffer,4)",1)
 }
																		// deprecated header...should never see this
 IF (hqDI.hdrCommon.I_nFlags & $0080 == $0080)
 {
	STACK_VAR INTEGER n
	STRING_TO_VARIABLE(n,"$E2,GET_BUFFER_STRING(buffer,2)",1)
 }
																		/************************/
																		/* parse SESSION header */
																		/************************/
 IF (hqDI.hdrCommon.I_nFlags & HIQ_MID_FLAG_SESSION_NUMBER == HIQ_MID_FLAG_SESSION_NUMBER)
 {
																	  // store incoming session number
	STRING_TO_VARIABLE(hqDI.hdrSession.Q_nIncoming,"$E2,GET_BUFFER_STRING(buffer,2)",1)
 }
 ret = ret = LENGTH_ARRAY(buffer)   // ret = # bytes removed from buffer

																		// info
 IF (LENGTH_ARRAY(buffer))         	// if anything is left...
 {
	IF (debugHQN >= DEBUG_VERBOSE)
	{
	 AMX_LOG(AMX_WARNING,"ITOA(nLine),' ',sDebug,' headers parsed ',ITOA(LENGTH_ARRAY(buffer)),' bytes left'")
	 IF ((hqDI.hdrMultiPart.P_lBytes_remaining == LENGTH_ARRAY(buffer)) && (hqDI.hdrCommon.I_nFlags & HIQ_MID_FLAG_MULTI_PART_MSG == HIQ_MID_FLAG_MULTI_PART_MSG))
	 {
		IF (debugHQN >= DEBUG_VERBOSE)
		{
		 AMX_LOG(AMX_WARNING,"ITOA(nLine),' ',sDebug,
												 ' Multi Part Message end ',
												 ITOA(hqDI.hdrMultiPart.P_lBytes_remaining),' bytes'")
		}
	 }
	 IF (debugHQN >= DEBUG_VVV)
	 {
		printToDebug(0:1:0,"ITOA(nLine),'     bytes(HEX) '",buffer,131,0,PRINT_HEX)
		printToDebug(0:1:0,"ITOA(nLine),'     bytes(ASC) '",buffer,131,0,PRINT_ASCII_HEX)
		printToDebug(0:1:0,"ITOA(nLine),'     bytes(DEC) '",buffer,131,0,PRINT_DECIMAL)
	 }
	}
 }
 RETURN ret
}

DEFINE_FUNCTION LONG parseHiQNetworkInfo (_SHiQnet_Network_Info hqIP, CHAR buffer[], INTEGER nLine, CHAR sDebug[])
{
// FUNCTION: parseHiQNetworkInfo()
// PURPOSE : destructively parse HiQNet Network data from left side of buffer[],
//           setting values in hqIP using the NetLinx Marshalling Protocol
// PARAMETERS:
// RETURN: #bytes parsed
 STACK_VAR LONG ret
 ret = LENGTH_ARRAY(buffer)

 STRING_TO_VARIABLE(hqIP.h_lMAC_OUI,"$E3,$00,GET_BUFFER_STRING(buffer,3)",1)
 STRING_TO_VARIABLE(hqIP.i_lMAC_NIC,"$E3,$00,GET_BUFFER_STRING(buffer,3)",1)
 STRING_TO_VARIABLE(hqIP.J_cflagDHCP,"$E1,GET_BUFFER_STRING(buffer,1)",1)
 STRING_TO_VARIABLE(hqIP.k_lIPAddress,"$E3,GET_BUFFER_STRING(buffer,4)",1)
 STRING_TO_VARIABLE(hqIP.l_lSubnetMask,"$E3,GET_BUFFER_STRING(buffer,4)",1)
 STRING_TO_VARIABLE(hqIP.m_lGateway,"$E3,GET_BUFFER_STRING(buffer,4)",1)
 ret = ret - LENGTH_ARRAY(buffer)   // ret is how many bytes were removed from buffer

 RETURN ret
}

DEFINE_FUNCTION LONG parseHiQPayloadDisco (_sHiQNet_DISCO_PAYLOAD hqDP, CHAR buffer[], INTEGER nLine, CHAR sDebug[])
// FUNCTION: parseHiQPayloadDisco()
// PURPOSE : destructively parse HiQNet DISCO data from left side of buffer[],
//           setting values in hqCH using the NetLinx Marshalling Protocol
// PARAMETERS:
//         LP - start variable at 0 for initial function call.
//              pass into successive parseHiQ functions to get the total message Length Parsed
// RETURN: #bytes parsed + #bytes previously parsed (from lp parameter)
{
 STACK_VAR LONG ret
 IF (debugHQN >= DEBUG_VVVV)
 {
	AMX_LOG(AMX_WARNING,"ITOA(nLine),' ',sDebug,' parseHiQPayloadDisco() start...'")
 }
 ret = LENGTH_ARRAY(buffer)

// store into _sHiQNet_DISCO_PAYLOAD struct using STRING_TO_VARIABLE Marshalling Protocol
																		// 1,2
 STRING_TO_VARIABLE(hqDP.a_nSourceNode,"$E2,GET_BUFFER_STRING(buffer,2)",1)
 																		// 3,4
 STRING_TO_VARIABLE(hqDP.b_cCost,"$E1,GET_BUFFER_STRING(buffer,1)",1)
 																	  // 5,6 ($00,$10 expected)
 STRING_TO_VARIABLE(hqDP.c_nLengthSerialNumber,"$E2,GET_BUFFER_STRING(buffer,2)",1)
 																		// 7 - 22 or more
 hqDP.d_sSerialNumber = GET_BUFFER_STRING(buffer,hqDP.c_nLengthSerialNumber)
																		// 23,24,25,26
 STRING_TO_VARIABLE(hqDP.e_lMaxMessageSize,"$E3,GET_BUFFER_STRING(buffer,4)",1)
																		// 27,28
 STRING_TO_VARIABLE(hqDP.f_nKeepAlivePeriod,"$E2,GET_BUFFER_STRING(buffer,2)",1)
																		// 29
 STRING_TO_VARIABLE(hqDP.g_cNetworkID,"$E1,GET_BUFFER_STRING(buffer,1)",1)
																	  // 19 bytes expected to be removed via parseHiQNetworkInfo()
 parseHiQNetworkInfo (hqDP.network, buffer, nLine, sDebug)
 ret = ret - LENGTH_ARRAY(buffer)

 // end DISCO PAYLOAD
 IF (debugHQN >= DEBUG_VVVV)
 {
	AMX_LOG(AMX_WARNING,"ITOA(nLine),' ',sDebug,
																	 ' DISCO PAYLOAD bytes expected ',ITOA(2+1+2+hqDP.c_nLengthSerialNumber+4+2+1+19),
																	 ' DISCO PAYLOAD bytes parsed ',ITOA(ret),
																	 ' ...parseHiQPayloadDisco() end'")
 }

 RETURN ret
}

DEFINE_FUNCTION printHiQHeaderCommon (_sHiQNet_Common_Header hqCH, INTEGER nLine, CHAR sDebug[])
{
 // Assume: The HiQNet network will present complete messages in each event,
 //         buffering will not be necessary...
 // Assume: first byte should be $02, as a 1 byte string
 IF (debugHQN >= DEBUG_VERBOSE)
 {
 	IF (debugHQN >= DEBUG_VVVV) { AMX_LOG(AMX_WARNING,"ITOA(nLine),' ',sDebug,' Common header'") }
	AMX_LOG(AMX_WARNING,"'HQn version ',FORMAT('$%02X',hqCH.A_cVersion),', ',
											 'length header ',FORMAT('$%02X',hqCH.B_cLengthHeader),' (',ITOA(hqCH.B_cLengthHeader),')',', ',
											 'length message ',FORMAT('$%08X',hqCH.C_lLengthMessage),' (',ITOA(hqCH.C_lLengthMessage),')'")
	AMX_LOG(AMX_WARNING,"'Source      HiQ Node ',FORMAT('$%04X',hqCH.D_nSourceNode),' (',FORMAT('%5d',hqCH.D_nSourceNode),')',', ',
											 'VDO ',FORMAT('$%08X',hqCH.E_lSourceVDObject),' (',ITOA(hqCH.E_lSourceVDObject),')'")
	AMX_LOG(AMX_WARNING,"'Destination HiQ Node ',FORMAT('$%04X',hqCH.F_nDestinationNode),' (',FORMAT('%5d',hqCH.F_nDestinationNode),')',', ',
											 'VDO ',FORMAT('$%08X',hqCH.G_lDestinationVDObject),' (',ITOA(hqCH.G_lDestinationVDObject),')'")
	AMX_LOG(AMX_WARNING,"'Message ID ',returnHiQMessageID(hqCH.H_nMessageID)")
	AMX_LOG(AMX_WARNING,"'flags ',returnHiqMessageFlags(hqCH.I_nFlags)")
	AMX_LOG(AMX_WARNING,"'hop count ',FORMAT('$%02X',hqCH.J_cHop_count),' (',ITOA(hqCH.J_cHop_count),')',', ',
											 'sequence ',FORMAT('$%04X',hqCH.K_nSequence),' (',ITOA(hqCH.K_nSequence),')'")
 }
 ELSE IF (debugHQN >= DEBUG_TERSE)
 {
	AMX_LOG(AMX_WARNING,"ITOA(nLine),' ',sDebug,' Common header'")
	AMX_LOG(AMX_WARNING,"FORMAT('$%02X',hqCH.A_cVersion),', ',         // 1
	                     FORMAT('$%02X',hqCH.B_cLengthHeader),', ',    // 2
	                     FORMAT('$%08X',hqCH.C_lLengthMessage)")       // 3,4,5,6
	AMX_LOG(AMX_WARNING,"FORMAT('$%04X',hqCH.D_nSourceNode),', ',      // 7,8
											 FORMAT('$%08X',hqCH.E_lSourceVDObject)")      // 9,10,11,12
	AMX_LOG(AMX_WARNING,"FORMAT('$%04X',hqCH.F_nDestinationNode),', ', // 13,14
											 FORMAT('$%08X',hqCH.G_lDestinationVDObject)") // 15,16,17,18
	AMX_LOG(AMX_WARNING,"returnHiQMessageID(hqCH.H_nMessageID)")       // 19,20
	AMX_LOG(AMX_WARNING,"returnHiqMessageFlags(hqCH.I_nFlags)")        // 21,22
	AMX_LOG(AMX_WARNING,"FORMAT('$%02X',hqCH.J_cHop_count),', ',       // 23
											 FORMAT('$%04X',hqCH.K_nSequence)")            // 24,25
 }
}

DEFINE_FUNCTION printHiQHeaderError (_sHiQnet_Error_Header hqE, INTEGER nLine, CHAR sDebug[])
{
 IF (debugHQN >= DEBUG_VVV)         // print the error code
 {
	AMX_LOG(AMX_ERROR,"ITOA(__LINE__),' ',sDebug,' ERROR header ',
										 'CODE ',returnHiQControlNetworkEventID(hqE.L_nCode),', ',
										 FORMAT('LENGTH $%04X ',hqE.M_nLength),
										 FORMAT('(%d)',hqE.M_nLength)")
	IF (hqE.M_nLength)
	{
	 AMX_LOG(AMX_ERROR,"ITOA(__LINE__),
											 ' ERROR ',
											 FORMAT(' (trimmed %d) ',LENGTH_ARRAY(hqE.str)),
											 '''',hqE.str,''''")
	}
 }
 ELSE IF (debugHQN >= DEBUG_TERSE)  // print the error code
 {
	AMX_LOG(AMX_ERROR,"ITOA(__LINE__),' ',sDebug,' ERROR ',
										 'CODE ',returnHiQControlNetworkEventID(hqE.L_nCode),' ',
										 '''',hqE.str,''''")
 }
}

DEFINE_FUNCTION printHiQHeaderMultiPart (_sHiQnet_MultiPart_Header hqMPH, INTEGER nLine, CHAR sDebug[])
{
 IF (debugHQN >= DEBUG_VERBOSE)
 {
	IF (debugHQN >= DEBUG_VVVV) { AMX_LOG(AMX_WARNING,"ITOA(nLine),' ',sDebug,' Multi Part header'") }
	AMX_LOG(AMX_WARNING,"FORMAT('Multi Part Start $%04X',hqMPH.O_nStart_sequence),
											' (',ITOA(hqMPH.O_nStart_sequence),')'")
	AMX_LOG(AMX_WARNING,"FORMAT('Multi Part Bytes Left $%08X',hqMPH.P_lBytes_remaining),
											' (',ITOA(hqMPH.P_lBytes_remaining),')'")
 }
 ELSE IF (debugHQN >= DEBUG_TERSE)
 {
 	AMX_LOG(AMX_WARNING,"ITOA(nLine),' ',sDebug,' Multi Part header'")
	AMX_LOG(AMX_WARNING,"FORMAT('$%04X',hqMPH.O_nStart_sequence)")
	AMX_LOG(AMX_WARNING,"FORMAT('$%08X',hqMPH.P_lBytes_remaining)")
 }
}

DEFINE_FUNCTION printHiQHeaders (_sHiQnet_headers hqD, INTEGER hdrSession, INTEGER nLine, CHAR sDebug[])
{
 printHiQHeaderCommon(hqD.hdrCommon,nLine,sDebug)
 IF (hqD.hdrCommon.I_nFlags & HIQ_MID_FLAG_ERROR == HIQ_MID_FLAG_ERROR)
 {
	printHiQHeaderError(hqD.hdrError,nLine,sDebug)
 }
 IF (hqD.hdrCommon.I_nFlags & HIQ_MID_FLAG_MULTI_PART_MSG == HIQ_MID_FLAG_MULTI_PART_MSG)
 {
	printHiQHeaderMultiPart(hqD.hdrMultiPart,nline,sDebug)
 }
 IF (hqD.hdrCommon.I_nFlags & HIQ_MID_FLAG_SESSION_NUMBER == HIQ_MID_FLAG_SESSION_NUMBER)
 {
	IF (debugHQN >= DEBUG_VERBOSE)
	{
	 AMX_LOG(AMX_WARNING,"ITOA(nLine),' ',sDebug,' Session header'")
	 AMX_LOG(AMX_WARNING,"FORMAT('session $%04X',hdrSession)")
	}
	ELSE IF (debugHQN >= DEBUG_TERSE)
	{
	 AMX_LOG(AMX_WARNING,"ITOA(nLine),' ',sDebug,' Session header'")
	 AMX_LOG(AMX_WARNING,"FORMAT('$%04X',hdrSession)")
	}
 }
}

DEFINE_FUNCTION printHiQPayloadDisco (_sHiQNet_DISCO_PAYLOAD hqPD, INTEGER nLine, CHAR sDebug[])
{
 //printHiQPayloadDisco(hqPD.common,nline)
// INTEGER a_nSourceNode                 // HiQnet address of sender == common.D_nSourceNode
// CHAR    b_cCost                       // HIQ_ROUTING_COST_ETHERNET for NetLinx TCP connection. If proxies are in play(?): "Aggregated cost of route back to src".
// CHAR    d_sSerialNumber[16]           // 10x 0's and the MAC address, if the device has a NIC
// LONG    e_lMaxMessageSize             // Max Msg size sender can handle
// INTEGER F_nKeepAliveTime              // Keep Alive rate in ms
// CHAR    g_cNetworkID                  // HIQ_ROUTING_NID_ETHERNET for Netlinx TCP connection
// _SHiQnet_Network_Info hqPD.network // MAC address, DHCP flag, IP address/subnet/gateway
 IF (debugHQN >= DEBUG_VERBOSE)
 {
	IF (debugHQN >= DEBUG_VVV) { AMX_LOG(AMX_WARNING,"ITOA(nLine),' ',sDebug,' DISCO payload'") }
	AMX_LOG(AMX_WARNING,"'Source HiQ Node ',FORMAT('$%04X',hqPD.a_nSourceNode),' (',ITOA(hqPD.a_nSourceNode),')',', ',
											 'Cost ',FORMAT('$%02X',hqPD.b_cCost),', ',
											'Length SN ',FORMAT('$%04X',hqPD.c_nLengthSerialNumber),' (',ITOA(hqPD.c_nLengthSerialNumber),')'")
	printToDebug(0:1:0,'SN ',hqPD.d_sSerialNumber,131,0,PRINT_HEX)

	AMX_LOG(AMX_WARNING,"'max message size ',FORMAT('$%08X',hqPD.e_lMaxMessageSize),' (',ITOA(hqPD.e_lMaxMessageSize),')',', ',
											 'keep alive period ',FORMAT('$%04X',hqPD.f_nKeepAlivePeriod),' (',ITOA(hqPD.f_nKeepAlivePeriod),')',', ',
											 'network ID ',returnHiQNetworkID(hqPD.g_cNetworkID)")

	printToDebug(0:1:0,'MAC address ',"
										 RIGHT_STRING(RAW_BE(hqPD.network.h_lMAC_OUI),3),
										 RIGHT_STRING(RAW_BE(hqPD.network.i_lMAC_NIC),3)",
										 131,0,PRINT_HEX)

	AMX_LOG(AMX_WARNING,FORMAT('DHCP $%02X',hqPD.network.J_cflagDHCP))
	printToDebug(0:1:0,'IP address  ',RAW_BE(hqPD.network.k_lIPAddress),131,0,PRINT_DECIMAL)
	printToDebug(0:1:0,'Subnet Mask ',RAW_BE(hqPD.network.l_lSubnetMask),131,0,PRINT_DECIMAL)
	printToDebug(0:1:0,'Gateway     ',RAW_BE(hqPD.network.m_lGateway),131,0,PRINT_DECIMAL)
 }
 ELSE IF (debugHQN >= DEBUG_TERSE)
 {
	AMX_LOG(AMX_WARNING,"ITOA(nLine),' ',sDebug,' DISCO payload'")
	AMX_LOG(AMX_WARNING,"FORMAT('$%04X',hqPD.a_nSourceNode),', ',
											 FORMAT('$%02X',hqPD.b_cCost),', ',
											 FORMAT('$%04X',hqPD.c_nLengthSerialNumber)")
	printToDebug(0:1:0,'',hqPD.d_sSerialNumber,131,0,PRINT_HEX)

	AMX_LOG(AMX_WARNING,"FORMAT('$%08X',hqPD.e_lMaxMessageSize),', ',
											 FORMAT('$%04X',hqPD.f_nKeepAlivePeriod),', ',
											 returnHiQNetworkID(hqPD.g_cNetworkID)")

	printToDebug(0:1:0,'',"
										 RIGHT_STRING(RAW_BE(hqPD.network.h_lMAC_OUI),3),
										 RIGHT_STRING(RAW_BE(hqPD.network.i_lMAC_NIC),3)",
										 131,0,PRINT_HEX)

	AMX_LOG(AMX_WARNING,FORMAT('$%02X',hqPD.network.J_cflagDHCP))
	printToDebug(0:1:0,'',RAW_BE(hqPD.network.k_lIPAddress),131,0,PRINT_DECIMAL)
	printToDebug(0:1:0,'',RAW_BE(hqPD.network.l_lSubnetMask),131,0,PRINT_DECIMAL)
	printToDebug(0:1:0,'',RAW_BE(hqPD.network.m_lGateway),131,0,PRINT_DECIMAL)
 }
}

DEFINE_FUNCTION CHAR[48] returnHiQControlNetworkEventID (INTEGER nID)
{
 STACK_VAR CHAR s[43]
 SWITCH (nID)
 {
	CASE HIQ_ERROR_BAD_VERSION:   { s = ' INVALID VERSION' }
	CASE HIQ_ERROR_BAD_LENGTH:    { s = ' INVALID LENGTH' }
	CASE HIQ_ERROR_BAD_VIRTUAL:   { s = ' INVALID VIRTUAL DEVICE' }
	CASE HIQ_ERROR_BAD_OBJECT:    { s = ' INVALID OBJECT' }
	CASE HIQ_ERROR_BAD_SV:        { s = ' INVALID PARAMETER' }
	CASE HIQ_ERROR_BAD_MSG_ID:    { s = ' INVALID MESSAGE ID' }
	CASE HIQ_ERROR_BAD_VALUE:     { s = ' INVALID VALUE' }
	CASE HIQ_ERROR_UNAVAILABLE:   { s = ' RESOURCE UNAVAILABLE' }
	CASE HIQ_ERROR_MSG_NA:        { s = ' UNSUPPORTED' }
	CASE HIQ_ERROR_BAD_VD_CLASS:  { s = ' INVALID VIRTUAL DEVICE CLASS' }
	CASE HIQ_ERROR_BAD_OBJ_CLASS: { s = ' INVALID OBJECT CLASS' }
	CASE HIQ_ERROR_BAD_SV_CLASS:  { s = ' INVALID PARAMETER CLASS' }
	CASE HIQ_ERROR_BAD_ATTRIBUTE: { s = ' INVALID ATTRIBUTE ID' }
	CASE HIQ_ERROR_BAD_DATA_TYPE: { s = ' INVALID DATA TYPE' }
	CASE HIQ_ERROR_BAD_CONFIG:    { s = ' INVALID CONFIGURATION' }
	CASE HIQ_ERROR_FLASH:         { s = ' FLASH ERROR' }
	CASE HIQ_ERROR_NOT_ROUTER:    { s = ' NOT A ROUTER' }
 }
 RETURN "FORMAT('$%04X',nID),s"
}

DEFINE_FUNCTION CHAR[32] returnHiQDataType (CHAR cID)
{
 STACK_VAR CHAR s[27]
 SWITCH (cID)
 {
	CASE HIQ_BYTE:      { s = ' SBYTE' }
	CASE HIQ_UBYTE:     { s = ' CHAR' }
	CASE HIQ_WORD:      { s = ' SINTEGER' }
	CASE HIQ_UWORD:     { s = ' INTEGER' }
	CASE HIQ_LONG:      { s = ' SLONG' }
	CASE HIQ_ULONG:     { s = ' LONG' }
	CASE HIQ_FLOAT32:   { s = ' FLOAT' }
	CASE HIQ_FLOAT64:   { s = ' DOUBLE' }
	CASE HIQ_BLOCK:     { s = ' STRING' }
	CASE HIQ_STRING:    { s = ' UTF16BE' }
	CASE HIQ_LONG64:    { s = ' LONG64' }
	CASE HIQ_ULONG64:   { s = ' ULONG64' }
	CASE HIQ_BIG_BLOCK: { s = ' LSTRING' }
 }
 RETURN "FORMAT('$%02X',cID),s"
}

DEFINE_FUNCTION CHAR[254] returnHiQMessageFlags (INTEGER I_nFlags)
{
 STACK_VAR CHAR s[250]
 
 IF (I_nFlags & $8000)                       { s = "s,FORMAT(' $%04X |',$8000)" }
 IF (I_nFlags & $4000)                       { s = "s,FORMAT(' $%04X |',$4000)" }
 IF (I_nFlags & $2000)                       { s = "s,FORMAT(' $%04X |',$2000)" }
 IF (I_nFlags & $1000)                       { s = "s,FORMAT(' $%04X |',$1000)" }
 IF (I_nFlags & $0800)                       { s = "s,FORMAT(' $%04X |',$0800)" }
 IF (I_nFlags & $0400)                       { s = "s,FORMAT(' $%04X |',$0400)" }
 IF (I_nFlags & $0200)                       { s = "s,FORMAT(' $%04X |',$0200)" }
 IF (I_nFlags & HIQ_MID_FLAG_SESSION_NUMBER) { s = "s,FORMAT(' SES. NUM $%04X |',HIQ_MID_FLAG_SESSION_NUMBER)" }
 IF (I_nFlags & $0080)                       { s = "s,FORMAT(' $%04X |',$0080)" }
 IF (I_nFlags & HIQ_MID_FLAG_MULTI_PART_MSG) { s = "s,FORMAT(' MULTI PART MSG $%04X |',HIQ_MID_FLAG_MULTI_PART_MSG)" }
 IF (I_nFlags & HIQ_MID_FLAG_GUARANTEED)     { s = "s,FORMAT(' GUARANTEED $%04X |',HIQ_MID_FLAG_GUARANTEED)" }
 IF (I_nFlags & $0010)                       { s = "s,FORMAT(' $%04X |',$0010)" }
 IF (I_nFlags & HIQ_MID_FLAG_ERROR)          { s = "s,FORMAT(' ERROR $%04X |',HIQ_MID_FLAG_ERROR)" }
 IF (I_nFlags & HIQ_MID_FLAG_INFORMATION)    { s = "s,FORMAT(' INFO $%04X |',HIQ_MID_FLAG_INFORMATION)" }
 IF (I_nFlags & HIQ_MID_FLAG_ACK)            { s = "s,FORMAT(' ACK $%04X |',HIQ_MID_FLAG_ACK)" }
 IF (I_nFlags & HIQ_MID_FLAG_REQUEST_ACK)    { s = "s,FORMAT(' REQ ACK $%04X |',HIQ_MID_FLAG_REQUEST_ACK)" }
																		// trim trailing ' |'
 IF (LENGTH_ARRAY(s) > 2) { SET_LENGTH_ARRAY(s,LENGTH_ARRAY(s) - 2) }

 RETURN "FORMAT('$%04X',I_nFlags),s"
}

DEFINE_FUNCTION CHAR[48] returnHiQMessageID (INTEGER nID)
{
 STACK_VAR CHAR s[43]
 SWITCH (nID)
 {
	CASE HIQ_MID_ROUTING_DISCO_INFO:       { s = ' DISCO INFO' }
	CASE HIQ_MID_ROUTING_GET_NETWORK_INFO: { s = ' GET NETWORK INFO' }
	CASE HIQ_MID_ROUTING_REQUEST_ADDRESS:  { s = ' REQUEST ADDRESS' }
	CASE HIQ_MID_ROUTING_ADDRESS_USED:     { s = ' ADDRESS USED' }
	CASE HIQ_MID_ROUTING_SET_ADDRESS:      { s = ' SET ADDRESS' }
	CASE HIQ_MID_ROUTING_GOODBYE:          { s = ' GOODBYE' }
	CASE HIQ_MID_ROUTING_HELLO:            { s = ' HELLO' }

	CASE HIQ_MID_DEVICE_GET_ATTRIBUTES:    { s = ' GET ATTRIBUTES' }
	CASE HIQ_MID_MULTISV_GET:                { s = ' MULTI-PARAMETER GET' }
 }
 RETURN "FORMAT('$%04X',nID),s"
}

DEFINE_FUNCTION CHAR[32] returnHiQNetworkID (CHAR cID)
{
 STACK_VAR CHAR s[29]
 SWITCH (cID)
 {
	CASE HIQ_ROUTING_NID_ETHERNET: { s = ' ETHERNET' }
	CASE HIQ_ROUTING_NID_RS232:    { s = ' RS232' }
	CASE HIQ_ROUTING_NID_USB_P2P:  { s = ' USB POINT TO POINT' }
 }
 
 RETURN "FORMAT('$%02X',cID),s"
}

DEFINE_FUNCTION setHiQHeaderCommon_NX_default (_sHiQNet_Common_Header hqD)
{
  // setup non-character data
  hqD.A_cVersion             = $02       // Always 2
  hqD.B_cLengthHeader        = $00       // TBD, $19 (25) for simple header, $1B (27) for header with Session Number, etc.
  hqD.C_lLengthMessage       = $00000000 // TBD, depending on message

	IF (!hqD.D_nSourceNode) { hqD.D_nSourceNode = nodeSelf }
	hqD.E_lSourceVDObject      = $00000000

  hqD.F_nDestinationNode     = $0000     // TBD, always
  hqD.G_lDestinationVDObject = $00000000 // TBD, depending on message

  hqD.H_nMessageID           = $0000     // TBD, always
  hqD.I_nFlags               = $0000     // TBD, always

  hqD.J_cHop_count           = $05       // Usually 5
  hqD.K_nSequence            = $0000     // Increment by one, each message
}

DEFINE_FUNCTION setHiQPayloadDisco_NX_default (CHAR sSN[], _SHiQnet_Network_Info hqNI, _sHiQNet_DISCO_PAYLOAD hqPD, INTEGER nLine)
// get ip
//Get IP Address
//--------------
// ...
//        Type        DHCP
//        IPv4 Address  192.168.1.74
//        IPv4 Subnet Mask 255.255.255.0
//        IPv4 Gateway IP  192.168.1.254
// ...
//        MAC Address 00:60:9f:98:bd:f0
{
 hqPD.a_nSourceNode = nodeSelf
 hqPD.b_cCost = HIQ_ROUTING_COST_ETHERNET
 hqPD.c_nLengthSerialNumber = LENGTH_ARRAY(sSN) // usually 10x 0's then the MAC address
 hqPD.d_sSerialNumber = sSN
 hqPD.e_lMaxMessageSize = 1988   // 1988 due to the ~2000 byte ICSP packet size. Strings can be much larger, but will be split into multiple events.
																 // sometimes they are split anyways, by the HiQnet device...
 hqPD.f_nKeepAlivePeriod = 10000 // 10s in ms is the HiQnet default
 hqPD.g_cNetworkID = HIQ_ROUTING_NID_ETHERNET

 hqPD.network.h_lMAC_OUI = hqNI.h_lMAC_OUI
 hqPD.network.i_lMAC_NIC = hqNI.i_lMAC_NIC
 hqPD.network.J_cflagDHCP = hqNI.J_cflagDHCP
 hqPD.network.k_lIPAddress = hqNI.k_lIPAddress
 hqPD.network.l_lSubnetMask = hqNI.l_lSubnetMask
 hqPD.network.m_lGateway = hqNI.m_lGateway
 IF (debugHQN >= DEBUG_TERSE)
 {
	SEND_STRING 0,"ITOA(nLine),' hqPD_DiscoBC is set, node ',FORMAT('$%04X',nodeSelf),' (',ITOA(nodeSelf),')'"
 }
}

DEFINE_FUNCTION setHqData_Default(CHAR sMAC[])
// assumption: ipHiQNet array has already been setup with default IP addresses, if you know them already
{
 STACK_VAR _sHiQnet_Network_Info hqNetInfo
 STACK_VAR CHAR octets[6][3]
 STACK_VAR CHAR sSerialNumber[16]
 
 STACK_VAR INTEGER i
 STACK_VAR INTEGER nTmp
 STACK_VAR INTEGER lenIPHQn

 // Convert NetLinx ASCII MAC info into two LONG values for HiQnet data
 // expected: i == 6
 i = stringTokenizer(sMAC,':',octets)
 STRING_TO_VARIABLE(hqNetInfo.h_lMAC_OUI,"$E3,$00,HEXTOI(octets[1]),HEXTOI(octets[2]),HEXTOI(octets[3])",1)
 STRING_TO_VARIABLE(hqNetInfo.i_lMAC_NIC,"$E3,$00,HEXTOI(octets[4]),HEXTOI(octets[5]),HEXTOI(octets[6])",1)
																		// SN = 10x 0's then MAC address
 sSerialNumber = "0,0,0,0,0,0,0,0,0,0,
									RIGHT_STRING(RAW_BE(hqNetInfo.h_lMAC_OUI),3),
									RIGHT_STRING(RAW_BE(hqNetInfo.i_lMAC_NIC),3)"

 hqNetInfo.J_cflagDHCP = ipNetLinx.FLAGS & $0001
 // Convert NetLinx ASCII IP info into a LONG value for HiQnet data
 // expected: i == 4, each time stringTokenizer() is called...
 i = stringTokenizer(ipNetLinx.IPADDRESS,'.',octets)
 STRING_TO_VARIABLE(hqNetInfo.k_lIPAddress,"$E3,ATOI(octets[1]),ATOI(octets[2]),ATOI(octets[3]),ATOI(octets[4])",1)
 i = stringTokenizer(ipNetLinx.SUBNETMASK,'.',octets)
 STRING_TO_VARIABLE(hqNetInfo.l_lSubnetMask,"$E3,ATOI(octets[1]),ATOI(octets[2]),ATOI(octets[3]),ATOI(octets[4])",1)
 i = stringTokenizer(ipNetLinx.GATEWAY,'.',octets)
 STRING_TO_VARIABLE(hqNetInfo.m_lGateway,"$E3,ATOI(octets[1]),ATOI(octets[2]),ATOI(octets[3]),ATOI(octets[4])",1)

 lenIPHQn = LENGTH_ARRAY(ipHiQNet)
 FOR (i = MAX_LENGTH_ARRAY(ipHiQNet); i; i--)
 {
	// set common values first...
	setHiQHeaderCommon_NX_default(hqDataOut[i].hdrCommon)

	// add HiQNet addresses we want to communicate with
	IF (i <= lenIPHQn)
	{
	 hqDataOut[i].hdrCommon.F_nDestinationNode = ATOI(ipHiQNet[i].node)
	}

	// After bootup...generate a random session number between 1 and 65535
	// Adding one for each HiQnet device we may talk to, to reduce the chance of duplication after reboot
	// For each new session...increment the session number
	nTmp = RANDOM_NUMBER(65535)    // Generate new session # at bootup, between 1 and 65535
	IF (hqPH_Out[i].session.a_nSelf = nTmp + 1)     // if session # will be same as the last one change it
	{
	 nTmp ++
	}
	hqPH_Out[i].session.a_nSelf = nTmp
	SEND_STRING 0,"ITOA(__LINE__),' set unique session # ',FORMAT('$%04X',hqPH_Out[i].session.a_nSelf),' for ',ipHiQNet[i].hostname,'...'"
 }
 SET_LENGTH_ARRAY(hqDataIn,lenIPHQn)
 SET_LENGTH_ARRAY(hqDataOut,lenIPHQn)
 SET_LENGTH_ARRAY(hqPD_DvMgr,lenIPHQn)

 // setup self Disco Payload (this could be an array, if you have NX multiple masters, Max Message Sizes, etc...)
 //setHiQPayloadDisco_NX_default(sSerialNumber,hqNetInfo,hqPD_DvMgr[ndxSelf],__LINE__)

 // setup the hqPD_DiscoBC array...start with the master
 setHiQPayloadDisco_NX_default(sSerialNumber,hqNetInfo,hqPD_DiscoBC[ndxSelf],__LINE__)
 SET_LENGTH_ARRAY(hqPD_DiscoBC,1)
 // add the pre-defined HiQnet devices
 FOR (i = LENGTH_ARRAY(hqPD_DiscoBC) + 1; i <= lenIPHQn; i++)
 {
	hqPD_DiscoBC[i] = hqPD_DvMgr[i]
 }
 SET_LENGTH_ARRAY(hqPD_DiscoBC,i - 1)

}

DEFINE_FUNCTION setIpHiQNet()       // set known HiQNet addresses and IP addresses here
{
 STACK_VAR INTEGER i
 STACK_VAR CHAR octets[4][3]
 GET_IP_ADDRESS(0:1:0,ipNetLinx)
 // ndxSelf = 1
 ipHiQNet[1].hostname = ipNetLinx.HOSTNAME
 ipHiQNet[1].node = ITOA(nodeSelf)  // $0033 (51)
 ipHiQNet[1].dvPort = dvSelf.PORT
 ipHiQNet[1].ipAddress = ipNetLinx.IPADDRESS

 ipHiQNet[2].hostname = 'Guy-CDi4x600BL' 
 ipHiQNet[2].node = '514'           // $0202
 ipHiQNet[2].dvPort = dvHQN[2].PORT
 ipHiQNet[2].ipAddress = '192.168.1.98'

 ipHiQNet[3].hostname = 'Guy-BLU-806DA'
 ipHiQNet[3].node = '46409'         // $B549
 ipHiQNet[3].dvPort = dvHQN[3].PORT
 ipHiQNet[3].ipAddress = '192.168.1.85'

 ipHiQNet[4].hostname = 'Guy-DCi8x300N'
 ipHiQNet[4].node = '2065'          // $0811
 ipHiQNet[4].dvPort = dvHQN[4].PORT
 ipHiQNet[4].ipAddress = '192.168.1.111'

 ipHiQNet[5].hostname = 'Guy-DCi8x600DA'
 ipHiQNet[5].node = '3089'          // $0C11
 ipHiQNet[5].dvPort = dvHQN[5].PORT
 ipHiQNet[5].ipAddress = '192.168.1.84'

 ipHiQNet[6].hostname = 'Guy-DCi4x2400N'
 ipHiQNet[6].node = '2305'          // $0901
 ipHiQNet[6].dvPort = dvHQN[6].PORT
 ipHiQNet[6].ipAddress = '192.168.1.88'

 ipHiQNet[7].hostname = 'Guy-4x3500HD'
 ipHiQNet[7].node = '2'             // $0002
 ipHiQNet[7].dvPort = dvHQN[7].PORT
 ipHiQNet[7].ipAddress = '192.168.1.81'
 
 ipHiQNet[8].hostname = 'Guy-MA4-D'
 ipHiQNet[8].node = '4389'          // $1125
 ipHiQNet[8].dvPort = dvHQN[8].PORT
 ipHiQNet[8].ipAddress = '192.168.1.104'

 SET_LENGTH_ARRAY(ipHiQNet,8)
 
 hqAtr_DM[ndxSelf].classname.str = DEVICE_ID_STRING(0:1:0)
 hqAtr_DM[ndxSelf].classname.utf16 = ASCII_TO_UTF16BE(ipHiQNet[ndxSelf].hostname)
 
 FOR (i = LENGTH_ARRAY(ipHiQNet); i; i--)
 {
	hqAtr_DM[i].name.str = ipHiQNet[i].hostname
	hqAtr_DM[i].name.utf16 = ASCII_TO_UTF16BE(ipHiQNet[i].hostname)
	
	hqPD_DvMgr[i].a_nSourceNode = ATOI(ipHiQNet[i].node)
	stringTokenizer(ipHiQNet[i].ipAddress,'.',octets)
	STRING_TO_VARIABLE(hqPD_DvMgr[i].network.k_lIPAddress,"$E3,ATOI(octets[1]),ATOI(octets[2]),ATOI(octets[3]),ATOI(octets[4])",1)
 }
 
 FOR (i = LENGTH_ARRAY(ipHiQNet) + 1; i <= LENGTH_ARRAY(dvHQN); i++)
 {
	ipHiQNet[i].dvPort = dvHQN[i].PORT
 }
 
 ipHiQDiscoBC.hostname  = 'HQN BROADCAST'
 ipHiQDiscoBC.node      = '65535'
 ipHiQDiscoBC.dvPort    = dvHQnDiscoBC.PORT
 ipHiQDiscoBC.ipAddress = '255.255.255.255'
}

DEFINE_FUNCTION CHAR[15] LONG_TO_IPV4(LONG lIP)
{
 STACK_VAR CHAR rawIP[6]
 VARIABLE_TO_STRING(lIP,rawIP,1)
 // skip the NetLinx Marshalling byte at rawIP[1]
 RETURN "ITOA(rawIP[2]),'.',ITOA(rawIP[3]),'.',ITOA(rawIP[4]),'.',ITOA(rawIP[5])"
}

DEFINE_FUNCTION INTEGER REMOVE_MARSHALLING_BYTES(char str[])
/* Testing found that concatenating "RAW_BE(),RAW_BE()..)" in Netlinx code was */
/* faster than VARIABLE_TO_STRING() followed by REMOVE_MARSHALLING_BYTES().    */
/* In other words, this function was used less often than it could have been,  */
/* but it can be useful for quickly written code using VARIABLE_TO_STRING().   */
/* This function does show a use case for REMOVE_STRING() start byte > 1.      */
// FUNCTION: REMOVE_MARSHALLING_BYTES()
// PURPOSE:  Destructively parse the NetLinx Marshaling bytes from str[].
//           E.g. used in building a HiQNet command string from variables
//           containing HiQnet data.
//           Structs must be defined in protocol order, e.g. HiQnet .version, .header length, .message length, etc.
//           After using VARIABLE_TO_STRING(<struct>,<string>,pos), run it through
//           this subroutine to leave only the data.
// RETURN:  new length of str[]
// for now, this should work with 1,2,4 & 8 byte data types, CHAR arrays, and these types in a struct.
{
 STACK_VAR INTEGER i
 STACK_VAR INTEGER s                // struct counter
 STACK_VAR INTEGER n                // for length of arrays
 STACK_VAR CHAR sTmp[16]

 // simple data types will start with $E1 - $E4
 i = 1
 s = 0
//printToDebug(0:1:0,"ITOA(__LINE__),' len = ',ITOA(LENGTH_ARRAY(str)),' i = ',ITOA(i),' '",str,131,0,PRINT_HEX)
 WHILE (i <= LENGTH_ARRAY(str))
 {
	SWITCH (STR[i])
	{
	 CASE $E1: // one byte data type: char
	 {
		REMOVE_STRING(STR,"$E1",i)
		i++
//printToDebug(0:1:0,"ITOA(__LINE__),' len = ',ITOA(LENGTH_ARRAY(str)),' i = ',ITOA(i),' '",str,131,0,PRINT_HEX)
	 }
	 CASE $E2: // two byte data types: int, sint
	 {
		REMOVE_STRING(STR,"$E2",i)
		i = i+2
//printToDebug(0:1:0,"ITOA(__LINE__),' len = ',ITOA(LENGTH_ARRAY(str)),' i = ',ITOA(i),' '",str,131,0,PRINT_HEX)
	 }
	 CASE $E3: // four byte data types: long, slong, float
	 {
		REMOVE_STRING(STR,"$E3",i)
		i = i+4
//printToDebug(0:1:0,"ITOA(__LINE__),' len = ',ITOA(LENGTH_ARRAY(str)),' i = ',ITOA(i),' '",str,131,0,PRINT_HEX)
	 }
	 CASE $E4: // eight byte data types: double
	 {
		REMOVE_STRING(STR,"$E4",i)
		i = i+4
//printToDebug(0:1:0,"ITOA(__LINE__),' len = ',ITOA(LENGTH_ARRAY(str)),' i = ',ITOA(i),' '",str,131,0,PRINT_HEX)
	 }
	 CASE $E5: // CHAR array not longer than 64K
	 {
		sTmp = MID_STRING(STR,FIND_STRING(STR,"$E5",i)+1,2)
		REMOVE_STRING(STR,"$E5,sTmp",i)
		// 2 bytes after $E5 are the length_array(char array) 1 to 65535
    STRING_TO_VARIABLE(n,"$E2,sTmp",1)
		i = i+n
//printToDebug(0:1:0,"ITOA(__LINE__),' len = ',ITOA(LENGTH_ARRAY(str)),' i = ',ITOA(i),' '",str,131,0,PRINT_HEX)
	 }
	 CASE $EA: // struct start
	 {
		REMOVE_STRING(STR,"$EA",i)
		s++
//printToDebug(0:1:0,"ITOA(__LINE__),' len = ',ITOA(LENGTH_ARRAY(str)),' i = ',ITOA(i),' '",str,131,0,PRINT_HEX)
	 }
	 CASE $EB: // struct end
	 {
		REMOVE_STRING(STR,"$EB",i)
		s--
//printToDebug(0:1:0,"ITOA(__LINE__),' len = ',ITOA(LENGTH_ARRAY(str)),' i = ',ITOA(i),' '",str,131,0,PRINT_HEX)
		IF (s & $8000) // if s rolled under
		{
		 AMX_LOG(AMX_ERROR,"ITOA(__LINE__),' ???unmatched $',ITOHEX(STR[i]),' struct end'")
		}
	 }
	 //#warn 'todo: add array and "skip data" types'
	 DEFAULT:
	 {
		// skip over unhandled bytes, don't delete them...in case you have a mix of RAW_BE() and VARIABLE_TO_STRING()
		i++                             
//printToDebug(0:1:0,"ITOA(__LINE__),' len = ',ITOA(LENGTH_ARRAY(str)),' i = ',ITOA(i),' '",str,131,0,PRINT_HEX)
	 }
	}
 }
//printToDebug(0:1:0,"ITOA(__LINE__),' len = ',ITOA(LENGTH_ARRAY(str)),' i = ',ITOA(i),' '",str,131,0,PRINT_HEX)
 RETURN LENGTH_ARRAY(str)
}



DEFINE_FUNCTION INTEGER UTF16BE_TO_ASCII(CHAR utf16[], CHAR ch[])
// hack, to quickly get the Unicode to print as ASCII in a terminal, instead of e.g. writing the Unicode to a file then reading it there...
// assumption 1: there are an even number of bytes in utf16
{
 STACK_VAR INTEGER i
 STACK_VAR INTEGER j
 STACK_VAR INTEGER l
 l = LENGTH_ARRAY(utf16)

 FOR (i = 2; i <= l; i = i + 2)     // copy the even bytes to ch[], as a Unicode to ASCII hack
 {
	j++
	ch[j] = utf16[i]
 }
 
 SET_LENGTH_ARRAY(ch,j)
 RETURN j
}

DEFINE_FUNCTION INTEGER WC_TO_ASCII(WIDECHAR WC[], CHAR ch[])
// hack, to quickly get the Unicode to print as ASCII in a terminal, instead of e.g. writing the Unicode to a file then reading it there...
// assumption 1: there are an even number of bytes in WC
{
 STACK_VAR INTEGER i
 STACK_VAR INTEGER j
 STACK_VAR INTEGER l
 STACK_VAR CHAR sTmp[2]
 l = LENGTH_ARRAY(WC)

 FOR (i = 2; i <= l; i = i + 2)     // copy the even bytes to ch[], as a Unicode to ASCII hack
 {
	sTmp = RAW_BE(WC[i])
	j++
	ch[j] = sTmp[2]
 }
 
 SET_LENGTH_ARRAY(ch,j)
 RETURN j
}
DEFINE_FUNCTION CHAR[256] ASCII_TO_UTF16BE(CHAR CH[])
// purpose: convert ASCII to Unicode UTF-16 BE
// hack: just adds a $00 MSB to each character
{
 STACK_VAR CHAR utf16[256]
 STACK_VAR INTEGER i
 FOR (i = 1; i <= LENGTH_ARRAY(CH); i++)
 {
	utf16 = "utf16,$00,CH[i]"
 }
 return utf16
}
                                    // formatHEX to send utf16 BE to TP with ^UNI or ^BMF %UN commands
DEFINE_FUNCTION CHAR[256] UTF16BE_FORMAT_TP(CHAR bytes[])
{
 STACK_VAR LONG i
 STACK_VAR CHAR sTmp[256]

 FOR (i = 1; i <= LENGTH_ARRAY(bytes); i++)
 {
	sTmp = "sTmp,FORMAT('%02X',bytes[i])"
 }
 RETURN sTmp
}

(***********************************************************)
(*                 STARTUP CODE GOES BELOW                 *)
(***********************************************************)
DEFINE_START

{ // local scope for disco page cleanup STACK_VAR
 STACK_VAR INTEGER i
 FOR (i = MAX_LENGTH_ARRAY(hqPD_DvMgr); i; i--)
 {
	flagDiscoMatched[i] = FALSE
	flagDiscoRowSelected[i] = FALSE
 }
 SET_LENGTH_ARRAY(hqPD_DiscoBC,LENGTH_ARRAY(hqPD_DvMgr))
                                    // This will clear text on the whole page on the first boot after a code load,
																		// as hqPD_DvMgr hasn't been initialized yet. 
 FOR (i = btnDiscoSelectCol1[LENGTH_ARRAY(hqPD_DvMgr)+1]; i <= btnDiscoSelectCol4[LENGTH_ARRAY(btnDiscoSelectCol4)]; i++ )
 {
	SEND_COMMAND dvTP,"'^TXT-',ITOA(i),',0,'"
 }
} // END local scope for disco page cleanup STACK_VAR

																		/****************************/
                                    /* init ip and disco arrays */
																		/****************************/
setIpHiQNet()
																		/************************/
                                    /* init HQn Data arrays */
																		/************************/
																		// completed via DATA_EVENT[dvSelf]/STRING: handler
flagsNX_GET_MAC = $1
IP_CLIENT_OPEN(dvSelf.PORT,'127.0.0.1',23,IP_TCP)

(***********************************************************)
(*                MODULE DEFINITIONS GO BELOW              *)
(***********************************************************)

(***********************************************************)
(*                THE EVENTS GO BELOW                      *)
(***********************************************************)
DEFINE_EVENT                        // NX

DATA_EVENT[dvSelf]
{
 ONLINE:
 {
	ON[flagHQN_online[1]]             // have to turn on NX Master HQn Online button somewhere...why not here?
	IF (flagsNX_GET_MAC & $1)
	{
																		// use $2 as flag dvSelf is ONLINE
	 flagsNX_GET_MAC = flagsNX_GET_MAC | $2
	 SEND_STRING DATA.DEVICE,"'MSG ON',13,10"
	 SEND_STRING DATA.DEVICE,"'GET IP',13,10"
	}
	ELSE
	{
	 IP_CLIENT_CLOSE(DATA.DEVICE.PORT)
	}
 }
 STRING:
 {
	IF (flagsNX_GET_MAC & $1)        // flag to GET MAC
	{
	 STACK_VAR INTEGER nBytesToTrash
	 nBytesToTrash = FIND_STRING(DATA.TEXT,'MAC Address',1)
	 IF (nBytesToTrash)
	 {	
		GET_BUFFER_STRING(DATA.TEXT,nBytesToTrash+11)
																		//  MAC as '00:60:9f:xx:xx:xx'
		setHqData_Default(GET_BUFFER_STRING(DATA.TEXT,17))
		IP_CLIENT_CLOSE(dvSelf.PORT)
		flagsNX_GET_MAC = flagsNX_GET_MAC & (!$1)
		
		IP_BOUND_CLIENT_OPEN (dvHQnDiscoBC.PORT,PORT_HQN,'192.168.1.255',PORT_HQN,IP_UDP_2WAY)
	 }
	}
 }
 ONERROR:
 {
	flagsNX_GET_MAC = flagsNX_GET_MAC & (!$2)
 }
 OFFLINE:
 {
	flagsNX_GET_MAC = flagsNX_GET_MAC & (!$2)
 }
}

DEFINE_EVENT                        // Command Line control
DATA_EVENT[vdvControl]
{
 ONLINE:
 {
 	wait 50                       // WAIT FOR DEVICE HOLDOFF
	{
	// WHAT VERSION?
	SEND_STRING 0:0:0,"'Using     ',__HIQNET_COMM_NAME__,
								     ', v',__HIQNET_COMM_VERSION__"
	}
 }
 COMMAND:
 {
	
  STACK_VAR CHAR CMD[255]
	// pre-pended Args for CMD
	STACK_VAR CHAR sArgs[44][16]
	STACK_VAR LONG lArgs[44]
	STACK_VAR INTEGER lenA
	IF (FIND_STRING(DATA.TEXT,' ',1))
	{
	 STACK_VAR LONG i
	 lenA = stringTokenizer(REMOVE_STRING(DATA.TEXT,' ',1),',',sArgs)
	                                  // remove the space
   sArgs[lenA] = LEFT_STRING(sArgs[lenA],LENGTH_ARRAY(sArgs[lenA])-1)
	 FOR (i = lenA; i; i--)
	 {
		SWITCH (sArgs[i][1])
		{
		 CASE '$': { lArgs[i] = HEXTOI(sArgs[i]) }
		 DEFAULT:  { lArgs[i] = ATOI(sArgs[i]) }
		}
	 }
	}

  IF (FIND_STRING(DATA.TEXT,'-',1)) { CMD = UPPER_STRING(REMOVE_STRING(DATA.TEXT,'-',1)) }
  ELSE                              { CMD = UPPER_STRING(LEFT_STRING(DATA.TEXT,MAX_LENGTH_ARRAY(CMD))) }

  SWITCH (CMD)
  {
	 CASE 'DISCOLISTENER-':
	 CASE 'DL-':
	 {
		flagHQnDiscoListener = ATOI(DATA.TEXT)
		AMX_LOG(AMX_ERROR,"ITOA(__LINE__),' DISCO-',ITOA(nHQnDiscoBC),' LISTENER-',ITOA(flagHQnDiscoListener)")
	 }
	 CASE 'DISCO-':
	 CASE 'DBC-':
	 {
		nHQnDiscoBC = ATOI(DATA.TEXT)
		IF (nHQnDiscoBC)
		{
		 flagHQnDiscoListener = TRUE
		 AMX_LOG(AMX_ERROR,"ITOA(__LINE__),' DISCO-',ITOA(nHQnDiscoBC),' LISTENER-',ITOA(flagHQnDiscoListener)")
		 IF (!TIMELINE_ACTIVE(tidHQnDiscoBC))
		 {
			TIMELINE_CREATE(tidHQnDiscoBC,lHQnDiscoBC,LENGTH_ARRAY(lHQnDiscoBC),TIMELINE_RELATIVE,TIMELINE_REPEAT)
		 }
		}
		ELSE
		{
		 AMX_LOG(AMX_ERROR,"ITOA(__LINE__),' DISCO-',ITOA(nHQnDiscoBC),' LISTENER-',ITOA(flagHQnDiscoListener)")
		 IP_CLIENT_CLOSE(dvHQnDiscoBC.PORT)
		 TIMELINE_KILL(tidHQnDiscoBC)
		}
	 }
	 CASE '?DISCO':
	 CASE '?DBC':
	 CASE '?DL:':
	 {
		AMX_LOG(AMX_ERROR,"ITOA(__LINE__),' DISCO-',ITOA(nHQnDiscoBC),' LISTENER-',ITOA(flagHQnDiscoListener)")
	 }
	 CASE 'HELLO':                    // SEND_C 33000,'HQNAddr1[,addr2,addrn...] HELLO'
	 {
		IF (lenA)
		{
		 STACK_VAR INTEGER i
		 STACK_VAR INTEGER j
		 STACK_VAR CHAR sTmp[32]
		 FOR (i = lenA; i; i--)
		 {
			FOR (j = LENGTH_ARRAY(hqPD_DiscoBC); j; j--)
			{
			 IF (hqPD_DiscoBC[j].a_nSourceNode == lArgs[i])
			 {
				IF (flagHQN_online[j])
				{
				 AMX_LOG(AMX_ERROR,"ITOA(__LINE__),' ',ITOA(j),' ',ipHiQNet[j].hostname,' ',ipHiQNet[j].ipAddress,' ',DEV_TO_STRING(0:ipHiQNet[j].dvPort:0),' already ONLINE'")
				}
				ELSE
				{
				 AMX_LOG(AMX_ERROR,"ITOA(__LINE__),' ',ITOA(j),' ',ipHiQNet[j].hostname,' ',ipHiQNet[j].ipAddress,' ',DEV_TO_STRING(0:ipHiQNet[j].dvPort:0),' connecting...'")
				 IP_CLIENT_OPEN(ipHiQNet[j].dvPort,
												ipHiQNet[j].ipAddress,
												PORT_HQN,IP_TCP)
			 }
			 BREAK
			 }
			}
		 }
		}
		ELSE
		{
		 AMX_LOG(AMX_ERROR,"ITOA(__LINE__),' try prepending HiQnet address to command, comma-delimited'")
		}
	 }
	 CASE 'HELLO-':                   // SEND_C 33000,'HELLO-1'
	 {
		STACK_VAR INTEGER i
		i = ATOI(DATA.TEXT)
		IF (i >= 1 && i <= LENGTH_ARRAY(ipHiQNet))
		{
		 IF (flagHQN_online[i])
		 {
			AMX_LOG(AMX_ERROR,"ITOA(__LINE__),' ',ipHiQNet[i].hostname,' ',ipHiQNet[i].ipAddress,' ',DEV_TO_STRING(0:ipHiQNet[i].dvPort:0),' already ONLINE'")
		 }
		 ELSE
		 {
			IP_CLIENT_OPEN(ipHiQNet[i].dvPort,
										 ipHiQNet[i].ipAddress,
										 PORT_HQN,IP_TCP)
		 }
		}
		ELSE
		{
		 AMX_LOG(AMX_ERROR,"ITOA(__LINE__),' try HELLO-1 thru HELLO-',ITOA(LENGTH_ARRAY(ipHiQNet))")
		}
	 }
 	 CASE '?DEBUGHQN':
	 CASE '?DH':
	 {
		AMX_LOG(AMX_ERROR,"ITOA(__LINE__),' DEBUGHQN-',ITOA(debugHQN)")
	 }
	 CASE 'DEBUGHQN-':
	 CASE 'DH-':
	 {
		debugHQN = ATOI(DATA.TEXT)
		AMX_LOG(AMX_ERROR,"ITOA(__LINE__),' DEBUGHQN-',ITOA(debugHQN)")
	 }
   CASE 'DISCONNECT-': // no goodbye, just close it...
	 {
		IP_CLIENT_CLOSE(dvHQN[ATOI(DATA.TEXT)].PORT)
	 }
	 CASE 'FLAGS?-':                  // manually check returnHiqMessageFlags() subroutine
	 CASE '?FLAGS-':
	 {
		AMX_LOG(AMX_ERROR,"ITOA(__LINE__),'?FLAGS-',returnHiqMessageFlags(HEXTOI(DATA.TEXT))")
	 }
	 CASE 'GETATTRIBUTES-':
	 CASE 'GA-': // 'GA-ID1,ID2,ID3...ID16'
	 {
		STACK_VAR CHAR sAID[16][8]
		STACK_VAR INTEGER nAID[16]
		STACK_VAR INTEGER l
		STACK_VAR INTEGER i
		STACK_VAR CHAR sOutgoing[128]
		l = stringTokenizer(DATA.TEXT,',',sAID)
		SET_LENGTH_ARRAY(nAID,l)
		FOR (i = l; i; i--)
		{
		 nAID[i] = ATOI(sAID[i])
		}
		#warn 'for now, send GET ATTRIBUTES to all HiQnet online'
		FOR (i = MAX_LENGTH_ARRAY(flagHQN_online); i; i--)
		{
		 IF (flagHQN_online[i])
		 {
			IF (i > 1)
			{
			 buildHiQGetAttributes(hqDataOut[i],l,nAID,sOutgoing,'GA-')
			 SEND_STRING dvHQN[i],sOutgoing 
			}
		 }
		}
	 }
	 CASE 'GET-':
	 {
		SWITCH (DATA.TEXT)
		{
		 CASE 'MAC':                   // test the get (self) MAC address algorithm used in DEFINE_START
		 {
			flagsNX_GET_MAC = flagsNX_GET_MAC | $1
			IF (flagsNX_GET_MAC & $2 = $0)  // if dvSelf is OFFLINE
			{
			 IP_CLIENT_OPEN(dvSelf.PORT,'127.0.0.1',23,IP_TCP)
			}
		 }
		 CASE 'MASTERSN':
		 CASE 'MSN':
		 {
			AMX_LOG(AMX_ERROR,MASTER_SN)
		 }
		 CASE 'MASTERVERSION':
		 CASE 'MV':
		 {
			AMX_LOG(AMX_ERROR,getMasterVersion(0:1:0))
		 }
		}
	 }
	 CASE 'GOODBYE':
	 {
		IF (lenA)
		{
		 STACK_VAR INTEGER i
		 STACK_VAR INTEGER j
		 STACK_VAR CHAR sTmp[32]
		 FOR (i = lenA; i; i--)
		 {
			//FOR (j = LENGTH_ARRAY(ipHiQNet); j; j--)
			FOR (j = 1; j <= LENGTH_ARRAY(hqPD_DiscoBC); j++)
			{
			 IF (hqPD_DiscoBC[j].a_nSourceNode == lArgs[i])
			 {
				IF (flagHQN_online[j])
				{
				 AMX_LOG(AMX_ERROR,"ITOA(__LINE__),' ',ipHiQNet[j].hostname,' ',ipHiQNet[j].ipAddress,' ',DEV_TO_STRING(0:ipHiQNet[j].dvPort:0),' goodbye..'")
				 HiQGoodbye(dvHQN[j],hqDataIn[j],"hqAtr_DM[j].name.str,' <-'")
				}
				ELSE
				{
				 AMX_LOG(AMX_ERROR,"ITOA(__LINE__),' ',ipHiQNet[j].hostname,' ',ipHiQNet[j].ipAddress,' ',DEV_TO_STRING(0:ipHiQNet[j].dvPort:0),' already OFFLINE'")
			 }
			 BREAK
			 }
			}
		 }
		}
		ELSE
		{
		 AMX_LOG(AMX_ERROR,"ITOA(__LINE__),' try prepending HiQnet address to command, comma-delimited'")
		}
	 }
	 CASE 'GOODBYE-':                 // SEND_C 33000,'GOODBYE-1'
   {
		STACK_VAR INTEGER i
		i = ATOI(DATA.TEXT)
		IF (i >= 1 && i <= LENGTH_ARRAY(ipHiQNet))
		{
		 IF (flagHQN_online[i])
		 {
			hiqGoodbye(dvHQN[i],hqDataOut[i],"ipHiQNet[i].hostname,' <-'")
		 }
		 ELSE
		 {
			AMX_LOG(AMX_ERROR,"ITOA(__LINE__),' ',ipHiQNet[i].hostname,' ',DEV_TO_STRING(0:ipHiQNet[i].dvPort:0),' not ONLINE'")
		 }
		}
		ELSE
		{
		 AMX_LOG(AMX_ERROR,"ITOA(__LINE__),' try GOODBYE-1 thru GOODBYE-',ITOA(LENGTH_ARRAY(ipHiQNet))")
		}
	 }
	 CASE 'MID?-':                    // manually check returnHiQMessageID() subroutine
	 CASE '?MID-':
	 {
		AMX_LOG(AMX_ERROR,"ITOA(__LINE__),'?MIB-',returnHiQMessageID(HEXTOI(DATA.TEXT))")
	 }

   CASE 'IP?':                      // get ip addresses
   CASE '?IP':
   {
		STACK_VAR INTEGER i
		FOR (i = LENGTH_ARRAY(ipHiQNet); i; i--)
		{
		 AMX_LOG(AMX_ERROR,"ITOA(__LINE__),' ip ',ITOA(i),' ',
																	 ipHiQNet[i].hostname,' = ',
																	 '''',IPHIQNET[I].IPADDRESS,''''")
		}
   }
	 CASE '?KEEPALIVEPERIODPERCENTAGE':
	 CASE '?KAPP':
	 {
		IF (lenA)
		{
		 STACK_VAR INTEGER i
		 FOR (i = lenA; i; i--)
		 {
			AMX_LOG(AMX_ERROR,"ITOA(__LINE__),' ',
													 ipHiQNet[lArgs[i]].hostname,
													' KAP_pct(',(sArgs[i]),')-',FTOA(lHQN_KAP_p[lArgs[i]]),
													' KAP ',ITOA(lHQN_KeepAlive[lArgs[i]][1]),'ms'")
		 }
		}
		ELSE
		{
		 AMX_LOG(AMX_ERROR,"ITOA(__LINE__),' try prepending 1 - ',ITOA(LENGTH_ARRAY(ipHiQNet)),' to command, comma-delimited'")
		}
	 }
	 CASE 'KEEPALIVEPERIODPERCENTAGE-':
	 CASE 'KAPP-':
	 {
		IF (lenA)
		{
		 STACK_VAR INTEGER i
		 FOR (i = lenA; i; i--)
		 {
			lHQN_KAP_p[lArgs[i]] = ATOF(DATA.TEXT)
			IF (!hqPD_DvMgr[i].f_nKeepAlivePeriod)
			{
			 hqPD_DvMgr[i].f_nKeepAlivePeriod = 10000
			}
			lHQN_KeepAlive[lArgs[i]][1] = TYPE_CAST(hqPD_DvMgr[i].f_nKeepAlivePeriod * lHQN_KAP_p[lArgs[i]])
			IF (lHQN_KeepAlive[lArgs[i]][1] < 125)
			{
			 lHQN_KeepAlive[lArgs[i]][1] = 125 // 250ms min KAP * 0.5
			}
			AMX_LOG(AMX_ERROR,"ITOA(__LINE__),' ',
													 ipHiQNet[lArgs[i]].hostname,
													' KAP_p(',(sArgs[i]),')-',FTOA(lHQN_KAP_p[lArgs[i]]),
													' KAP ',ITOA(lHQN_KeepAlive[lArgs[i]][1]),'ms'")
			IF (TIMELINE_ACTIVE(lArgs[i]))
			{
			 TIMELINE_RELOAD(lArgs[i],lHQN_KeepAlive[lArgs[i]],1)
			}
		 }
		}
		ELSE
		{
		 AMX_LOG(AMX_ERROR,"ITOA(__LINE__),' try prepending 1 - ',ITOA(LENGTH_ARRAY(ipHiQNet)),' to command, comma-delimited'")
		}
	 }
 	 CASE '?MAXMESSAGESIZE':
	 CASE '?MMS':
	 {
		AMX_LOG(AMX_ERROR,"ITOA(__LINE__),' MAXMESSAGESIZE-',ITOA(hqPD_DiscoBC[ndxSelf].e_lMaxMessageSize)")
	 }
	 CASE 'MAXMESSAGESIZE-':
	 CASE 'MMS-':
	 {
		hqPD_DiscoBC[ndxSelf].e_lMaxMessageSize = ATOI(DATA.TEXT)
		AMX_LOG(AMX_ERROR,"ITOA(__LINE__),' MAXMESSAGESIZE-',ITOA(hqPD_DiscoBC[ndxSelf].e_lMaxMessageSize)")
	 }
	 CASE '?NOSESSION':
	 CASE '?NS':
	 {
		IF (lenA)
		{
		 STACK_VAR INTEGER i
		 FOR (i = lenA; i; i--)
		 {
			AMX_LOG(AMX_ERROR,"ITOA(__LINE__),' NOSESSION(',(sArgs[i]),')-',ITOA(flagHQN_NoSession[lArgs[i]])")
		 }
		}
		ELSE
		{
		 AMX_LOG(AMX_ERROR,"ITOA(__LINE__),' try prepending 1 - ',ITOA(LENGTH_ARRAY(ipHiQNet)),' to command, comma-delimited'")
		}
	 }
	 CASE 'NOSESSION-':
	 CASE 'NS-':
	 {
		IF (lenA)
		{
		 STACK_VAR INTEGER i
		 FOR (i = lenA; i; i--)
		 {
			flagHQN_NoSession[lArgs[i]] = ATOI(DATA.TEXT)
			AMX_LOG(AMX_ERROR,"ITOA(__LINE__),' NOSESSION(',(sArgs[i]),')-',ITOA(flagHQN_NoSession[lArgs[i]])")
		 }
		}
		ELSE
		{
		 AMX_LOG(AMX_ERROR,"ITOA(__LINE__),' try prepending 1 - ',ITOA(LENGTH_ARRAY(ipHiQNet)),' to command, comma-delimited'")
		}
	 }
	 CASE 'ONLINE?':
	 CASE '?ONLINE':
	 {
		STACK_VAR INTEGER i
		STACK_VAR INTEGER j
		FOR (i = MAX_LENGTH_ARRAY(flagHQN_online); i; i--)
		{
		 IF (flagHQN_online[i])
		 {
			j++
			AMX_LOG(AMX_ERROR,"ITOA(__LINE__),
												' ONLINE ',ITOA(i),' ',
												ipHiQNet[i].hostname,' = ',
												'''',IPHIQNET[i].IPADDRESS,''''")
		 }
		}
		IF (!j)
		{
			AMX_LOG(AMX_ERROR,"ITOA(__LINE__),' ALL HiQNet OFFLINE'")
		}
		
	 }
	 CASE 'PARSEHEADERS-':            // parse and print header data. Must be in "CH,EH,MPH,SH" order
	 CASE 'PH-':
	 {
		STACK_VAR CHAR sHeader[255]
		STACK_VAR CHAR sBytes[255][5]
		STACK_VAR INTEGER l             // Length
		STACK_VAR INTEGER i
		STACK_VAR _sHiQNet_Common_Header hqCH

		IF (FIND_STRING(DATA.TEXT,',',1))     // comma delimited hex, with or without '$' prefix
		{
		 l = stringTokenizer(DATA.TEXT,',',sBytes)
		 AMX_LOG(AMX_ERROR,"ITOA(__LINE__),' PH-length ',ITOA(l)")
		}
		ELSE IF (FIND_STRING(DATA.TEXT,"$5C,$58",1))  // bytes that start with \x prefix, no trailing delimiter
		{
		 REMOVE_STRING(CMD,"$5C,$58",1)         // trash the first '0x'
		 l = stringTokenizer(DATA.TEXT,"$5C,$58",sBytes)
		 AMX_LOG(AMX_ERROR,"ITOA(__LINE__),' l = ',ITOA(l)")
		}
		ELSE
		{
		 printToDebug(0:1:0,'PH fail:',DATA.TEXT,131,1,PRINT_HEX)
		}
		sHeader = ''
		FOR (i = l;i;i--)
		{
		 sHeader[i] = HEXTOI(sBytes[i])
		}
		SET_LENGTH_ARRAY(sHeader,l)
		printToDebug(0:1:0,'PH-',sHeader,131,1,PRINT_HEX)
		
		parseHiQHeaderCommon(hqCH,sHeader,__LINE__,'PH-')
		printHiQHeaderCommon(hqCH,__LINE__,'PH-')

		{ // local scope for STACK_VAR
		 STACK_VAR _sHiQnet_Error_Header hqEH

		 WHILE(LENGTH_ARRAY(sHeader))    // if there is more...
		 {
																		 // error messages may contain "CH,EH.L_nCode" in the body of the error, e.g. ITHD4,DCIN,DCIDA, between the Unicode
			IF ((hqCH.I_nFlags & HIQ_MID_FLAG_ERROR == HIQ_MID_FLAG_ERROR) && (!hqEh.L_nCode))
			{
			 STRING_TO_VARIABLE(hqEH.L_nCode,"$E2,GET_BUFFER_STRING(sHeader,2)",1)
			 IF (debugHQN >= DEBUG_TERSE)
			 {
				AMX_LOG(AMX_ERROR,"ITOA(__LINE__),' ','PH-','Error Header CODE ',returnHiQControlNetworkEventID(hqEH.L_nCode)")
			 }
			}
			#WARN 'todo: add Multi Part Header and Session Header support here'
			ELSE
			{
			 printToDebug(0:1:0,'PH-unparsed:',GET_BUFFER_STRING(sHeader,LENGTH_ARRAY(sHeader)),131,1,PRINT_HEX)
			 BREAK
			}
		 }
		} // END local scope for STACK_VAR
	 }
	 CASE 'SUBSCRIBEPARAMALL-':       // VDO (4 bytes), TYPE (1 byte), sensor rate ms (2 bytes), flags (2 bytes)
	 CASE 'SPA-':                     // for example: SPA-0,1,50,0 appears to work for DSi2.0 and ITHD4, session or no session. More testing TBD
	 {
		LOCAL_VAR VOLATILE INTEGER i
		LOCAL_VAR VOLATILE _sHiQnet_headers hqD
		FOR (i = LENGTH_ARRAY(ipHiQNet); i; i--)
		{
		 IF (flagHQN_online[i]) { BREAK }
		}
		IF (i)
		{
		 STACK_VAR CHAR args[4][16]
		 LOCAL_VAR VOLATILE CHAR sOutgoing[128]
		 STACK_VAR INTEGER len
		 LOCAL_VAR VOLATILE INTEGER nNode
		 LOCAL_VAR LONG lVDO
		 LOCAL_VAR CHAR cType
		 LOCAL_VAR INTEGER nRate
		 LOCAL_VAR INTEGER nFlag
		 SEND_STRING 0,"ITOA(__LINE__),' SPA-',DATA.TEXT"
		 hqD = hqDataOut[i]
		 hqD.hdrCommon.I_nFlags = hqD.hdrCommon.I_nFlags | HIQ_MID_FLAG_GUARANTEED
		 IF(flagHQN_NoSession[i])
		 {
			hqD.hdrCommon.I_nFlags = hqD.hdrCommon.I_nFlags & (!HIQ_MID_FLAG_SESSION_NUMBER)
		 }
		 hqD.hdrCommon.B_cLengthHeader = lengthHiQHeaders(hqD.hdrCommon.I_nFlags)
		 hqD.hdrCommon.C_lLengthMessage = hqD.hdrCommon.B_cLengthHeader
																							 // 2 bytes for HiQnet Address
																							 // 4 bytes for HiQnet VD OBject
																							 // 1 bytes for subscribe type bitmap
																							 // 2 bytes for sensor rate
																							 // 2 bytes for flags
																							 + 11
		 hqD.hdrCommon.H_nMessageID = HIQ_MID_MULTISV_SUBSCRIBE_ALL
		 
		 printHiQHeaders(hqD,hqD.hdrSession.Q_nOther,__LINE__,"ipHiQNet[i].hostname,' <-'")
		 buildHiQHeaders(hqD.hdrCommon,
										 hqD.hdrCommon.I_nFlags,
										 hqD.hdrError.L_nCode,
										 hqD.hdrMultiPart,
										 hqD.hdrSession.Q_nOther,
										 sOutgoing)

		 nNode = hqD.hdrCommon.F_nDestinationNode
		 SEND_STRING 0,"ITOA(__LINE__),FORMAT(' node $%04X',nNode)"
		
		 len = stringTokenizer(DATA.TEXT,',',args)
		 FOR (; len; len--)
		 {
			SWITCH (len)
			{
			 CASE 1:
			 {
				lVDo = HEXTOI(args[len])
				SEND_STRING 0,"ITOA(__LINE__),FORMAT(' VDO $%08X',lVDO)"
			 }
			 CASE 2:
			 {
				cType = HEXTOI(args[len])
				SEND_STRING 0,"ITOA(__LINE__),FORMAT(' type $%02X',cType)"
			 }
			 CASE 3:
			 {
				nRate = HEXTOI(args[len])
				SEND_STRING 0,"ITOA(__LINE__),FORMAT(' sensor rate $%04X',nRate)"
			 }
			 CASE 4:
			 {
				nFlag = HEXTOI(args[len])
				SEND_STRING 0,"ITOA(__LINE__),FORMAT(' flags $%04X',nFlag)"
			 }
			}
		 }
		 IF (!cType) { cType = $05 }    // HIQ_SUB_TYPE_SV_NON_SENSOR | HIQ_SUB_TYPE_ATTRIBUTES
		 IF (nRate < 50) { nRate = HIQ_SUB_SENSOR_RATE_NX }
		 sOutgoing = "sOutgoing,
									RAW_BE(nNode),
									RAW_BE(lVDO),
									cType,
									RAW_BE(nRate),
									RAW_BE(nFlag)"
		 printToDebug(0:1:0,"ITOA(__LINE__),
												FORMAT(' $%02X ',ITOA(LENGTH_ARRAY(sOutgoing))),' ',
												ipHiQNet[i].hostname,' <-',
												' SUBSCRIBE ALL '",
												sOutgoing,131,1,PRINT_HEX)
		 SEND_STRING dvHQN[i],sOutgoing
		 #WARN 'todo: add UNSUBSCRIBE ALL'
//		 WAIT 600 'unsubscribe' // this seems to have no effect on the flood...
//		 {
//			hqD.hdrCommon.H_nMessageID = HIQ_MID_MULTISV_SUBSCRIBE_ALL
//			buildHiQHeaders(hqD.hdrCommon,
//								hqD.hdrCommon.I_nFlags,
//								hqD.hdrError.L_nCode,
//								hqD.hdrMultiPart,
//								hqD.hdrSession.Q_nOther,
//								sOutgoing)
//		 sOutgoing = "sOutgoing,
//									RAW_BE(hqD.hdrCommon.D_nSourceNode),
//									RAW_BE(hqD.hdrCommon.E_lSourceVDObject),
//									cType"
//		 printToDebug(0:1:0,"ITOA(__LINE__),
//												FORMAT(' $%02X ',ITOA(LENGTH_ARRAY(sOutgoing))),' ',
//												ipHiQNet[i].hostname,' <-',
//												' UNSUBSCRIBE ALL '",
//												sOutgoing,131,1,PRINT_HEX)
//		 SEND_STRING dvHQN[i],sOutgoing
//		 }
		}
		ELSE   { SEND_STRING 0,"ITOA(__LINE__),' no HiQnet ONLINE'" }
	 }
	 CASE '?VERSION':
	 CASE '?VER':
	 {
	 	SEND_STRING 0:1:0,"'Using     ',__HIQNET_COMM_NAME__,
											 ', v',__HIQNET_COMM_VERSION__"
	 }
	 DEFAULT:                         // HEX strings from Audio Architect
	 {
		STACK_VAR CHAR sOutgoing[255]
		STACK_VAR CHAR sBytes[255][5]
		STACK_VAR INTEGER l             // Length
		STACK_VAR INTEGER i

		IF (FIND_STRING(CMD,',',1))     // comma delimited hex, with or without '$' prefix
		{
		 l = stringTokenizer(CMD,',',sBytes)
		 AMX_LOG(AMX_ERROR,"ITOA(__LINE__),' l = ',ITOA(l)")
		}
		ELSE IF (FIND_STRING(CMD,"$5C,$58",1))  // bytes that start with \x prefix, no trailing delimiter
		{
		 REMOVE_STRING(CMD,"$5C,$58",1)         // trash the first '0x'
		 l = stringTokenizer(CMD,"$5C,$58",sBytes)
		 AMX_LOG(AMX_ERROR,"ITOA(__LINE__),' l = ',ITOA(l)")
		}
		ELSE
		{
		 printToDebug(0:1:0,'CMD fail:',CMD,131,1,PRINT_HEX)
		}
		sOutgoing = ''
		FOR (i = l;i;i--)
		{
		 sOutgoing[i] = HEXTOI(sBytes[i])
		}
		SET_LENGTH_ARRAY(sOutgoing,l)

		{ // local scope
		 STACK_VAR INTEGER nDestinationNode
		 STACK_VAR _sHiQnet_headers hqDO
		 STACK_VAR INTEGER nSession;
		 STACK_VAR LONG ret
		 STACK_VAR CHAR sDebug[64]
		 STACK_VAR CHAR sTmp[128]
		 STACK_VAR CHAR sNode[64]
		 
		 STRING_TO_VARIABLE(nDestinationNode,"$E2,sOutgoing[13],sOutgoing[14]",1)
		 sNode = "FORMAT('$%04X',nDestinationNode),' ',FORMAT('(%5d)',nDestinationNode)"
SEND_STRING 0,"ITOA(__LINE__),' Destination Node ',sNode"

		 //  Destination Node gets the string
		 FOR (i = 1; i <= LENGTH_ARRAY(hqPD_DiscoBC); i++)
		 {
			IF (nDestinationNode == hqPD_DiscoBC[i].a_nSourceNode)
			{
			 BREAK // first device found gets the string
			}
		 }

//  old code: first device found online gets the string
//		 FOR (i = MAX_LENGTH_ARRAY(flagHQN_online); i; i--)
//		 {
//			IF (flagHQN_online[i])
//			{
//			 BREAK // first device found gets the string
//			}
//		 }
		 IF (i) { sDebug = "ipHiQNet[i].hostname,' <-'" }
		 ELSE   { sDebug = ' <-' }

		 IF (debugHQN >= DEBUG_TERSE)
		 {
			printToDebug(0:1:0,"ITOA(__LINE__),FORMAT(' $%02X ',l),sDebug,' '",sOutgoing,131,1,PRINT_HEX)
		 }
		 
		 #WARN 'todo: parse and rebuild string, updating sequence # and adding session # if necessary'
		 
		 IF (i) { SEND_STRING dvHQN[i],sOutgoing }
		 ELSE   { SEND_STRING 0,"ITOA(__LINE__),' HiQnet ',sNode,' not found'" }
		}
	 }
	}
 }
}

DEFINE_EVENT                        // HiQnet Array

DATA_EVENT[dvHQN]
{
 ONERROR:
 {
	STACK_VAR INTEGER ndxDev
	ndxDev = GET_LAST(dvHQN)
	IF (ndxDev <> ndxSelf)
	{
	 IF (debugHQN >= DEBUG_TERSE)
	 {
		AMX_LOG(AMX_ERROR,"ITOA(__LINE__),' ',
											 LDATE,' ',TIME,' ',
											 DEV_TO_STRING(DATA.DEVICE),' ',
											 'err-',ITOA(DATA.NUMBER),' ',
											 '''',ipHiQNet[ndxDev].hostname,''''")
	 }
	 IF (!(DATA.NUMBER == 0 &&         // this happens sometimes, older(?) NetLinx masters
				 DATA.NUMBER == 9 &&         // already closed
				 DATA.NUMBER == 14))         // local port already used
	 {
		flagHQN_online[ndxDev] = FALSE
	 }
	 IF (TIMELINE_ACTIVE(tidHQN[ndxDev]))
	 {
		TIMELINE_KILL(tidHQN[ndxDev])
		IF (debugHQN >= DEBUG_VVV)
		{
		 AMX_LOG(AMX_WARNING,"ITOA(__LINE__),' ',ipHiQNet[ndxDev].hostname,' TIMELINE_KILL(',ITOA(tidHQN[ndxDev]),') ',ITOA(TIMELINE_ACTIVE(tidHQN[ndxDev]))")
		}
	 }
	}
 }
 ONLINE:
 {
	STACK_VAR INTEGER ndxDev
	ndxDev = GET_LAST(dvHQN)
	IF (ndxDev <> ndxSelf)
	{
	 // 1. Send DISCO QUERY
	 // 2. Send MULTI PARAMETER GET 0
	 // 3.a Send HELLO QUERY with SESSION as PAYLOAD, not header
	 // 3.b Respond to HELLO INFO response with ERROR if sessions are not desired
	 // 4. Send GET ATTRIBUTES 1,0
	 // 5. Send Subscribe All to Object 0,Non-Sensor, 50ms Sensor Rate, NO initial update (to mimic AA behavior with CDi/DSi2.0).
	 // 6. Send GET ATTRIBUTES for Device Name (not done yet by AA)
	 
	 STACK_VAR CHAR sDebug[64]
	 STACK_VAR CHAR sOutgoing[256]
	 LOCAL_VAR INTEGER nAttributeCount
	 LOCAL_VAR INTEGER nAID[16]
	 
	 flagHQN_online[ndxDev] = TRUE
	 sDebug = "ipHiQNet[ndxDev].hostname,' <-'"
	 
	 IF (debugHQN >= DEBUG_TERSE)
	 {
		SEND_STRING 0,"ITOA(__LINE__),' ',LDATE,' ',TIME,' ',DEV_TO_STRING(DATA.DEVICE),' ONLINE ',DATA.SOURCEIP,' ''',ipHiQNet[ndxDev].hostname,''''"
	 }
																		 // updateSession # for initial Hello Query
	 hqPH_Out[ndxDev].session.a_nSelf++ // session number from 1 to 65535, increments each new session
	 IF (!hqPH_Out[ndxDev].session.a_nSelf) { hqPH_Out[ndxDev].session.a_nSelf++ }
	 
																		 // set sequence to 0 here, it will be incremented when the strings are built
	 hqDataOut[ndxDev].hdrCommon.K_nSequence = 0
																		 // flag system to send a Subscribe All after the initial GET ATTRIBUTES is received.
	 flagSubscribeAll_after_GA[ndxDev] = TRUE
	 
																		 // info flag not set for initial DISCO, this makes it a QUERY
	 buildHiQDisco(hqPD_DiscoBC[ndxSelf],hqDataOut[ndxDev],HIQ_MID_FLAG_GUARANTEED,sOutgoing,__LINE__,sDebug)
	 SEND_STRING dvHQN[ndxDev],sOutgoing

	 buildHiQMultiParameterGet(hqDataOut[ndxDev],sOutgoing,sDebug)	
	 SEND_STRING dvHQN[ndxDev],sOutgoing
	 
																		 // info flag not set for initial HELLO, this makes it a QUERY
																		 // session flag is not set either, .nSelf session is sent as a payload not a header
	 buildHiQHello(hqPH_Out[ndxDev],hqDataOut[ndxDev],HIQ_MID_FLAG_GUARANTEED,sOutgoing,__LINE__,sDebug)
	 SEND_STRING dvHQN[ndxDev],sOutgoing
	 
	 IF (!nAttributeCount)
	 {
		nAttributeCount = 1
		nAID[1] = 0 // Device Manager Class Name
		SET_LENGTH_ARRAY(nAID,nAttributeCount)
	 }
	 buildHiQGetAttributes(hqDataOut[ndxDev],nAttributeCount,nAID,sOutgoing,sDebug)
	 SEND_STRING dvHQN[ndxDev],sOutgoing
	}
 }
 STRING:
 {
	STACK_VAR INTEGER ndxDev
	ndxDev = GET_LAST(dvHQN)
	IF (ndxDev <> ndxSelf)
	{
	 STACK_VAR LONG len
	 STACK_VAR LONG lenParsed
	 STACK_VAR INTEGER nFlags
	 STACK_VAR CHAR sOutgoing[2000]
	 STACK_VAR CHAR sDebug[64]

	 len = LENGTH_ARRAY(DATA.TEXT)
	 sDebug = "ipHiQNet[ndxDev].hostname,' ->'"

		AMX_LOG(AMX_WARNING,"ITOA(__LINE__),' DATA_EVENT[',DEV_TO_STRING(DATA.DEVICE),'] STRING: ',ITOA(LENGTH_ARRAY(DATA.TEXT)),' ',DATA.SOURCEIP")
	 IF (debugHQN >= DEBUG_VERBOSE)
	 {

		IF (debugHQN >= DEBUG_VVV)
		{
		 printToDebug(0:1:0,"ITOA(__LINE__),' ',ipHiQNet[ndxDev].hostname,' ',FORMAT('$%X',len),' -> '",DATA.TEXT,131,0,PRINT_HEX)
		}
	 }
	 // this is an attempt to workaround the message flood received from 4x3500HD 
	 // after it gets a bad DISCO INFO() Keep Alive message from NetLinx
 //	IF (TIMELINE_ACTIVE(ndxDev))
 //	{
 //	 TIMELINE_PAUSE(ndxDev)
 //	 IF (debugHQN >= DEBUG_VVV) { AMX_LOG(AMX_WARNING,"ITOA(__LINE__),' ',sDebug,' TIMELINE_PAUSE(',ITOA(tidHQN[ndxDev]),') ',ITOA(TIMELINE_ACTIVE(tidHQN[ndxDev]))") }
 //	}

	 lenParsed = parseHiQData(hqDataIn[ndxDev],hqPH_In[ndxDev],hqPD_DvMgr[ndxDev],hqAtr_DM[ndxDev],DATA.TEXT, __LINE__,"ipHiQNet[ndxDev].hostname,' ->'")
	 nFlags = hqDataIn[ndxDev].hdrCommon.I_nFlags
	
																		 /********************************************/
																		 /* now that the data stored, evaluate it... */
																		 /********************************************/
	 IF (debugHQN >= DEBUG_VERBOSE)
	 {
		AMX_LOG(AMX_WARNING,"ITOA(__LINE__),' ',sDebug,' EVALUATING parsed data...'")

		AMX_LOG(AMX_WARNING,"ITOA(__LINE__),' ',sDebug,' checking msg length'")
	 }

																		 /************************/
																		 /* check message length */
																		 /************************/
	 IF (len <> hqDataIn[ndxDev].hdrCommon.C_lLengthMessage)
	 {
		AMX_LOG(AMX_ERROR,"ITOA(__LINE__),' ',sDebug,
																			' length message ERROR ',ITOA(hqDataIn[ndxDev].hdrCommon.C_lLengthMessage),' <>',
																			' len data ',ITOA(len)")
	 }
	 ELSE
	 {
		IF (debugHQN >= DEBUG_VERBOSE)
		{
		 AMX_LOG(AMX_WARNING,"ITOA(__LINE__),' ',sDebug,' length message OK ',ITOA(len)")
		}
	 }

																		 /**************************/
																		 /* check destination node */
																		 /**************************/
	 // if your Hello Query() had an incorrect Destination Node, but the device responded anyways...(e.g. BLU806, CDi DC et al.)
	 IF (debugHQN >= DEBUG_VERBOSE)
	 {
		AMX_LOG(AMX_WARNING,"ITOA(__LINE__),' ',sDebug,' checking expected HiQ Address'")
	 }
	 IF (hqDataOut[ndxDev].hdrCommon.F_nDestinationNode <> hqDataIn[ndxDev].hdrCommon.D_nSourceNode)
	 {
		AMX_LOG(AMX_ERROR,"ITOA(__LINE__),' ',sDebug,
											' Outgoing Destination NODE updated by Incoming Source NODE response:',
											ITOA(hqDataOut[ndxDev].hdrCommon.F_nDestinationNode),' to ',
											ITOA(hqDataIn[ndxDev].hdrCommon.D_nSourceNode)")
		hqDataOut[ndxDev].hdrCommon.F_nDestinationNode = hqDataIn[ndxDev].hdrCommon.D_nSourceNode
	 }
	 ELSE
	 {
		IF (debugHQN >= DEBUG_VERBOSE)
		{
		 AMX_LOG(AMX_WARNING,"ITOA(__LINE__),' ',sDebug,FORMAT(' HiQ Address $%04X OK',hqDataIn[ndxDev].hdrCommon.D_nSourceNode)")
		}
	 }
																		 /********************/
																		 /* check error flag */
																		 /********************/
	 IF (hqDataIn[ndxDev].hdrCommon.I_nFlags & HIQ_MID_FLAG_ERROR == HIQ_MID_FLAG_ERROR)
	 {
		IF (debugHQN >= DEBUG_VERBOSE) 
		{
		 AMX_LOG(AMX_WARNING,"ITOA(__LINE__),' ',sDebug,' checking for additional errors...'") // e.g. ITHD4
		}
		WHILE (LENGTH_ARRAY(DATA.TEXT))
		{
		 _sHiQnet_headers hqE
		 IF (DATA.TEXT[1] == $02)
		 {
			parseHiQHeaders(hqE,DATA.TEXT,__LINE__,sdebug) // e.g. ITHD4 sends multiple complete headers with errors when there are multiple errors
			IF (debugHQN >= DEBUG_VVVV)
			{
			 printHiQHeaders(hqE,hqE.hdrSession.Q_nIncoming,__LINE__,sDebug)
			}
			ELSE IF (debugHQN >= DEBUG_TERSE)
			{
			 printHiQHeaderError(hqE.hdrError,__LINE__,sDebug)
			}
		 }
		 ELSE
		 {
			STACK_VAR CHAR sErr[256]
			STACK_VAR CHAR utf16[256]
			STACK_VAR INTEGER i
			STRING_TO_VARIABLE(i,"$E2,GET_BUFFER_STRING(DATA.TEXT,2)",1)
			utf16 = GET_BUFFER_STRING(DATA.TEXT,i)
			SET_LENGTH_ARRAY(utf16,i - 2)
			UTF16BE_TO_ASCII(utf16,sErr)
			printToDebug(0:1:0,"ITOA(__LINE__),
												 ' ERROR ',
												 FORMAT(' (trimmed %d) ',LENGTH_ARRAY(sErr))",
												 sErr,131,0,PRINT_ASCII_HEX)
		 }
		}
	 }

																		 /**********************/
																		 /* check session flag */
																		 /**********************/
	 IF (debugHQN >= DEBUG_VERBOSE) 
	 {
		AMX_LOG(AMX_WARNING,"ITOA(__LINE__),' ',sDebug,' checking incoming hdr session # flag'")
	 }
	 IF (hqDataIn[ndxDev].hdrCommon.I_nFlags & HIQ_MID_FLAG_SESSION_NUMBER == HIQ_MID_FLAG_SESSION_NUMBER)
	 {
																	// ... and the session # is incorrect
		IF (hqDataIn[ndxDev].hdrSession.Q_nOther <> hqDataIn[ndxDev].hdrSession.Q_nIncoming)
		{
																	// CDiDC and A4 may be using the NX session#, before sending us theirs in a HELLO INFO
		 IF (hqDataIn[ndxDev].hdrSession.Q_nIncoming == hqPH_Out[ndxDev].session.a_nSelf)
		 {
			hqDataIn[ndxDev].hdrSession.Q_nOther = hqPH_Out[ndxDev].session.a_nSelf
			IF (debugHQN >= DEBUG_VERBOSE)
			{
			 AMX_LOG(AMX_WARNING,"ITOA(__LINE__),' ',sDebug,
													 ' Session # confirmed: expected incoming header session # updated to ',
													 FORMAT('$%04X',hqDataIn[ndxDev].hdrSession.Q_nOther)")
			}
		 }
		 ELSE
		 {
			#warn 'todo: send error message to HiQnet device???'
	// Per the HiQnet 3rd party Quick Start guide, 
	//  the error is CASE HIQ_ERROR_BAD_LENGTH "$00,$02", and its payload is "$00,$00"...
	// If an error message is returned in response to a Hello message, 
	//  MultiParamGet(NumParams=0) messages shall be used for backward compatibility in order to start Keep Alives
	// Otherwise use DiscoInfo for Keep Alives:
	//  For each guaranteed connection (such as TCP), 
	//  a Device must transmit a DiscoInfo(I) message no less than 
	//   hqData.disco.nKeepAliveTime milliseconds after it last transmitted a 
	//   DiscoInfo(I) or Application Layer message.
			
			// per HiQNet spec, close the session...if this is not HELLO QUERY Info (e.g. CDi Drivecore, A4)
			IF (debugHQN >= DEBUG_TERSE)
			{
			 AMX_LOG(AMX_ERROR,"ITOA(__LINE__),' ',sDebug,
																				 FORMAT(' expected session $%04X',hqDataIn[ndxDev].hdrSession.Q_nOther),',',
																				 FORMAT(' received session $%04X',hqDataIn[ndxDev].hdrSession.Q_nIncoming)")

			 AMX_LOG(AMX_ERROR,"ITOA(__LINE__),' ',ipHiQNet[ndxDev].hostname,' <- GOODBYE(0:',ITOA(ipHiQNet[ndxDev].dvPort),':0,...) due to SESSION # ERROR'")
			}
			hiqGoodbye(DATA.DEVICE,hqDataOut[ndxDev],"ipHiQNet[ndxDev].hostname,' <-'")
		 }
		}
		ELSE
		{
		 IF (debugHQN >= DEBUG_VERBOSE)
		 {
			AMX_LOG(AMX_WARNING,"ITOA(__LINE__),' ',sDebug,'',
																					FORMAT(' expected session $%04X',hqDataIn[ndxDev].hdrSession.Q_nIncoming),' OK'")
		 }
		}
	 }
	 ELSE
	 {
		IF (debugHQN >= DEBUG_VERBOSE)
		{
		 AMX_LOG(AMX_WARNING,"ITOA(__LINE__),' ',sDebug,' hdr session # not required for this message'")
		}
	 }
																		 /******************************/
																		 /* check 'ack requested' flag */
																		 /******************************/
	 // send back the same message to the sender with the ReqAck flag cleared and the Ack flag set
	 // any data in its payload, that data is not sent back with the acknowledge message
	 #warn 'be sure to perform the acknowledged action :)'
	 IF (hqDataIn[ndxDev].hdrCommon.I_nFlags & HIQ_MID_FLAG_REQUEST_ACK == HIQ_MID_FLAG_REQUEST_ACK)
	 {
		STACK_VAR _sHiQnet_headers hqDO
		STACK_VAR INTEGER nNode_tmp
		STACK_VAR LONG lVDO_tmp
		STACK_VAR LONG lData
		IF (debugHQN >= DEBUG_VERBOSE) 
		{
		 AMX_LOG(AMX_WARNING,"ITOA(__LINE__),' ',sDebug,' ACK REQUESTED '")
		}

		hqDO = hqDataIn[ndxDev]
		lData = TYPE_CAST(hqDO.hdrCommon.C_lLengthMessage - hqDO.hdrCommon.B_cLengthHeader)
		
																		 // assumption: if multipart message, wait for the whole thing...
		IF ((hqDO.hdrCommon.I_nFlags & HIQ_MID_FLAG_MULTI_PART_MSG == HIQ_MID_FLAG_MULTI_PART_MSG) &&
				(hqDO.hdrMultiPart.P_lBytes_remaining <> lData))
		{
		 AMX_LOG(AMX_WARNING,"ITOA(__LINE__),' ',
													sDebug,
													' cannot ACK until complete multi-part message is received  ',
													ITOA(hqDO.hdrMultiPart.P_lBytes_remaining - lData),' bytes TBR...'")
		}
		ELSE
		{
																		 // sending back headers only, not data payload
		 hqDO.hdrCommon.C_lLengthMessage = TYPE_CAST(hqDO.hdrCommon.B_cLengthHeader)
																		 // assumption: swap source/destination addresses
		 nNode_tmp = hqDO.hdrCommon.D_nSourceNode
		 lVDO_tmp = hqDO.hdrCommon.E_lSourceVDObject
		 hqDO.hdrCommon.D_nSourceNode = hqDO.hdrCommon.F_nDestinationNode
		 hqDO.hdrCommon.E_lSourceVDObject = hqDO.hdrCommon.G_lDestinationVDObject
		 hqDO.hdrCommon.F_nDestinationNode = nNode_tmp
		 hqDO.hdrCommon.G_lDestinationVDObject = lVDO_tmp
																		 // clear RegAck flag, set Ack flag
		 hqDO.hdrCommon.I_nFlags = hqDO.hdrCommon.I_nFlags & (!HIQ_MID_FLAG_REQUEST_ACK)
		 hqDO.hdrCommon.I_nFlags = hqDO.hdrCommon.I_nFlags | HIQ_MID_FLAG_ACK
																		 // if sessions are active, use the outgoing session header not the one received
		 IF (hqDO.hdrCommon.I_nFlags & HIQ_MID_FLAG_SESSION_NUMBER == HIQ_MID_FLAG_SESSION_NUMBER)
		 {
			hqDO.hdrSession.Q_nOther = hqDataOut[ndxDev].hdrSession.Q_nOther
		 }
		 IF (debugHQN >= DEBUG_VERBOSE)
		 {
			printHiQHeaders(hqDO,hqDO.hdrSession.Q_nOther,__LINE__,"ipHiQNet[ndxDev].hostname,' <-'")
		 }
		 buildHiQHeaders(hqDO.hdrCommon,hqDO.hdrCommon.I_nFlags,hqDO.hdrError.L_nCode,hqDO.hdrMultiPart,hqDO.hdrSession.Q_nOther,sOutgoing)
		 IF (debugHQN >= DEBUG_VERBOSE)
		 {
			printToDebug(0:1:0,"ITOA(__LINE__),
										FORMAT(' $%02X ',ITOA(LENGTH_ARRAY(sOutgoing))),' ',
										ipHiQNet[ndxDev].hostname,' <- ',
										'ACK '",
										sOutgoing,131,1,PRINT_HEX)
		 }
		 SEND_STRING dvHQN[ndxDev],sOutgoing
		}
	 }
																		 /**************************/
																		 /* message specific bytes */
																		 /**************************/
	 IF (debugHQN >= DEBUG_VERBOSE)
	 {
		AMX_LOG(AMX_WARNING,"ITOA(__LINE__),' ',sDebug,' checking MESSAGE ID''s'")
	 }
	 SWITCH (hqDataIn[ndxDev].hdrCommon.H_nMessageID)
	 {
		CASE HIQ_MID_ROUTING_DISCO_INFO:
		{
		 IF (debugHQN >= DEBUG_VERBOSE)
		 {
			AMX_LOG(AMX_WARNING,"ITOA(__LINE__),' ',sDebug,' ',returnHiQMessageID(hqDataIn[ndxDev].hdrCommon.H_nMessageID)")
		 }
																		 // if this is DISCO INFO, not a DISCO QUERY
		 IF (nFlags & HIQ_MID_FLAG_INFORMATION)
		 {
			IF (debugHQN >= DEBUG_VERBOSE)
			{
			 AMX_LOG(AMX_WARNING,"ITOA(__LINE__),' ',sDebug,' DISCO INFO()'")
			}
 IF (debugHQN >= DEBUG_VVV)
 {
	AMX_LOG(AMX_WARNING,"ITOA(__LINE__),' ',sDebug,' tidActive[',ITOA(tidHQN[ndxDev]),'] ',ITOA(TIMELINE_ACTIVE(tidHQN[ndxDev]))")
 }
			IF (hqPD_DvMgr[ndxDev].f_nKeepAlivePeriod) // if KAP value has been parsed
			{
			 IF (!TIMELINE_ACTIVE(tidHQN[ndxDev]))
			 {
				lHQN_KeepAlive[ndxDev][1] = TYPE_CAST(hqPD_DvMgr[ndxDev].f_nKeepAlivePeriod * lHQN_KAP_p[ndxDev])
				IF (lHQN_KeepAlive[ndxDev][1] < 125)
				{
				 lHQN_KeepAlive[ndxDev][1] = 125// 250ms min KAP * 0.5
				}
				TIMELINE_CREATE(tidHQN[ndxDev],lHQN_KeepAlive[ndxDev],1,TIMELINE_RELATIVE,TIMELINE_REPEAT)
			 }
			}
 IF (debugHQN >= DEBUG_VERBOSE)
 {
	STACK_VAR FLOAT f1
	STACK_VAR FLOAT f2
	f1 = TYPE_CAST(lHQN_KeepAlive[ndxDev][1])
	f2 = TYPE_CAST(hqPD_DvMgr[ndxDev].f_nKeepAlivePeriod)
	AMX_LOG(AMX_WARNING,"ITOA(__LINE__),' ',sDebug,
											' tidActive[',ITOA(tidHQN[ndxDev]),'] ',
											ITOA(TIMELINE_ACTIVE(tidHQN[ndxDev])),',',
											' KA time = ',ITOA(lHQN_KeepAlive[ndxDev][1]),'ms ',
											'(',FTOA(f1/f2),' KAP)'")
 }
		 }
		 ELSE
		 {
			IF (debugHQN >= DEBUG_VERBOSE)
			{
			 AMX_LOG(AMX_WARNING,"ITOA(__LINE__),' ',sDebug,' DISCO QUERY() EVAL ToDo...'")
			}
		 }
		}
		CASE HIQ_MID_ROUTING_HELLO:
		{
		 IF (debugHQN >= DEBUG_VERBOSE)
		 {
			AMX_LOG(AMX_WARNING,"ITOA(__LINE__),' ',sDebug,' ',returnHiQMessageID(hqDataIn[ndxDev].hdrCommon.H_nMessageID)")
			AMX_LOG(AMX_WARNING,"ITOA(__LINE__),' ',sDebug,' checking flags ',returnHiqMessageFlags(nFlags)")
		 }

																		 // if this is HELLO INFO, not a HELLO QUERY
		 IF (nFlags & HIQ_MID_FLAG_INFORMATION)
		 {
			IF (debugHQN >= DEBUG_VERBOSE)
			{
			 AMX_LOG(AMX_WARNING,"ITOA(__LINE__),' ',sDebug,' HELLO INFO()'")
			 AMX_LOG(AMX_WARNING,"ITOA(__LINE__),' ',sDebug,' checking incoming Payload Session # ',FORMAT('$%04X',hqPH_In[ndxDev].session.a_nSelf)")
			}
																		 // if we've parsed a valid session #
			IF (hqPH_In[ndxDev].session.a_nSelf <> $0000)
			{
																		 // ...update the expected INCOMING payload header session # to that sent OUT previously...
																		 // ...AND update the outgoing common header session # to this payload session #
			 hqDataIn[ndxDev].hdrSession.Q_nOther = hqPH_Out[ndxDev].session.a_nSelf
			 hqDataOut[ndxDev].hdrSession.Q_nOther = hqPH_In[ndxDev].session.a_nSelf
			 hqDataOut[ndxDev].hdrCommon.I_nFlags = hqDataOut[ndxDev].hdrCommon.I_nFlags | HIQ_MID_FLAG_SESSION_NUMBER
			 IF (debugHQN >= DEBUG_TERSE)
			 {
				AMX_LOG(AMX_WARNING,"ITOA(__LINE__),' ',sDebug,
														' Session #''s confirmed: INCOMING ',FORMAT('$%04X,',hqDataIn[ndxDev].hdrSession.Q_nOther),
																										' OUTGOING ',FORMAT('$%04X',hqDataOut[ndxDev].hdrSession.Q_nOther)")
			 }
																		 // BUT...
			 IF (flagHQN_NoSession[ndxDev])// if a session-less connection is flagged, return an error to the HiQnet device
			 {
				hqDataOut[ndxDev].hdrCommon.I_nFlags = hqDataOut[ndxDev].hdrCommon.I_nFlags & (!HIQ_MID_FLAG_SESSION_NUMBER)
				hqDataOut[ndxDev].hdrCommon.I_nFlags = HIQ_MID_FLAG_ERROR | HIQ_MID_FLAG_GUARANTEED | HIQ_MID_FLAG_INFORMATION
				buildHiQHelloError (hqDataOut[ndxDev],hqDataOut[ndxDev].hdrCommon.I_nFlags,sOutgoing,__LINE__,"ipHiQNet[ndxDev].hostname,' <-'")
				SEND_STRING dvHQN[ndxDev],sOutgoing
			 }
			}
			ELSE IF (nFlags & HIQ_MID_FLAG_ERROR == HIQ_MID_FLAG_ERROR)
			{
			 // assume the session # is not supported? ...suspect more eval may be needed
			 hqDataOut[ndxDev].hdrCommon.I_nFlags = hqDataOut[ndxDev].hdrCommon.I_nFlags & (!HIQ_MID_FLAG_SESSION_NUMBER)
			 IF (debugHQN >= DEBUG_TERSE)
			 {
				AMX_LOG(AMX_ERROR,"ITOA(__LINE__),' ',
													sDebug,
													' hqDO[',ITOA(ndxDev),'] SESSION & ERROR ',
													returnHiQMessageFlags(hqDataOut[ndxDev].hdrCommon.I_nFlags & HIQ_MID_FLAG_SESSION_NUMBER)")
			 }
			}
			ELSE IF (hqPH_In[ndxDev].session.a_nSelf == $0000)
			{
			 // assume the session # is not supported? ...suspect more eval may be needed
			 hqDataOut[ndxDev].hdrCommon.I_nFlags = hqDataOut[ndxDev].hdrCommon.I_nFlags & (!HIQ_MID_FLAG_SESSION_NUMBER)
			 IF (debugHQN >= DEBUG_TERSE)
			 {
				AMX_LOG(AMX_WARNING,"ITOA(__LINE__),' ',
														sDebug,
														' hdr session $0000 not valid, session #''s NOT supported for this session'")
			 }
			}
		 }
		 ELSE
		 {
			IF (debugHQN >= DEBUG_VERBOSE)
			{
			 AMX_LOG(AMX_WARNING,"ITOA(__LINE__),' ',sDebug,' HELLO QUERY() EVAL ToDo...'")
			}
		 }
		}
 //	 CASE HIQ_MID_ROUTING_GET_NETWORK_INFO:
 //	 CASE HIQ_MID_ROUTING_REQUEST_ADDRESS:
 //	 CASE HIQ_MID_ROUTING_ADDRESS_USED:
 //	 CASE HIQ_MID_ROUTING_SET_ADDRESS:
 	 CASE HIQ_MID_ROUTING_GOODBYE:
	 {
		flagDiscoMatched[ndxDev] = FALSE
		SEND_STRING 0,"ITOA(__LINE__),' ',sDebug,
									' HiQnet ',ITOA(hqDataIn[ndxDev].hdrCommon.D_nSourceNode),
									' GOODBYE to ',ITOA(hqDataIn[ndxDev].hdrCommon.F_nDestinationNode),',',
									' flagDiscoMatched[',ITOA(ndxDev),'] == FALSE'"
	 }
 //	 
 //	 // Device Methods
		CASE HIQ_MID_DEVICE_GET_ATTRIBUTES:
		{
																		/***************************/
																		/* if this is INFORMATION  */
																		/***************************/
		 IF (hqDataIn[ndxDev].hdrCommon.I_nFlags & HIQ_MID_FLAG_INFORMATION == HIQ_MID_FLAG_INFORMATION)
		 {
			LOCAL_VAR VOLATILE INTEGER nNode
			LOCAL_VAR LONG lVDO
			LOCAL_VAR CHAR cType
			LOCAL_VAR INTEGER nRate
			LOCAL_VAR INTEGER nFlag

		 // how do we know this is the device manager data? Check source dev & VDO...
			#warn'todo: filter attributes eval???'
																		 // if it's an amplifier, subscribe to all non-sensor attributes
			IF ((LEFT_STRING(hqAtr_DM[ndxDev].classname.str,5) == 'Crown') && flagSubscribeAll_after_GA[ndxDev])
			{
			 #warn'this should be cleaned up, made into a subroutine...'
			 flagSubscribeAll_after_GA[ndxDev] = FALSE
			 AMX_LOG(AMX_WARNING,"ITOA(__LINE__),' ',sDebug,' message ID ',returnHiQMessageID(hqDataIn[ndxDev].hdrCommon.H_nMessageID),' ... now sending SubscribeALL...'")
			 hqDataOut[ndxDev].hdrCommon.I_nFlags = hqDataOut[ndxDev].hdrCommon.I_nFlags | HIQ_MID_FLAG_GUARANTEED
			 IF(flagHQN_NoSession[ndxDev])
			 {
				hqDataOut[ndxDev].hdrCommon.I_nFlags = hqDataOut[ndxDev].hdrCommon.I_nFlags & (!HIQ_MID_FLAG_SESSION_NUMBER)
			 }
			 hqDataOut[ndxDev].hdrCommon.B_cLengthHeader = lengthHiQHeaders(hqDataOut[ndxDev].hdrCommon.I_nFlags)
			 hqDataOut[ndxDev].hdrCommon.C_lLengthMessage = hqDataOut[ndxDev].hdrCommon.B_cLengthHeader
																								 // 2 bytes for HiQnet Address
																								 // 4 bytes for HiQnet VD OBject
																								 // 1 bytes for subscribe type bitmap
																								 // 2 bytes for sensor rate
																								 // 2 bytes for flags
																								 + 11
			 hqDataOut[ndxDev].hdrCommon.H_nMessageID = HIQ_MID_MULTISV_SUBSCRIBE_ALL
			 
			 printHiQHeaders(hqDataOut[ndxDev],hqDataOut[ndxDev].hdrSession.Q_nOther,__LINE__,"ipHiQNet[ndxDev].hostname,' <-'")
			 buildHiQHeaders(hqDataOut[ndxDev].hdrCommon,
											 hqDataOut[ndxDev].hdrCommon.I_nFlags,
											 hqDataOut[ndxDev].hdrError.L_nCode,
											 hqDataOut[ndxDev].hdrMultiPart,
											 hqDataOut[ndxDev].hdrSession.Q_nOther,
											 sOutgoing)

			 nNode = hqDataOut[ndxDev].hdrCommon.F_nDestinationNode
			 SEND_STRING 0,"ITOA(__LINE__),FORMAT(' node $%04X',nNode)"
			 lVDo = $00000000
			 SEND_STRING 0,"ITOA(__LINE__),FORMAT(' VDO $%08X',lVDO)"
			 cType = $01 // HIQ_SUB_TYPE_ATTRIBUTES
			 SEND_STRING 0,"ITOA(__LINE__),FORMAT(' type $%02X',cType)"
			 nRate = 50  // this is too short for NX, but is what AA sends so it's here for now...
			 SEND_STRING 0,"ITOA(__LINE__),FORMAT(' sensor rate $%04X',nRate)"
			 nFlag = $00 // don't return initial values
			 SEND_STRING 0,"ITOA(__LINE__),FORMAT(' flag $%04X',nFlag)"
			
			 IF (!cType) { cType = $05 }    // HIQ_SUB_TYPE_SV_NON_SENSOR | HIQ_SUB_TYPE_ATTRIBUTES
			 IF (nRate < 50) { nRate = HIQ_SUB_SENSOR_RATE_NX }
			 sOutgoing = "sOutgoing,
										RAW_BE(nNode),
										RAW_BE(lVDO),
										cType,
										RAW_BE(nRate),
										RAW_BE(nFlag)"
			 printToDebug(0:1:0,"ITOA(__LINE__),
													FORMAT(' $%02X ',ITOA(LENGTH_ARRAY(sOutgoing))),' ',
													ipHiQNet[ndxDev].hostname,' <-',
													'SUBSCRIBE ALL '",
													sOutgoing,131,1,PRINT_HEX)
			 SEND_STRING dvHQN[ndxDev],sOutgoing
			}

			#warn 'filter data??? - need a flag for attribute updated'
			IF (LENGTH_ARRAY(UTF16BE_FORMAT_TP(hqAtr_DM[ndxDev].classname.utf16)) > nDiscoUniMax)
			{
			 SEND_COMMAND dvTPG4,"'^BMF-',ITOA(adrHQnDeviceType[ndxDev]),',0,',
														'%UN',UTF16BE_FORMAT_TP(hqAtr_DM[ndxDev].classname.utf16),
														'%F',sDiscoFontSmallG4"
			 SEND_COMMAND dvTPG5,"'^BMF-',ITOA(adrHQnDeviceType[ndxDev]),',0,',
														'%UN',UTF16BE_FORMAT_TP(hqAtr_DM[ndxDev].classname.utf16),
														'%F',sDiscoFontSmallG5"
			}
			ELSE
			{
			 SEND_COMMAND dvTPG4,"'^BMF-',ITOA(adrHQnDeviceType[ndxDev]),',0,',
														'%UN',UTF16BE_FORMAT_TP(hqAtr_DM[ndxDev].classname.utf16),
														'%F',sDiscoFontLargeG4"
			 SEND_COMMAND dvTPG5,"'^BMF-',ITOA(adrHQnDeviceType[ndxDev]),',0,',
														'%UN',UTF16BE_FORMAT_TP(hqAtr_DM[ndxDev].classname.utf16),
														'%F',sDiscoFontLargeG5"
			}
			IF (LENGTH_ARRAY(UTF16BE_FORMAT_TP(hqAtr_DM[ndxDev].name.utf16)) > nDiscoUniMax)
			{
			 SEND_COMMAND dvTPG4,"'^BMF-',ITOA(adrHQnDeviceName[ndxDev]),',0,',
														'%UN',UTF16BE_FORMAT_TP(hqAtr_DM[ndxDev].name.utf16),
														'%F',sDiscoFontSmallG4"
			 SEND_COMMAND dvTPG5,"'^BMF-',ITOA(adrHQnDeviceName[ndxDev]),',0,',
														'%UN',UTF16BE_FORMAT_TP(hqAtr_DM[ndxDev].name.utf16),
														'%F',sDiscoFontSmallG5"
			}
			ELSE
			{
			 SEND_COMMAND dvTPG4,"'^BMF-',ITOA(adrHQnDeviceName[ndxDev]),',0,',
														'%UN',UTF16BE_FORMAT_TP(hqAtr_DM[ndxDev].name.utf16),
														'%F',sDiscoFontLargeG4"
			 SEND_COMMAND dvTPG5,"'^BMF-',ITOA(adrHQnDeviceName[ndxDev]),',0,',
														'%UN',UTF16BE_FORMAT_TP(hqAtr_DM[ndxDev].name.utf16),
														'%F',sDiscoFontLargeG5"
			}
		 }
		}
 //	 CASE HIQ_MID_DEVICE_GET_VD_LIST:
 //	 CASE HIQ_MID_DEVICE_STORE:
 //	 CASE HIQ_MID_DEVICE_RECALL:
 //	 CASE HIQ_MID_DEVICE_LOCATE:
 //
 //	 // Log Methods
 //	 CASE HIQ_MID_EVENT_LOG_SUBSCRIBE:
 //	 CASE HIQ_MID_EVENT_LOG_UNSUBSCRIBE:
 //	 CASE HIQ_MID_EVENT_LOG_REQUEST:
 //
 //	 // Multi Parameter Methods
 //	 CASE HIQ_MID_MULTISV_SET:
 //	 CASE HIQ_MID_MULTISV_SET_PERCENT:
 //	 CASE HIQ_MID_MULTISV_GET:
 //	 CASE HIQ_MID_MULTISV_SUBSCRIBE:
 //	 CASE HIQ_MID_MULTISV_SUBSCRIBE_PERCENT:
 //	 CASE HIQ_MID_MULTISV_UNSUBSCRIBE:
		DEFAULT:
		{
		 IF (debugHQN >= DEBUG_VERBOSE)
		 {
			AMX_LOG(AMX_WARNING,"ITOA(__LINE__),' ',sDebug,' message ID ',returnHiQMessageID(hqDataIn[ndxDev].hdrCommon.H_nMessageID),' no evaluation'")
		 }
		}
	 }
 //	IF (debugHQN >= DEBUG_VERBOSE)
 //	{
		AMX_LOG(AMX_WARNING,"ITOA(__LINE__),' end DATA_EVENT[',DEV_TO_STRING(DATA.DEVICE),'] STRING: ',DATA.SOURCEIP,' ''',ipHiQNet[ndxDev].hostname,''''")
 //	}
 
 //	IF (TIMELINE_ACTIVE(ndxDev)) // timeline is not paused, so doesn't need to be restarted
 //	{
 //	 TIMELINE_RESTART(ndxDev)
 //	 IF (debugHQN >= DEBUG_VVV) { AMX_LOG(AMX_WARNING,"ITOA(__LINE__),' ',sDebug,' TIMELINE_RESTART(',ITOA(tidHQN[ndxDev]),') ',ITOA(TIMELINE_ACTIVE(tidHQN[ndxDev]))") }
 //	}
	}
 }
 OFFLINE:
 {
	STACK_VAR INTEGER ndxDev
	ndxDev = GET_LAST(dvHQN)
	IF (ndxDev <> ndxSelf)
	{
	 flagHQN_online[ndxDev] = FALSE
	 IF (debugHQN >= DEBUG_TERSE)
	 {
		AMX_LOG(AMX_WARNING,"ITOA(__LINE__),' ',LDATE,' ',TIME,' ',DEV_TO_STRING(DATA.DEVICE),' OFFLINE ''',ipHiQNet[ndxDev].hostname,''''")
	 }
	 IF (TIMELINE_ACTIVE(tidHQN[ndxDev]))
	 {
		TIMELINE_KILL(tidHQN[ndxDev])
		IF (debugHQN >= DEBUG_VVV)
		{
		 AMX_LOG(AMX_WARNING,"ITOA(__LINE__),' ',ipHiQNet[ndxDev].hostname,' TIMELINE_KILL(',ITOA(tidHQN[ndxDev]),') ',ITOA(TIMELINE_ACTIVE(tidHQN[ndxDev]))")
		}
	 }
	}
 }
}


TIMELINE_EVENT[2] TIMELINE_EVENT[3] TIMELINE_EVENT[4] TIMELINE_EVENT[5] TIMELINE_EVENT[6] TIMELINE_EVENT[7]
TIMELINE_EVENT[8] TIMELINE_EVENT[9] TIMELINE_EVENT[10] TIMELINE_EVENT[11] TIMELINE_EVENT[12] TIMELINE_EVENT[13]
TIMELINE_EVENT[14] TIMELINE_EVENT[15] TIMELINE_EVENT[16] TIMELINE_EVENT[17] TIMELINE_EVENT[18] TIMELINE_EVENT[19]
TIMELINE_EVENT[20] TIMELINE_EVENT[21] TIMELINE_EVENT[22] TIMELINE_EVENT[23] TIMELINE_EVENT[24] TIMELINE_EVENT[25]
																		/**********************/
																		/* HiQnet Keep Alives */
																		/**********************/
{
 STACK_VAR CHAR sOutgoing[256]
 STACK_VAR CHAR sDebug[64]
 STACK_VAR INTEGER nFlags
 
 sDebug = "ipHiQNet[TIMELINE.ID].hostname,' <-'"
 
 #warn 'todo?: add older keep alive type when DISCO INFO not supported'
 IF (debugHQN >= DEBUG_VVV)
 {
	AMX_LOG(AMX_WARNING,"ITOA(__LINE__),' ',sDebug,
											' tid ',ITOA(TIMELINE.ID),
											' KAP ',ITOA(TIMELINE.TIME),'ms'")
 }
 IF (debugHQN >= DEBUG_VVVV)
 {
	SEND_STRING 0,"ITOA(__LINE__),' hqDataOut[',TIMELINE.ID,'].hdrCommon.I_nFlags ',returnHiQMessageFlags(hqDataOut[TIMELINE.ID].hdrCommon.I_nFlags)"
 }
 nFlags = (hqDataOut[TIMELINE.ID].hdrCommon.I_nFlags & HIQ_MID_FLAG_SESSION_NUMBER) | HIQ_MID_FLAG_GUARANTEED | HIQ_MID_FLAG_INFORMATION

 IF (debugHQN >= DEBUG_VVVV)
 {
	SEND_STRING 0,"ITOA(__LINE__),' hqDataOut[',TIMELINE.ID,'].hdrCommon.I_nFlags ',returnHiQMessageFlags(nFlags)"
 }
 buildHiQDisco (hqPD_DiscoBC[ndxSelf], hqDataOut[TIMELINE.ID], nFlags, sOutgoing, __LINE__, sDebug)
 SEND_STRING dvHQN[TIMELINE.ID],sOutgoing
}

DEFINE_EVENT                        // HiQnet Broadcast

DATA_EVENT[dvHQnDiscoBC]
{
 ONERROR:
 {
	IF (debugHQN >= DEBUG_TERSE)
	{
	 AMX_LOG(AMX_ERROR,"ITOA(__LINE__),' ',
											LDATE,' ',TIME,' ',
											DEV_TO_STRING(DATA.DEVICE),' ',
											'err-',ITOA(DATA.NUMBER),' ',
											'''',ipHiQDiscoBC.hostname,''''")
	}
	IF (!(DATA.NUMBER == 0 &&         // this happens sometimes, older(?) NetLinx masters
			  DATA.NUMBER == 9 &&         // already closed
			  DATA.NUMBER == 14))         // local port already used
	{
	 flagHQnDiscoBC_Online = FALSE
	}
 }
 OFFLINE:
 {
	flagHQnDiscoBC_Online = FALSE       // expected : this should never happen, as it is UDP
 }
 ONLINE:
 {
	flagHQnDiscoBC_Online = TRUE
	//SEND_COMMAND vdvControl,'DISCO-5' // announce the arrival of this NetLinx master
	//SEND_COMMAND vdvControl,'DISCO-1' // announce the arrival of this NetLinx master
 }
 STRING:
 {
	IF (flagHQnDiscoListener)
	{
	 STACK_VAR _sHiQnet_headers hqDI
	 STACK_VAR sDebug[32]

	 sDebug = "ipHiQDiscoBC.hostname,' ->'"
	 IF (debugHQN >= DEBUG_TERSE)
	 {
		printToDebug(0:1:0,"ITOA(__LINE__),' ',
												FORMAT('$%02X',ITOA(LENGTH_ARRAY(DATA.TEXT))),' ',
												sDebug,' '"
												,DATA.TEXT,131,0,PRINT_HEX)
	 }
	 parseHiQHeaders(hqDI,DATA.TEXT,__LINE__,sDebug)
	 IF (debugHQN >= DEBUG_TERSE)
	 {
		printHiQHeaders(hqDI,hqDI.hdrSession.Q_nIncoming,__LINE__,sDebug)
	 }
	 
	 SWITCH (hqDI.hdrCommon.H_nMessageID)
	 {
		CASE HIQ_MID_ROUTING_DISCO_INFO:
		{
		 IF (hqDI.hdrCommon.H_nMessageID == HIQ_MID_ROUTING_DISCO_INFO)
		 {
			STACK_VAR _sHiQNet_DISCO_PAYLOAD hqPD
			STACK_VAR _sNetLinx_HiQNet_control_info hqIP

			LOCAL_VAR VOLATILE INTEGER i
			
			parseHiQPayloadDisco(hqPD,DATA.TEXT,__LINE__,sDebug)
			IF (debugHQN >= DEBUG_TERSE)
			{
			 printHiQPayloadDisco(hqPD,__LINE__,sDebug)
			}
			
			hqIP.ipAddress = LONG_TO_IPV4(hqPD.network.k_lIPAddress)
			hqIP.node = ITOA(hqPD.a_nSourceNode)

			// Keep It Simple match criteria:
			// hqPD_DvMgr[].a_nSourceNode - HiQnet Node
			// hqPD_DvMgr[].network.k_lIPAddress - IP Address
			// First look at the hqPD_DiscoBC data...
			FOR (i = LENGTH_ARRAY(hqPD_DiscoBC); i; i--)
			{
			 IF ((hqPD.a_nSourceNode == hqPD_DiscoBC[i].a_nSourceNode) &&
					 (hqPD.network.k_lIPAddress == hqPD_DiscoBC[i].network.k_lIPAddress))
			 {
				flagDiscoMatched[i] = TRUE
				SEND_COMMAND dvTP,"'^TXT-',ITOA(adrDiscoRowNumber[i]),',0,',ITOA(i)"
				AMX_LOG(AMX_ERROR,"ITOA(__LINE__),
													' ************************ DISCO MATCH ',ITOA(i),' ',
													ITOA(hqPD.a_nSourceNode),' ',LONG_TO_IPV4(hqPD.network.k_lIPAddress),' '")
				
				SEND_COMMAND dvTP,"'^TXT-',ITOA(adrHQnDeviceType[i]),',0,',hqAtr_DM[i].classname.str"
				SEND_COMMAND dvTP,"'^TXT-',ITOA(adrHQnDeviceNode[i]),',0,',FORMAT('$%04X',hqPD.a_nSourceNode),' (',hqIP.node,')'"
				SEND_COMMAND dvTP,"'^TXT-',ITOA(adrHQn_IPAddress[i]),',0,',hqIP.ipAddress"
				SEND_COMMAND dvTP,"'^TXT-',ITOA(adrHQnDeviceName[i]),',0,',hqAtr_DM[i].name.str"
				BREAK
			 }
			}
			IF (!i)                          // no match, new device
			{
			 IF (LENGTH_ARRAY(hqPD_DiscoBC) < MAX_LENGTH_ARRAY(hqPD_DiscoBC))
			 {
				i = LENGTH_ARRAY(hqPD_DiscoBC) + 1
				hqPD_DiscoBC[i] = hqPD
				SET_LENGTH_ARRAY(hqPD_DiscoBC,i)
				AMX_LOG(AMX_ERROR,"ITOA(__LINE__),
													' ************************ DISCO NEW ROW ',ITOA(i),' ',
													ITOA(hqPD.a_nSourceNode),' ',LONG_TO_IPV4(hqPD.network.k_lIPAddress),' '")
				flagDiscoMatched[i] = TRUE
				SEND_COMMAND dvTP,"'^TXT-',ITOA(adrDiscoRowNumber[i]),',0,',ITOA(i)"
				
				SEND_COMMAND dvTP,"'^TXT-',ITOA(adrHQnDeviceType[i]),',0,'"
				SEND_COMMAND dvTP,"'^TXT-',ITOA(adrHQnDeviceNode[i]),',0,',FORMAT('$%04X',hqPD.a_nSourceNode),' (',hqIP.node,')'"
				SEND_COMMAND dvTP,"'^TXT-',ITOA(adrHQn_IPAddress[i]),',0,',hqIP.ipAddress"
				SEND_COMMAND dvTP,"'^TXT-',ITOA(adrHQnDeviceName[i]),',0,'"
			 }
			 ELSE
			 {
				SEND_STRING 0,"ITOA(__LINE__),' hqPD_DiscoBC is already FULL'"
			 }
			}
																		// if this was a DISCO INFO QUERY
			IF (hqDI.hdrCommon.I_nFlags & HIQ_MID_FLAG_INFORMATION <> HIQ_MID_FLAG_INFORMATION)
			{
			 IF (i <= LENGTH_ARRAY(hqPD_DiscoBC))
			 {
				IF (hqPD.a_nSourceNode <> nodeSelf)
				{
				 // 1. Respond to DISCO QUERY with Disco Info
				 // 2. what does AA/Netsetter do?
				 
				 LOCAL_VAR VOLATILE CHAR sOutgoing[256]
				 
				 sDebug = "ITOA(__LINE__),' ',ITOA(hqPD.a_nSourceNode),' <-'"
				 
																				// updateSession # for initial Hello Query
				 hqPH_Out[i].session.a_nSelf++ // session number from 1 to 65535, increments each new session
				 IF (!hqPH_Out[i].session.a_nSelf) { hqPH_Out[i].session.a_nSelf++ }
				 
																					 // set sequence to 0 here, it will be incremented when the strings are built
				 hqDataOut[i].hdrCommon.K_nSequence = 0
																					 // flag system to send a Subscribe All after the initial GET ATTRIBUTES is received.
				 //flagSubscribeAll_after_GA[i] = TRUE
				 
																					 // info flag set for response DISCO
				 buildHiQDisco(hqPD_DiscoBC[ndxSelf],hqDataOut[i],HIQ_MID_FLAG_GUARANTEED ^ HIQ_MID_FLAG_INFORMATION,sOutgoing,__LINE__,sDebug)
				 #warn 'where to send this string???'
				}
			 }
			}
		 }
		}
		CASE HIQ_MID_ROUTING_GOODBYE:
		{
		 STACK_VAR INTEGER i
		 FOR (i = LENGTH_ARRAY(hqPD_DiscoBC); i; i--)
		 {
			IF (hqDI.hdrCommon.D_nSourceNode == hqPD_DiscoBC[i].a_nSourceNode)
			{
			 flagDiscoMatched[i] = FALSE
			 IF (debugHQN >= DEBUG_VERBOSE)
			 {
				SEND_STRING 0,"ITOA(__LINE__),' ',sDebug,
									' HiQnet ',ITOA(hqDI.hdrCommon.D_nSourceNode),
									' BROADCAST GOODBYE to ',ITOA(hqDI.hdrCommon.F_nDestinationNode),',',
									' flagDiscoMatched[',ITOA(i),'] == FALSE'"
			 }
			 BREAK
			}
		 }
		}
		DEFAULT:
		{
		 IF (debugHQN >= DEBUG_TERSE)
		 {
			SEND_STRING 0,"ITOA(__LINE__),
									 ' UNHANDLED BROADCAST ',
									 returnHiQMessageID(hqDI.hdrCommon.H_nMessageID),' ',
									 returnHiQMessageFlags(hqDI.hdrCommon.I_nFlags)"
		 }
		}
	 }
	}
 }
}

TIMELINE_EVENT[tidHQnDiscoBC]       // HiQnet Disco
{
 LOCAL_VAR _sHiQnet_headers hqDataDiscoBC
 STACK_VAR CHAR sOutGoing[256]
 STACK_VAR INTEGER i
 hqDataDiscoBC.hdrCommon.D_nSourceNode = nodeSelf
 hqDataDiscoBC.hdrCommon.H_nMessageID = HIQ_MID_ROUTING_DISCO_INFO
 hqDataDiscoBC.hdrCommon.F_nDestinationNode = $FFFF
 buildHiQDisco(hqPD_DiscoBC[ndxSelf],hqDataDiscoBC,HIQ_MID_FLAG_GUARANTEED,sOutgoing,__LINE__,"ipHiQDiscoBC.hostname,' <-'")
 SEND_STRING dvHQnDiscoBC,sOutgoing
 OFF[dvTP,7]                        // blink button once
 rowTP = 1
 
																	  // clear the rows of any unsaved discoveries
 SET_LENGTH_ARRAY(hqPD_DiscoBC,LENGTH_ARRAY(ipHiQNet))
 FOR (i = LENGTH_ARRAY(ipHiQNet) + 1; i <= MAX_LENGTH_ARRAY(ipHiQNet); i++ )
 {
	SEND_COMMAND dvTP,"'^TXT-',ITOA(adrDiscoRowNumber[i]),',0,'"
	
	SEND_COMMAND dvTP,"'^TXT-',ITOA(adrHQnDeviceType[i]),',0,'"
	SEND_COMMAND dvTP,"'^TXT-',ITOA(adrHQnDeviceNode[i]),',0,'"
	SEND_COMMAND dvTP,"'^TXT-',ITOA(adrHQn_IPAddress[i]),',0,'"
	SEND_COMMAND dvTP,"'^TXT-',ITOA(adrHQnDeviceName[i]),',0,'"
	flagDiscoRowSelected[i] = FALSE
	OFF[dvTP,btnDiscoSelectCol1[i]]   // fast feedback here
	OFF[dvTP,btnDiscoSelectCol2[i]]   // fast feedback here
	OFF[dvTP,btnDiscoSelectCol3[i]]   // fast feedback here
	OFF[dvTP,btnDiscoSelectCol4[i]]   // fast feedback here
 }
 FOR (i = MAX_LENGTH_ARRAY(flagDiscoMatched); i; i--)
 {
	flagDiscoMatched[i] = FALSE
	OFF[dvTP,btnDiscoMatched[i]]      // fast feedback here
 }
 
 nHQnDiscoBC--
 SEND_COMMAND dvTP,"'^TXT-7,0,HiQnet Disco BC-',ITOA(nHQnDiscoBC)"
 IF (!nHQnDiscoBC)
 {
	TIMELINE_KILL(TIMELINE.ID)
	 SEND_COMMAND dvTP,"'^TXT-7,0,HiQnet Disco Broadcast'"
 }
}

DEFINE_EVENT                        // TP

DATA_EVENT[dvTP1]
{
 ONLINE:
 {
	LOCAL_VAR VOLATILE INTEGER i
	WAIT 60
	{
	 FOR (i = 1; i <= LENGTH_ARRAY(hqPD_DiscoBC); i++)
	 {
		SEND_COMMAND dvTP1,"'^TXT-',ITOA(adrDiscoRowNumber[i]),',0,',ITOA(i)"
		SEND_COMMAND dvTP1,"'^TXT-',ITOA(adrHQnDeviceType[i]),',0,',hqAtr_DM[i].classname.str"
		SEND_COMMAND dvTP1,"'^TXT-',ITOA(adrHQnDeviceNode[i]),',0,',
												FORMAT('$%04X',hqPD_DiscoBC[i].a_nSourceNode),
												' (',ITOA(hqPD_DiscoBC[i].a_nSourceNode),')'"
		SEND_COMMAND dvTP1,"'^TXT-',ITOA(adrHQn_IPAddress[i]),',0,',ipHiQNet[i].ipAddress"
		SEND_COMMAND dvTP1,"'^TXT-',ITOA(adrHQnDeviceName[i]),',0,',hqAtr_DM[i].name.str"
	 }
	 FOR (; i <= LENGTH_ARRAY(btnDiscoMatched); i++)
	 {
		SEND_COMMAND dvTP1,"'^TXT-',ITOA(adrDiscoRowNumber[i]),',0,'"
		SEND_COMMAND dvTP1,"'^TXT-',ITOA(adrHQnDeviceType[i]),',0,'"
		SEND_COMMAND dvTP1,"'^TXT-',ITOA(adrHQnDeviceNode[i]),',0,'"
		SEND_COMMAND dvTP1,"'^TXT-',ITOA(adrHQn_IPAddress[i]),',0,'"
		SEND_COMMAND dvTP1,"'^TXT-',ITOA(adrHQnDeviceName[i]),',0,'"
		OFF[flagHQN_online[i]]
	 }
	 FOR (i = 1; i <= LENGTH_ARRAY(hqPD_DvMgr); i++)
	 {
		SEND_COMMAND dvTP1,"'^TXT-',ITOA(adrHQnSavedRowNumber[i]),',0,',ITOA(i)"
	 }
	 FOR (; i <= LENGTH_ARRAY(btnHQnOnline); i++)
	 {
		SEND_COMMAND dvTP1,"'^TXT-',ITOA(adrHQnSavedRowNumber[i]),',0,'"
	 }
	}
 }
}

DATA_EVENT[dvTP2]
{
 ONLINE:
 {
	LOCAL_VAR VOLATILE INTEGER i
	WAIT 60
	{
	 FOR (i = 1; i <= LENGTH_ARRAY(hqPD_DiscoBC); i++)
	 {
		SEND_COMMAND dvTP2,"'^TXT-',ITOA(adrDiscoRowNumber[i]),',0,',ITOA(i)"
		SEND_COMMAND dvTP2,"'^TXT-',ITOA(adrHQnDeviceType[i]),',0,',hqAtr_DM[i].classname.str"
		SEND_COMMAND dvTP2,"'^TXT-',ITOA(adrHQnDeviceNode[i]),',0,',
												FORMAT('$%04X',hqPD_DiscoBC[i].a_nSourceNode),
												' (',ITOA(hqPD_DiscoBC[i].a_nSourceNode),')'"
		SEND_COMMAND dvTP2,"'^TXT-',ITOA(adrHQn_IPAddress[i]),',0,',ipHiQNet[i].ipAddress"
		SEND_COMMAND dvTP2,"'^TXT-',ITOA(adrHQnDeviceName[i]),',0,',hqAtr_DM[i].name.str"
	 }
	 FOR (; i <= LENGTH_ARRAY(btnDiscoMatched); i++)
	 {
		SEND_COMMAND dvTP2,"'^TXT-',ITOA(adrDiscoRowNumber[i]),',0,'"
		SEND_COMMAND dvTP2,"'^TXT-',ITOA(adrHQnDeviceType[i]),',0,'"
		SEND_COMMAND dvTP2,"'^TXT-',ITOA(adrHQnDeviceNode[i]),',0,'"
		SEND_COMMAND dvTP2,"'^TXT-',ITOA(adrHQn_IPAddress[i]),',0,'"
		SEND_COMMAND dvTP2,"'^TXT-',ITOA(adrHQnDeviceName[i]),',0,'"
		OFF[flagHQN_online[i]]
	 }
 	 FOR (i = 1; i <= LENGTH_ARRAY(hqPD_DvMgr); i++)
	 {
		SEND_COMMAND dvTP2,"'^TXT-',ITOA(adrHQnSavedRowNumber[i]),',0,',ITOA(i)"
	 }
	 FOR (; i <= LENGTH_ARRAY(btnHQnOnline); i++)
	 {
		SEND_COMMAND dvTP2,"'^TXT-',ITOA(adrHQnSavedRowNumber[i]),',0,'"
	 }
	}
 }
}

DATA_EVENT[dvTP3]
{
 ONLINE:
 {
	LOCAL_VAR VOLATILE INTEGER i
	WAIT 60
	{
	 FOR (i = 1; i <= LENGTH_ARRAY(hqPD_DiscoBC); i++)
	 {
		SEND_COMMAND dvTP3,"'^TXT-',ITOA(adrDiscoRowNumber[i]),',0,',ITOA(i)"
		SEND_COMMAND dvTP3,"'^TXT-',ITOA(adrHQnDeviceType[i]),',0,',hqAtr_DM[i].classname.str"
		SEND_COMMAND dvTP3,"'^TXT-',ITOA(adrHQnDeviceNode[i]),',0,',
												FORMAT('$%04X',hqPD_DiscoBC[i].a_nSourceNode),
												' (',ITOA(hqPD_DiscoBC[i].a_nSourceNode),')'"
		SEND_COMMAND dvTP3,"'^TXT-',ITOA(adrHQn_IPAddress[i]),',0,',ipHiQNet[i].ipAddress"
		SEND_COMMAND dvTP3,"'^TXT-',ITOA(adrHQnDeviceName[i]),',0,',hqAtr_DM[i].name.str"
	 }
	 FOR (; i <= LENGTH_ARRAY(btnDiscoMatched); i++)
	 {
		SEND_COMMAND dvTP3,"'^TXT-',ITOA(adrDiscoRowNumber[i]),',0,'"
		SEND_COMMAND dvTP3,"'^TXT-',ITOA(adrHQnDeviceType[i]),',0,'"
		SEND_COMMAND dvTP3,"'^TXT-',ITOA(adrHQnDeviceNode[i]),',0,'"
		SEND_COMMAND dvTP3,"'^TXT-',ITOA(adrHQn_IPAddress[i]),',0,'"
		SEND_COMMAND dvTP3,"'^TXT-',ITOA(adrHQnDeviceName[i]),',0,'"
		OFF[flagHQN_online[i]]
	 }
	}
	 FOR (i = 1; i <= LENGTH_ARRAY(hqPD_DvMgr); i++)
	 {
		SEND_COMMAND dvTP3,"'^TXT-',ITOA(adrHQnSavedRowNumber[i]),',0,',ITOA(i)"
	 }
	 FOR (; i <= LENGTH_ARRAY(btnHQnOnline); i++)
	 {
		SEND_COMMAND dvTP3,"'^TXT-',ITOA(adrHQnSavedRowNumber[i]),',0,'"
	 }
 }
}

DATA_EVENT[dvTP]
{
 ONLINE:
 {
	STACK_VAR INTEGER i
	STACK_VAR INTEGER j
																		// check if panel is G5
	FOR (i = LENGTH_ARRAY(DEVICE_ID_G5); i; i--)
	{
	 IF (DEVICE_ID_G5[i] = DEVICE_ID(DATA.DEVICE))
	 {
		FOR (j = 1; j <= LENGTH_ARRAY(dvTPG5); j++)
		{
		 IF(dvTPG5[j] = DATA.DEVICE)
		 {
			BREAK
		 }
		}
		IF (j > LENGTH_ARRAY(dvTPG5))   // ID is G5 but no device match, then add it to dvTPG5
		{
		 dvTPG5[j] = DATA.DEVICE
		 SET_LENGTH_ARRAY(dvTPG5,j)
SEND_STRING 0,"ITOA(__LINE__),' dvTPG5 ',ITOA(j)"
		 // REBUILD_EVENT() // dvTP5 only used in SEND_COMMANDs, not to trigger events
		}
		BREAK
	 }
	}
	IF (!i)                           // assume it's a G4 panel
	{
	 FOR (j = 1; j <= LENGTH_ARRAY(dvTPG4); j++)
	 {
		IF(dvTPG4[j] = DATA.DEVICE)
		{
		 BREAK
		}
	 }
	 IF (j > LENGTH_ARRAY(dvTPG4))   // no device match, then add it to dvTPG4
	 {
		dvTPG4[j] = DATA.DEVICE
		SET_LENGTH_ARRAY(dvTPG4,j)
SEND_STRING 0,"ITOA(__LINE__),' dvTPG4 ',ITOA(j)"
		// REBUILD_EVENT() // dvTP4 only used in SEND_COMMANDs, not to trigger events
	 }
	}
 }
}

BUTTON_EVENT[dvTP,7]                // Disco Broadcast
{
 PUSH:
 {
	CANCEL_WAIT 'disco announce'
	nHQnDiscoBC++
	AMX_LOG(AMX_DEBUG,"ITOA(__LINE__),' PUSH',DEVCHAN_TO_STRING(BUTTON.INPUT)")
	IF (nHQnDiscoBC)
	{
	 //SEND_COMMAND dvTP,"'^TXT-7,0,',returnAsciiUTF16BE('HiQnet Disco WAIT...')"
	 SEND_COMMAND dvTP,"'^TXT-7,0,HiQnet Disco BC-',ITOA(nHQnDiscoBC)"
	}
	ELSE
	{
	 SEND_COMMAND dvTP,"'^TXT-7,0,HiQnet Disco Broadcast'"
	 IF (TIMELINE_ACTIVE(tidHQnDiscoBC))
	 {
		TIMELINE_KILL(tidHQnDiscoBC)
	 }
	}
 }
 RELEASE:
 {
	AMX_LOG(AMX_DEBUG,"ITOA(__LINE__),' RELEASE',DEVCHAN_TO_STRING(BUTTON.INPUT)")

	IF (nHQnDiscoBC)
	{
	 LOCAL_VAR VOLATILE INTEGER i
	 WAIT 10 'disco announce'
	 {
																			// setup for 5 disco announcements
		// IF (nHQnDiscoBC) { nHQnDiscoBC = 5 }
		SEND_COMMAND vdvControl,"'DISCO-',ITOA(nHQnDiscoBC)"
		//SEND_COMMAND dvTP,"'^TXT-7,0,HiQnet Disco BC-',ITOA(nHQnDiscoBC)"
	  SEND_COMMAND dvTP,"'^TXT-7,0,HiQnet Disco WAIT...'"
		OFF[dvTP,7]                      // blink button once
	 }
	}
	ELSE
	{
	 CANCEL_WAIT 'disco announce'
	}
 }
}

BUTTON_EVENT[dvTP,8]                // Disco Listener
{
 PUSH:
 {
	flagHQnDiscoListener = !flagHQnDiscoListener
 }
}

BUTTON_EVENT[dvTP,btnDiscoSelectCol1]
{
 PUSH:
 {
	STACK_VAR INTEGER ndxBtn
	ndxBtn = GET_LAST(btnDiscoSelectCol1)
	IF (ndxBtn <> ndxSelf)            // can't select NX Master
	{
																		// if there is a HiQnet device in this row
	 IF (ndxBtn <= LENGTH_ARRAY(hqPD_DiscoBC))
	 {
		flagDiscoRowSelected[ndxBtn] = !flagDiscoRowSelected[ndxBtn]
	 }
	}
 }
}
BUTTON_EVENT[dvTP,btnDiscoSelectCol2]
{
 PUSH:
 {
	STACK_VAR INTEGER ndxBtn
	ndxBtn = GET_LAST(btnDiscoSelectCol2)
	IF (ndxBtn <> ndxSelf)            // can't select NX Master
	{
																		// if there is a HiQnet device in this row
	 IF (ndxBtn <= LENGTH_ARRAY(hqPD_DiscoBC))
	 {
		flagDiscoRowSelected[ndxBtn] = !flagDiscoRowSelected[ndxBtn]
	 }
	}
 }
}
BUTTON_EVENT[dvTP,btnDiscoSelectCol3]
{
 PUSH:
 {
	STACK_VAR INTEGER ndxBtn
	ndxBtn = GET_LAST(btnDiscoSelectCol3)
	IF (ndxBtn <> ndxSelf)            // can't select NX Master
	{
																		// if there is a HiQnet device in this row
	 IF (ndxBtn <= LENGTH_ARRAY(hqPD_DiscoBC))
	 {
		flagDiscoRowSelected[ndxBtn] = !flagDiscoRowSelected[ndxBtn]
	 }
	}
 }
}
BUTTON_EVENT[dvTP,btnDiscoSelectCol4]
{
 PUSH:
 {
	STACK_VAR INTEGER ndxBtn
	ndxBtn = GET_LAST(btnDiscoSelectCol4)
	IF (ndxBtn <> ndxSelf)            // can't select NX Master
	{
																		// if there is a HiQnet device in this row
	 IF (ndxBtn <= LENGTH_ARRAY(hqPD_DiscoBC))
	 {
		flagDiscoRowSelected[ndxBtn] = !flagDiscoRowSelected[ndxBtn]
	 }
	}
 }
}

BUTTON_EVENT[dvTP,1]                // CONNECT
{
 PUSH:
 {
	STACK_VAR INTEGER i
	STACK_VAR CHAR sNodes[252]
	AMX_LOG(AMX_ERROR,"ITOA(__LINE__),' PUSH',DEVCHAN_TO_STRING(BUTTON.INPUT)")
	FOR (i = MAX_LENGTH_ARRAY(flagDiscoRowSelected); i; i--)
	{
	 IF (flagDiscoRowSelected[i])
	 {
SEND_STRING 0,"ITOA(__LINE__),' flagDiscoRowSelected[',ITOA(i),'] ',ITOA(hqPD_DiscoBC[i].a_nSourceNode)"
		sNodes = "sNodes,ITOA(hqPD_DiscoBC[i].a_nSourceNode),','"
	 }
	}
	IF (LENGTH_ARRAY(sNodes))
	{
	 SET_LENGTH_ARRAY(sNodes,LENGTH_ARRAY(sNodes)-1) // remove trailing comma
	 AMX_LOG(AMX_ERROR,"ITOA(__LINE__),' ',DEV_TO_STRING(vdvControl),' <- ',sNodes,' HELLO'")
	 SEND_COMMAND vdvControl,"sNodes,' HELLO'"
	}
 }
}

BUTTON_EVENT[dvTP,2]                // DISCONNECT
{
 PUSH:
 {
	STACK_VAR INTEGER i
	STACK_VAR CHAR sNodes[252]
	AMX_LOG(AMX_ERROR,"ITOA(__LINE__),' PUSH',DEVCHAN_TO_STRING(BUTTON.INPUT)")
	FOR (i = MAX_LENGTH_ARRAY(flagDiscoRowSelected); i; i--)
	{
	 IF (flagDiscoRowSelected[i])
	 {
		sNodes = "sNodes,ITOA(hqPD_DiscoBC[i].a_nSourceNode),','"
	 }
	}
	IF (LENGTH_ARRAY(sNodes))
	{
	 SET_LENGTH_ARRAY(sNodes,LENGTH_ARRAY(sNodes)-1) // remove trailing comma
	 AMX_LOG(AMX_ERROR,"ITOA(__LINE__),' ',DEV_TO_STRING(vdvControl),' <- ',sNodes,' GOODBYE'")
	 SEND_COMMAND vdvControl,"sNodes,' GOODBYE'"
	}
 }
}

BUTTON_EVENT[dvTP,109]              // GET ATTRIBUTES
{
 PUSH:
 {
	SEND_STRING 0,"ITOA(__LINE__),' PUSH',DEVCHAN_TO_STRING(BUTTON.INPUT)"
	SEND_STRING 0,"ITOA(__LINE__),' ',DEV_TO_STRING(vdvControl),' <- GA-1,0'"
	SEND_COMMAND vdvControl,"'GA-1,0'"
 }
}


(***********************************************************)
(*               OPTIONAL OLD CODE GOES BELOW              *)
(***********************************************************)
DEFINE_PROGRAM                      // GET MAC - loopback connection close

IF (flagsNX_GET_MAC & $2 == $2)     // if dvSelf is ONLINE
{
 WAIT 50 'GET MAC timeout'
 {
	IP_CLIENT_CLOSE(dvSelf.PORT)
 }
}
ELSE
{
 CANCEL_WAIT 'GET MAC timeout'
}

DEFINE_PROGRAM                      // button feedback - slow

WAIT 1 'blink'
{
 [dvTP,7] = nHQnDiscoBC
}

[dvTP,8] = flagHQnDiscoListener

[dvTP,btnDiscoSelectCol1[1]]  = flagDiscoRowSelected[1]
[dvTP,btnDiscoSelectCol1[2]]  = flagDiscoRowSelected[2]
[dvTP,btnDiscoSelectCol1[3]]  = flagDiscoRowSelected[3]
[dvTP,btnDiscoSelectCol1[4]]  = flagDiscoRowSelected[4]
[dvTP,btnDiscoSelectCol1[5]]  = flagDiscoRowSelected[5]
[dvTP,btnDiscoSelectCol1[6]]  = flagDiscoRowSelected[6]
[dvTP,btnDiscoSelectCol1[7]]  = flagDiscoRowSelected[7]
[dvTP,btnDiscoSelectCol1[8]]  = flagDiscoRowSelected[8]
[dvTP,btnDiscoSelectCol1[9]]  = flagDiscoRowSelected[9]
[dvTP,btnDiscoSelectCol1[10]] = flagDiscoRowSelected[10]
[dvTP,btnDiscoSelectCol1[11]] = flagDiscoRowSelected[11]
[dvTP,btnDiscoSelectCol1[12]] = flagDiscoRowSelected[12]
[dvTP,btnDiscoSelectCol1[13]] = flagDiscoRowSelected[13]
[dvTP,btnDiscoSelectCol1[14]] = flagDiscoRowSelected[14]
[dvTP,btnDiscoSelectCol1[15]] = flagDiscoRowSelected[15]
[dvTP,btnDiscoSelectCol1[16]] = flagDiscoRowSelected[16]
[dvTP,btnDiscoSelectCol1[17]] = flagDiscoRowSelected[17]
[dvTP,btnDiscoSelectCol1[18]] = flagDiscoRowSelected[18]
[dvTP,btnDiscoSelectCol1[19]] = flagDiscoRowSelected[19]
[dvTP,btnDiscoSelectCol1[20]] = flagDiscoRowSelected[20]
[dvTP,btnDiscoSelectCol1[21]] = flagDiscoRowSelected[21]
[dvTP,btnDiscoSelectCol1[22]] = flagDiscoRowSelected[22]
[dvTP,btnDiscoSelectCol1[23]] = flagDiscoRowSelected[23]
[dvTP,btnDiscoSelectCol1[24]] = flagDiscoRowSelected[24]
[dvTP,btnDiscoSelectCol1[25]] = flagDiscoRowSelected[25]

[dvTP,btnDiscoSelectCol2[1]]  = flagDiscoRowSelected[1]
[dvTP,btnDiscoSelectCol2[2]]  = flagDiscoRowSelected[2]
[dvTP,btnDiscoSelectCol2[3]]  = flagDiscoRowSelected[3]
[dvTP,btnDiscoSelectCol2[4]]  = flagDiscoRowSelected[4]
[dvTP,btnDiscoSelectCol2[5]]  = flagDiscoRowSelected[5]
[dvTP,btnDiscoSelectCol2[6]]  = flagDiscoRowSelected[6]
[dvTP,btnDiscoSelectCol2[7]]  = flagDiscoRowSelected[7]
[dvTP,btnDiscoSelectCol2[8]]  = flagDiscoRowSelected[8]
[dvTP,btnDiscoSelectCol2[9]]  = flagDiscoRowSelected[9]
[dvTP,btnDiscoSelectCol2[10]] = flagDiscoRowSelected[10]
[dvTP,btnDiscoSelectCol2[11]] = flagDiscoRowSelected[11]
[dvTP,btnDiscoSelectCol2[12]] = flagDiscoRowSelected[12]
[dvTP,btnDiscoSelectCol2[13]] = flagDiscoRowSelected[13]
[dvTP,btnDiscoSelectCol2[14]] = flagDiscoRowSelected[14]
[dvTP,btnDiscoSelectCol2[15]] = flagDiscoRowSelected[15]
[dvTP,btnDiscoSelectCol2[16]] = flagDiscoRowSelected[16]
[dvTP,btnDiscoSelectCol2[17]] = flagDiscoRowSelected[17]
[dvTP,btnDiscoSelectCol2[18]] = flagDiscoRowSelected[18]
[dvTP,btnDiscoSelectCol2[19]] = flagDiscoRowSelected[19]
[dvTP,btnDiscoSelectCol2[20]] = flagDiscoRowSelected[20]
[dvTP,btnDiscoSelectCol2[21]] = flagDiscoRowSelected[21]
[dvTP,btnDiscoSelectCol2[22]] = flagDiscoRowSelected[22]
[dvTP,btnDiscoSelectCol2[23]] = flagDiscoRowSelected[23]
[dvTP,btnDiscoSelectCol2[25]] = flagDiscoRowSelected[25]

[dvTP,btnDiscoSelectCol3[1]]  = flagDiscoRowSelected[1]
[dvTP,btnDiscoSelectCol3[2]]  = flagDiscoRowSelected[2]
[dvTP,btnDiscoSelectCol3[3]]  = flagDiscoRowSelected[3]
[dvTP,btnDiscoSelectCol3[4]]  = flagDiscoRowSelected[4]
[dvTP,btnDiscoSelectCol3[5]]  = flagDiscoRowSelected[5]
[dvTP,btnDiscoSelectCol3[6]]  = flagDiscoRowSelected[6]
[dvTP,btnDiscoSelectCol3[7]]  = flagDiscoRowSelected[7]
[dvTP,btnDiscoSelectCol3[8]]  = flagDiscoRowSelected[8]
[dvTP,btnDiscoSelectCol3[9]]  = flagDiscoRowSelected[9]
[dvTP,btnDiscoSelectCol3[10]] = flagDiscoRowSelected[10]
[dvTP,btnDiscoSelectCol3[11]] = flagDiscoRowSelected[11]
[dvTP,btnDiscoSelectCol3[12]] = flagDiscoRowSelected[12]
[dvTP,btnDiscoSelectCol3[13]] = flagDiscoRowSelected[13]
[dvTP,btnDiscoSelectCol3[14]] = flagDiscoRowSelected[14]
[dvTP,btnDiscoSelectCol3[15]] = flagDiscoRowSelected[15]
[dvTP,btnDiscoSelectCol3[16]] = flagDiscoRowSelected[16]
[dvTP,btnDiscoSelectCol3[17]] = flagDiscoRowSelected[17]
[dvTP,btnDiscoSelectCol3[18]] = flagDiscoRowSelected[18]
[dvTP,btnDiscoSelectCol3[19]] = flagDiscoRowSelected[19]
[dvTP,btnDiscoSelectCol3[20]] = flagDiscoRowSelected[20]
[dvTP,btnDiscoSelectCol3[21]] = flagDiscoRowSelected[21]
[dvTP,btnDiscoSelectCol3[22]] = flagDiscoRowSelected[22]
[dvTP,btnDiscoSelectCol3[23]] = flagDiscoRowSelected[23]
[dvTP,btnDiscoSelectCol3[24]] = flagDiscoRowSelected[24]
[dvTP,btnDiscoSelectCol3[25]] = flagDiscoRowSelected[25]

[dvTP,btnDiscoSelectCol4[1]]  = flagDiscoRowSelected[1]
[dvTP,btnDiscoSelectCol4[2]]  = flagDiscoRowSelected[2]
[dvTP,btnDiscoSelectCol4[3]]  = flagDiscoRowSelected[3]
[dvTP,btnDiscoSelectCol4[4]]  = flagDiscoRowSelected[4]
[dvTP,btnDiscoSelectCol4[5]]  = flagDiscoRowSelected[5]
[dvTP,btnDiscoSelectCol4[6]]  = flagDiscoRowSelected[6]
[dvTP,btnDiscoSelectCol4[7]]  = flagDiscoRowSelected[7]
[dvTP,btnDiscoSelectCol4[8]]  = flagDiscoRowSelected[8]
[dvTP,btnDiscoSelectCol4[9]]  = flagDiscoRowSelected[9]
[dvTP,btnDiscoSelectCol4[10]] = flagDiscoRowSelected[10]
[dvTP,btnDiscoSelectCol4[11]] = flagDiscoRowSelected[11]
[dvTP,btnDiscoSelectCol4[12]] = flagDiscoRowSelected[12]
[dvTP,btnDiscoSelectCol4[13]] = flagDiscoRowSelected[13]
[dvTP,btnDiscoSelectCol4[14]] = flagDiscoRowSelected[14]
[dvTP,btnDiscoSelectCol4[15]] = flagDiscoRowSelected[15]
[dvTP,btnDiscoSelectCol4[16]] = flagDiscoRowSelected[16]
[dvTP,btnDiscoSelectCol4[17]] = flagDiscoRowSelected[17]
[dvTP,btnDiscoSelectCol4[18]] = flagDiscoRowSelected[18]
[dvTP,btnDiscoSelectCol4[19]] = flagDiscoRowSelected[19]
[dvTP,btnDiscoSelectCol4[20]] = flagDiscoRowSelected[20]
[dvTP,btnDiscoSelectCol4[21]] = flagDiscoRowSelected[21]
[dvTP,btnDiscoSelectCol4[22]] = flagDiscoRowSelected[22]
[dvTP,btnDiscoSelectCol4[23]] = flagDiscoRowSelected[23]
[dvTP,btnDiscoSelectCol4[24]] = flagDiscoRowSelected[24]
[dvTP,btnDiscoSelectCol4[25]] = flagDiscoRowSelected[25]

WAIT 1
{
 [dvTP,btnDiscoMatched[1]]  = flagDiscoMatched[1]
 [dvTP,btnDiscoMatched[2]]  = flagDiscoMatched[2]
 [dvTP,btnDiscoMatched[3]]  = flagDiscoMatched[3]
 [dvTP,btnDiscoMatched[4]]  = flagDiscoMatched[4]
 [dvTP,btnDiscoMatched[5]]  = flagDiscoMatched[5]
 [dvTP,btnDiscoMatched[6]]  = flagDiscoMatched[6]
 [dvTP,btnDiscoMatched[7]]  = flagDiscoMatched[7]
 [dvTP,btnDiscoMatched[8]]  = flagDiscoMatched[8]
 [dvTP,btnDiscoMatched[9]]  = flagDiscoMatched[9]
 [dvTP,btnDiscoMatched[10]] = flagDiscoMatched[10]
 [dvTP,btnDiscoMatched[11]] = flagDiscoMatched[11]
 [dvTP,btnDiscoMatched[12]] = flagDiscoMatched[12]
 [dvTP,btnDiscoMatched[13]] = flagDiscoMatched[13]
 [dvTP,btnDiscoMatched[14]] = flagDiscoMatched[14]
 [dvTP,btnDiscoMatched[15]] = flagDiscoMatched[15]
 [dvTP,btnDiscoMatched[16]] = flagDiscoMatched[16]
 [dvTP,btnDiscoMatched[17]] = flagDiscoMatched[17]
 [dvTP,btnDiscoMatched[18]] = flagDiscoMatched[18]
 [dvTP,btnDiscoMatched[19]] = flagDiscoMatched[19]
 [dvTP,btnDiscoMatched[20]] = flagDiscoMatched[20]
 [dvTP,btnDiscoMatched[21]] = flagDiscoMatched[21]
 [dvTP,btnDiscoMatched[22]] = flagDiscoMatched[22]
 [dvTP,btnDiscoMatched[23]] = flagDiscoMatched[23]
 [dvTP,btnDiscoMatched[24]] = flagDiscoMatched[24]
 [dvTP,btnDiscoMatched[25]] = flagDiscoMatched[25]
}

[dvTP,btnHQnOnline[1]]  = flagHQN_Online[1]
[dvTP,btnHQnOnline[2]]  = flagHQN_Online[2]
[dvTP,btnHQnOnline[3]]  = flagHQN_Online[3]
[dvTP,btnHQnOnline[4]]  = flagHQN_Online[4]
[dvTP,btnHQnOnline[5]]  = flagHQN_Online[5]
[dvTP,btnHQnOnline[6]]  = flagHQN_Online[6]
[dvTP,btnHQnOnline[7]]  = flagHQN_Online[7]
[dvTP,btnHQnOnline[8]]  = flagHQN_Online[8]
[dvTP,btnHQnOnline[9]]  = flagHQN_Online[9]
[dvTP,btnHQnOnline[10]] = flagHQN_Online[10]
[dvTP,btnHQnOnline[11]] = flagHQN_Online[11]
[dvTP,btnHQnOnline[12]] = flagHQN_Online[12]
[dvTP,btnHQnOnline[13]] = flagHQN_Online[13]
[dvTP,btnHQnOnline[14]] = flagHQN_Online[14]
[dvTP,btnHQnOnline[15]] = flagHQN_Online[15]
[dvTP,btnHQnOnline[16]] = flagHQN_Online[16]
[dvTP,btnHQnOnline[17]] = flagHQN_Online[17]
[dvTP,btnHQnOnline[18]] = flagHQN_Online[18]
[dvTP,btnHQnOnline[19]] = flagHQN_Online[19]
[dvTP,btnHQnOnline[20]] = flagHQN_Online[20]
[dvTP,btnHQnOnline[21]] = flagHQN_Online[21]
[dvTP,btnHQnOnline[22]] = flagHQN_Online[22]
[dvTP,btnHQnOnline[23]] = flagHQN_Online[23]
[dvTP,btnHQnOnline[24]] = flagHQN_Online[24]
[dvTP,btnHQnOnline[25]] = flagHQN_Online[25]

(*****************************************************************)
(*                       END OF PROGRAM                          *)
(*                                                               *)
(*         !!!  DO NOT PUT ANY CODE BELOW THIS COMMENT  !!!      *)
(*                                                               *)
(*****************************************************************)

